/******************************************************************************

                               Copyright (c) 2007
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

#define DSL_INTERN

#include "drv_dsl_cpe_api.h"

#undef DSL_DBG_BLOCK
#define DSL_DBG_BLOCK DSL_DBG_LED

/** \addtogroup DRV_DSL_CPE_COMMON
 @{ */

#if defined(INCLUDE_ADSL_LED)

#if defined (INCLUDE_DSL_CPE_API_DANUBE)
   #warning "WARNING: ADSL Data LED handling included but not implemented for Danube!"
#elif (defined (INCLUDE_DSL_CPE_API_AMAZON_SE) || defined (INCLUDE_DSL_CPE_API_AMAZON_S))
/* wakeup and clean led module */
static int stop_led_module = 0;

static DSL_Context_t *pGlobalContext;

static spinlock_t dsl_led_lock = SPIN_LOCK_UNLOCKED;

static int flash_freq = 0;
static DSL_uint16_t flash, off;

static DSL_DRV_ThreadCtrl_t LedControl;

/**
   \brief API for atm driver to notify led thread a data comming/sending
   This function provide a API used by ATM driver, the atm driver call this
   function once a atm packet sent/received.

   \return  0 success else failed

   \ingroup Internal
 */
static DSL_int_t DSL_DRV_LED_Flash ( DSL_void_t )
{
   DSL_uint32_t flags;
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_Context_t *pContext = pGlobalContext;

   DSL_DEBUG(DSL_DBG_MSG, (pContext, "DSL[%02d]: IN - DSL_DRV_LED_Flash"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CHECK_POINTER(pContext, pContext->pDevCtx);
   DSL_CHECK_ERR_CODE();

   spin_lock_irqsave(&dsl_led_lock, flags);
   if (!stop_led_module)
   {
      /* asking to flash led */
      pContext->bLedNeedToFlash = 1;
   }
   spin_unlock_irqrestore(&dsl_led_lock, flags);
   if (!stop_led_module)
   {
      wake_up_interruptible(&pContext->ledPollingEvent);
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_LED_Flash"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return 0;
}

/*
   Led Thread Main function
*/
static DSL_int_t DSL_DRV_LED_Poll(
   DSL_DRV_ThreadParams_t *param)
{
   DSL_uint32_t flags;
   DSL_int_t nOsRet = 0;
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_Context_t *pContext = (DSL_Context_t *)param->nArg1;
   DSL_LineStateValue_t nCurrentState = DSL_LINESTATE_UNKNOWN;

   /* begin polling ... */
   stop_led_module = 0;

   while (!stop_led_module)
   {
      interruptible_sleep_on_timeout (&pContext->ledPollingEvent, HZ);
      /* Only proceed if the specified line is in SHOWTIME state*/
      DSL_CTX_READ_SCALAR(pContext, nErrCode, nLineState, nCurrentState);
      if( nErrCode != DSL_SUCCESS )
      {
         DSL_DEBUG(DSL_DBG_ERR, (pContext, "DSL[%02d]: ERROR - [%s %d]: Failed to get "
            "current line state!"DSL_DRV_CRLF, DSL_DEV_NUM(pContext), __func__, __LINE__));

         spin_lock_irqsave(&dsl_led_lock, flags);
         stop_led_module = 1;
         pContext->bLedNeedToFlash = 0;
         pContext->bLedStatusOn    = 0;
         spin_unlock_irqrestore(&dsl_led_lock, flags);
         break;
      }

      /* Only proceed if the specified line is in SHOWTIME state*/
      if ((nCurrentState != DSL_LINESTATE_SHOWTIME_TC_SYNC) &&
         (nCurrentState != DSL_LINESTATE_SHOWTIME_NO_SYNC))
      {
         spin_lock_irqsave(&dsl_led_lock, flags);
         pContext->bLedNeedToFlash = 0;
         pContext->bLedStatusOn    = 0;
         spin_unlock_irqrestore(&dsl_led_lock, flags);
         continue;
      }
      spin_lock_irqsave(&dsl_led_lock, flags);
      if (pContext->bLedNeedToFlash)
      {
         /* FLASH */
         /* use GPIO9 for TR68 data led off. */
         if (DSL_DRV_DANUBE_CmvWrite(pContext, DSL_CMV_GROUP_INFO,
            DSL_CMV_ADDRESS_INFO_GPIO_CONTROL, 5, 1, &flash) != DSL_SUCCESS)
         {
            DSL_DEBUG(DSL_DBG_ERR,
               (pContext, "[%s %d]: CMV Write fail!" DSL_DRV_CRLF,
               __func__, __LINE__));
         }
         pContext->bLedNeedToFlash = 0;
      }
      else
      {
         /* OFF */
         /* use GPIO9 for TR68 data led off. */
         if (DSL_DRV_DANUBE_CmvWrite(pContext, DSL_CMV_GROUP_INFO,
            DSL_CMV_ADDRESS_INFO_GPIO_CONTROL, 5, 1, &off) != DSL_SUCCESS)
         {
            DSL_DEBUG(DSL_DBG_ERR, (pContext, "[%s %d]: CMV Write fail!" DSL_DRV_CRLF,
               __func__, __LINE__));
         }
      }
      spin_unlock_irqrestore(&dsl_led_lock, flags);
   }

   nOsRet = DSL_DRV_ErrorToOS(nErrCode);

   DSL_DEBUG(DSL_DBG_MSG, (pContext, "DSL[%02d]: OUT - DSL_DRV_LED_Poll thread"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nOsRet;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_led.h'
*/
DSL_uint32_t DSL_DRV_LED_ModuleInit(DSL_Context_t *pContext)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_LED_ModuleInit"DSL_DRV_CRLF,
       DSL_DEV_NUM(pContext)));
       
   DSL_CHECK_POINTER(pContext, pContext->pDevCtx);
   DSL_CHECK_ERR_CODE();

   pGlobalContext = pContext;
   pContext->bLedNeedToFlash = 0;
   pContext->bLedStatusOn = 0;

#ifndef DSL_DATA_LED_FLASH_FREQUENCY
#error Please use --with-adsl-data-led-flash-frequency=[1-32] to define ADSL Data Flash Frequency
#endif
   DSL_DEBUG(DSL_DBG_MSG, (pContext, "DSL[%02d]: ADSL data led flashing at %d Hz" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), DSL_DATA_LED_FLASH_FREQUENCY));

   flash_freq = DSL_DATA_LED_FLASH_FREQUENCY;
   if (flash_freq == 16)
   {
      flash = 0xA01;
   }
   else
   {
      flash_freq = ((flash_freq * 2) << 2) & 0xFC;
      flash = 0xA02 | flash_freq;
   }
   off = 0xA00;

   /* adsl led for led function */
   DSL_DRV_INIT_EVENT ("atm_led", pContext->ledPollingEvent);

   DSL_BSP_ATMLedCBRegister (DSL_DRV_LED_Flash);

   return DSL_DRV_THREAD (&LedControl, "atm_led_completion", DSL_DRV_LED_Poll, (int)pContext);
}

#endif /* INCLUDE_DSL_CPE_API_AMAZON_SE || INCLUDE_DSL_CPE_API_AMAZON_S*/

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_led.h'
*/
DSL_uint32_t DSL_DRV_LED_FirmwareInit(DSL_Context_t *pContext)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_LED_FirmwareInit" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* This function is currently not used.
      LED configuration is done within device specific function
      DSL_DRV_DEV_AutobootHandleStart() */
   /*
   DSL_CHECK_CTX_POINTER(pContext);
   nErrCode = DSL_DRV_DEV_LED_FirmwareInit(pContext);
   */

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_LED_FirmwareInit, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

#endif /* INCLUDE_ADSL_LED */

/** @} DRV_DSL_CPE_COMMON */

