/******************************************************************************

                               Copyright (c) 2007
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

#define DSL_INTERN

#include "drv_dsl_cpe_api.h"

#undef DSL_DBG_BLOCK
#define DSL_DBG_BLOCK DSL_DBG_SAR

/** \file
   Segmentation and Reassembly (SAR) implementation
*/

#if (INCLUDE_DSL_CPE_API_VDSL_SUPPORT == 1)

#undef DSL_DBG_BLOCK
#define DSL_DBG_BLOCK DSL_DBG_SAR

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_SAR_CONFIG_SET
*/
DSL_uint32_t DSL_DRV_SAR_ConfigSet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_SAR_Config_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_CHANNEL_RANGE(pData->nChannel);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_SAR_ConfigSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check VCI range*/
   if (pData->data.nDefVci < 32)
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - nDefVci parameter range check error!"
         " Valid range 32...65535."DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));

      nErrCode = DSL_ERR_INVALID_PARAMETR;
   }

   /* Check VPI range*/
   if (pData->data.nDefVpi > 255)
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - nDefVpi parameter range check error!"
         " Valid range 0...255."DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));

      nErrCode = DSL_ERR_INVALID_PARAMETR;
   }

   if (nErrCode == DSL_SUCCESS)
   {
      /* Copy SAR configuration data to internal VINAX device context*/
      DSL_CTX_WRITE(pContext, nErrCode, pDevCtx->data.deviceCfg.SARcfg, pData->data);
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_SAR_ConfigSet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_SAR_CONFIG_GET
*/
DSL_uint32_t DSL_DRV_SAR_ConfigGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_SAR_Config_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_CHANNEL_RANGE(pData->nChannel);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_SAR_ConfigSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Copy SAR configuration data from the internal VINAX device context*/
   DSL_CTX_READ(pContext, nErrCode, pDevCtx->data.deviceCfg.SARcfg, pData->data);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_SAR_ConfigSet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_SAR_SEGMENTATION_TABLE_SET
*/
DSL_uint32_t DSL_DRV_SAR_SegmentationTableSet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_SAR_SegmentationTable_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_uint16_t nArrayIdx = 0, i = 0;
   DSL_SAR_SegmentationTableEntry_t const *pEntry = DSL_NULL;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_CHANNEL_RANGE(pData->nChannel);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_SAR_SegmentationTableSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   if(DSL_DRV_MUTEX_LOCK(pContext->dataMutex))
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - data mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   /* Configure Segmentation Table*/
   nArrayIdx = 0;
   pContext->pDevCtx->data.deviceCfg.SegmentationTable.nNumEntries = pData->data.nNumEntries;

   for(nArrayIdx = 0; nArrayIdx < pData->data.nNumEntries; nArrayIdx++)
   {
      pEntry = pData->data.pTableEntry + nArrayIdx;

      for( i = 0; i < 6; i++ )
      {
         pContext->pDevCtx->data.deviceCfg.SegmentationTable.pTableEntry[nArrayIdx].pMacAddress[i] =
            pEntry->pMacAddress[i];
      }

      pContext->pDevCtx->data.deviceCfg.SegmentationTable.pTableEntry[nArrayIdx].nVlanId   = pEntry->nVlanId;
      pContext->pDevCtx->data.deviceCfg.SegmentationTable.pTableEntry[nArrayIdx].nVlanPrio = pEntry->nVlanPrio;
      pContext->pDevCtx->data.deviceCfg.SegmentationTable.pTableEntry[nArrayIdx].nVpi      = pEntry->nVpi;
      pContext->pDevCtx->data.deviceCfg.SegmentationTable.pTableEntry[nArrayIdx].nVci      = pEntry->nVci;
   }

   DSL_DRV_MUTEX_UNLOCK(pContext->dataMutex);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_SAR_SegmentationTableSet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_SAR_SEGMENTATION_TABLE_GET
*/
DSL_uint32_t DSL_DRV_SAR_SegmentationTableGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_SAR_SegmentationTable_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_uint16_t nArrayIdx = 0, i = 0;
   DSL_SAR_SegmentationTableEntry_t const *pEntry = DSL_NULL;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_CHANNEL_RANGE(pData->nChannel);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_SAR_SegmentationTableGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   if(DSL_DRV_MUTEX_LOCK(pContext->dataMutex))
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - data mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   /* Get Segmentation Table from the VINAX device context*/
   nArrayIdx = 0;
   pData->data.nNumEntries = pContext->pDevCtx->data.deviceCfg.SegmentationTable.nNumEntries;
   for(nArrayIdx = 0; nArrayIdx < pData->data.nNumEntries; nArrayIdx++)
   {
      pEntry = pContext->pDevCtx->data.deviceCfg.SegmentationTable.pTableEntry + nArrayIdx;
      for( i = 0; i < 6; i++ )
      {
         pData->data.pTableEntry[nArrayIdx].pMacAddress[i] = pEntry->pMacAddress[i];
      }

      pData->data.pTableEntry[nArrayIdx].nVlanId   = pEntry->nVlanId;
      pData->data.pTableEntry[nArrayIdx].nVlanPrio = pEntry->nVlanPrio;
      pData->data.pTableEntry[nArrayIdx].nVpi      = pEntry->nVpi;
      pData->data.pTableEntry[nArrayIdx].nVci      = pEntry->nVci;
   }

   DSL_DRV_MUTEX_UNLOCK(pContext->dataMutex);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_SAR_SegmentationTableGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_SAR_REASSEMBLY_TABLE_SET
*/
DSL_uint32_t DSL_DRV_SAR_ReassemblyTableSet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_SAR_ReassemblyTable_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_uint16_t nArrayIdx = 0;
   DSL_SAR_ReassemblyTableEntry_t const *pEntry = DSL_NULL;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_CHANNEL_RANGE(pData->nChannel);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_SAR_ReassemblyTableSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   if(DSL_DRV_MUTEX_LOCK(pContext->dataMutex))
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - data mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   nArrayIdx = 0;
   pContext->pDevCtx->data.deviceCfg.ReassemblyTable.nNumEntries = pData->data.nNumEntries;

   for( nArrayIdx = 0; nArrayIdx < pData->data.nNumEntries; nArrayIdx++)
   {
      pEntry = pData->data.pTableEntry + nArrayIdx;

      pContext->pDevCtx->data.deviceCfg.ReassemblyTable.pTableEntry[nArrayIdx].nVci = pEntry->nVci;
      pContext->pDevCtx->data.deviceCfg.ReassemblyTable.pTableEntry[nArrayIdx].nVpi = pEntry->nVpi;
   }

   DSL_DRV_MUTEX_UNLOCK(pContext->dataMutex);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_SAR_ReassemblyTableSet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_SAR_REASSEMBLY_TABLE_GET
*/
DSL_uint32_t DSL_DRV_SAR_ReassemblyTableGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_SAR_ReassemblyTable_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_uint16_t nArrayIdx = 0;
   DSL_SAR_ReassemblyTableEntry_t const *pEntry = DSL_NULL;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_CHANNEL_RANGE(pData->nChannel);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_SAR_ReassemblyTableGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   if(DSL_DRV_MUTEX_LOCK(pContext->dataMutex))
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - data mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   nArrayIdx = 0;
   pData->data.nNumEntries = pContext->pDevCtx->data.deviceCfg.ReassemblyTable.nNumEntries;
   for( nArrayIdx = 0; nArrayIdx < pData->data.nNumEntries; nArrayIdx++)
   {
      pEntry = pContext->pDevCtx->data.deviceCfg.ReassemblyTable.pTableEntry + nArrayIdx;

      pData->data.pTableEntry[nArrayIdx].nVci = pEntry->nVci;
      pData->data.pTableEntry[nArrayIdx].nVpi = pEntry->nVpi;
   }

   DSL_DRV_MUTEX_UNLOCK(pContext->dataMutex);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_SAR_ReassemblyTableGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_SAR_SEGMENTATION_COUNTERS_GET
*/
DSL_uint32_t DSL_DRV_SAR_SegmentationCountersGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_SAR_SegmentationCounters_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_LineStateValue_t   nCurrentState = DSL_LINESTATE_UNKNOWN;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_CHANNEL_RANGE(pData->nChannel);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_SAR_SegmentationCountersGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Get current line state*/
   DSL_CTX_READ_SCALAR(pContext, nErrCode, nLineState, nCurrentState);

   /* Only proceed if the line is in SHOWTIME or IDLE state.*/
   if ((nCurrentState == DSL_LINESTATE_SHOWTIME_TC_SYNC) ||
       (nCurrentState == DSL_LINESTATE_IDLE))
   {
      /* Call VINAX device specific implementation*/
      nErrCode = DSL_DRV_SAR_VNX_SegmentationCountersGet(pContext, &(pData->data));
   }
   else
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Segmentation Counters only available in SHOWTIME or IDLE states!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERROR;
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_SAR_SegmentationCountersGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_SAR_REASSEMBLY_COUNTERS_GET
*/
DSL_uint32_t DSL_DRV_SAR_ReassemblyCountersGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_SAR_ReassemblyCounters_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_LineStateValue_t   nCurrentState = DSL_LINESTATE_UNKNOWN;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_CHANNEL_RANGE(pData->nChannel);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_SAR_ReassemblyCountersGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Get current line state*/
   DSL_CTX_READ_SCALAR(pContext, nErrCode, nLineState, nCurrentState);

   /* Only proceed if the line is in SHOWTIME or IDLE state.*/
   if ((nCurrentState == DSL_LINESTATE_SHOWTIME_TC_SYNC) ||
       (nCurrentState == DSL_LINESTATE_IDLE))
   {
      /* Call VINAX device specific implementation*/
      nErrCode = DSL_DRV_SAR_VNX_ReassemblyCountersGet(pContext, &(pData->data));
   }
   else
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Reassembly Counters only available in SHOWTIME or IDLE states!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERROR;
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_SAR_ReassemblyCountersGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

#endif /* (INCLUDE_DSL_CPE_API_VDSL_SUPPORT == 1) */

/** \addtogroup DRV_DSL_CPE_SAR
 @{ */


/** @} LIB_DSL_SAR */

