/******************************************************************************
**
** FILE NAME    : danube_ppa_api.c
** PROJECT      : Twinpass-E
** MODULES      : PPA (Routing Acceleration)
**
** DATE         : 12 SEP 2006
** AUTHOR       : Xu Liang
** DESCRIPTION  : PPA (Routing Acceleration) Protocol Stack Hook API
** COPYRIGHT    :   Copyright (c) 2006
**          Infineon Technologies AG
**          Am Campeon 1-12, 85579 Neubiberg, Germany
**
**   Any use of this software is subject to the conclusion of a respective
**   License agreement. Without such a License agreement no rights to the
**   software are granted
**
** HISTORY
** $Date        $Author         $Comment
** 12 SEP 2006  Xu Liang        Initiate Version
** 26 OCT 2006  Xu Liang        Add Infineon license header.
** 11 JAN 2007  Xu Liang        Check if it's WAN interface when add MC session
*******************************************************************************/



/*
 * ####################################
 *              Version No.
 * ####################################
 */
#define VER_FAMILY      0x08        //  bit 0: res
                                    //      1: Danube
                                    //      2: Twinpass
                                    //      3: Amazon-SE
                                    //      4: res
                                    //      5: AR9
                                    //      6: GR9
#define VER_DRTYPE      0x20        //  bit 0: Normal Data Path driver
                                    //      1: Indirect-Fast Path driver
                                    //      2: HAL driver
                                    //      3: Hook driver
                                    //      4: Stack/System Adaption Layer driver
                                    //      5: PPA API driver
#define VER_INTERFACE   0x05        //  bit 0: MII 0
                                    //      1: MII 1
                                    //      2: ATM WAN
                                    //      3: PTM WAN
#define VER_ACCMODE     0x03        //  bit 0: Routing
                                    //      1: Bridging
#define VER_MAJOR       0
#define VER_MINOR       1



/*
 * ####################################
 *              Head File
 * ####################################
 */

/*
 *  Common Head File
 */
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/netdevice.h>
//#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <net/sock.h>
#include <net/ip_vs.h>
#include <asm/time.h>

/*
 *  Chip Specific Head File
 */
#include <asm/amazon_se/ifx_ppa_api.h>
#include <asm/amazon_se/ifx_ppa_interface.h>
#include <asm/amazon_se/ifx_ppa_hook.h>
#include <asm/amazon_se/ifx_ppa_directpath.h>



/*
 * ####################################
 *              Definition
 * ####################################
 */

#define DANUBE_CHIP                             1
#define AMAZON_SE_CHIP                          2

/*
 *  Compilation Switch
 */

#define CHIP_TYPE                               AMAZON_SE_CHIP

#define ENABLE_DEBUG                            1

#define ENABLE_ASSERT                           1

// moved to ifx_ppa_api.h
//#define ENABLE_SESSION_DEBUG_FLAGS              1

#define DEBUG_DUMP_LIST_ITEM                    0

#define ENABLE_DANUBE                           1
#if CHIP_TYPE == AMAZON_SE_CHIP
  #undef ENABLE_DANUBE
  #define ENABLE_AMAZON_SE                      1
#endif

#if defined(ENABLE_DEBUG) && ENABLE_DEBUG
  #define ENABLE_DEBUG_PRINT                    1
  #define DISABLE_INLINE                        1
#endif

#if defined(DISABLE_INLINE) && DISABLE_INLINE
  #define INLINE
#else
  #define INLINE                                inline
#endif

#if defined(ENABLE_DEBUG_PRINT) && ENABLE_DEBUG_PRINT
  #undef  dbg
  #define dbg(format, arg...)                   do { if ( g_dbg_enable ) printk(KERN_WARNING __FILE__ ":%d:%s: " format "\n", __LINE__, __FUNCTION__, ##arg); } while ( 0 )
#elif !defined(dbg)
  #define dbg(format, arg...)
#endif

#if defined(ENABLE_ASSERT) && ENABLE_ASSERT
  #define ASSERT(cond, format, arg...)          do { if ( g_dbg_enable && !(cond) ) printk(KERN_ERR __FILE__ ":%d:%s: " format "\n", __LINE__, __FUNCTION__, ##arg); } while ( 0 )
#else
  #define ASSERT(cond, format, arg...)
#endif

#undef  err
#define err(format, arg...)                     printk(KERN_ERR __FILE__ ":%d:%s: " format "\n", __LINE__, __FUNCTION__, ##arg);

#if (defined(ENABLE_DEBUG_PRINT) && ENABLE_DEBUG_PRINT)          \
    || (defined(ENABLE_ASSERT) && ENABLE_ASSERT)
  #define ENABLE_DBG_PROC                       1
#else
  #define ENABLE_DBG_PROC                       0
#endif

/*
 *  device constant
 */
#define PPA_CHR_DEV_MAJOR                       180
#define PPA_DEV_NAME                            "ifx_ppa"

#define FWCODE_ROUTING_ACC_D2                   0x02
#define FWCODE_BRIDGING_ACC_D3                  0x03
#define FWCODE_ROUTING_BRIDGING_ACC_D4          0x04
#define FWCODE_ROUTING_BRIDGING_ACC_A4          0x14

/*
 *  default configuration
 */
#define ETX_WFQ_D4                              1
#define DPLUS_WFQ_D4                            3
#define FASTPATH_WFQ_D4                         0
#define DEFAULT_TIMEOUT_IN_SEC                  3600    //  1 hour
#define DEFAULT_BRIDGING_TIMEOUT_IN_SEC         60      //  1 minute
#if 0 /* XXX: Santosh - set to 1500 */
#define DEFAULT_MTU                             0x0003FFFF
#else
#define DEFAULT_MTU                             0x5DC // 1500 bytes
#endif

#define DEFAULT_CH_ID                           0
#define DEFAULT_HIT_POLLING_TIME                60      //  1 minute
#define DEFAULT_BRIDGING_HIT_POLLING_TIME       2       //  2 seconds
#define MIN_HITS                                0

#define SESSION_LIST_HASH_SHIFT                 10
#define SESSION_LIST_HASH_BIT_LENGTH            9
#define SESSION_LIST_HASH_MASK                  ((1 << SESSION_LIST_HASH_BIT_LENGTH) - 1)
#define SESSION_LIST_HASH_TABLE_SIZE            (1 << (SESSION_LIST_HASH_BIT_LENGTH + 1))
//#define SESSION_LIST_HASH_VALUE(x, is_reply)    (((((uint32_t)(x) >> SESSION_LIST_HASH_BIT_LENGTH) & SESSION_LIST_HASH_MASK) << 1) | ((is_reply) ? 1 : 0))
#define SESSION_LIST_HASH_VALUE(x, is_reply)    (((((uint32_t)(x) >> SESSION_LIST_HASH_SHIFT) & SESSION_LIST_HASH_MASK) << 1) | ((is_reply) ? 1 : 0))

#define SESSION_LIST_MC_HASH_SHIFT              0
#define SESSION_LIST_MC_HASH_BIT_LENGTH         6
#define SESSION_LIST_MC_HASH_MASK               ((1 << SESSION_LIST_MC_HASH_BIT_LENGTH) - 1)
#define SESSION_LIST_MC_HASH_TABLE_SIZE         (1 << SESSION_LIST_MC_HASH_BIT_LENGTH)
#define SESSION_LIST_MC_HASH_VALUE(x)           (((uint32_t)(x) >> SESSION_LIST_MC_HASH_SHIFT) & SESSION_LIST_MC_HASH_MASK)

#define BRIDGING_SESSION_LIST_HASH_BIT_LENGTH   8
#define BRIDGING_SESSION_LIST_HASH_MASK         ((1 << BRIDGING_SESSION_LIST_HASH_BIT_LENGTH) - 1)
#define BRIDGING_SESSION_LIST_HASH_TABLE_SIZE   (1 << BRIDGING_SESSION_LIST_HASH_BIT_LENGTH)
#define BRIDGING_SESSION_LIST_HASH_VALUE(x)     (*(uint32_t *)((uint16_t *)(x) + 1) & BRIDGING_SESSION_LIST_HASH_MASK)

/*
 *  net interface type
 */
#define NETIF_VLAN                              0x0001
#define NETIF_BRIDGE                            0x0002
#define NETIF_PPPOE                             0x0004
#define NETIF_BR2684                            0x0008
#define NETIF_LAN_IF                            0x0100
#define NETIF_WAN_IF                            0x0200
#define NETIF_PHY_IF_GOT                        0x0800
#define NETIF_MAC_ENTRY_CREATED                 0x1000

/*
 *  internal flag
 */
#if 0 /* moved to ifx_ppa_api.h */
#define SESSION_INTERNAL_FLAG_BASE              0
#define SESSION_IS_REPLY                        0x00000001
#define SESSION_BRIDGING_VCI_CHECK              0x00000002
#define SESSION_ADDED_IN_HW                     0x00000010
#define SESSION_CAN_NOT_ACCEL                   0x00000020
#define SESSION_STATIC                          0x00000040
#define SESSION_DROP                            0x00000080
#define SESSION_VALID_NAT_IP                    0x00000100
#define SESSION_VALID_NAT_PORT                  0x00000200
#define SESSION_VALID_NAT_SNAT                  0x00000400  //  src IP is replaced, otherwise dest IP is replaced
#define SESSION_VALID_VLAN_INS                  0x00001000
#define SESSION_VALID_VLAN_RM                   0x00002000
#define SESSION_VALID_OUT_VLAN_INS              0x00004000
#define SESSION_VALID_OUT_VLAN_RM               0x00008000
#define SESSION_VALID_PPPOE                     0x00010000
#define SESSION_VALID_NEW_SRC_MAC               0x00020000
#define SESSION_VALID_SRC_MAC                   SESSION_VALID_NEW_SRC_MAC
#define SESSION_VALID_MTU                       0x00040000
#define SESSION_VALID_NEW_DSCP                  0x00080000
#define SESSION_LAN_ENTRY                       0x00100000
#define SESSION_WAN_ENTRY                       0x00200000
#define SESSION_SRC_MAC_DROP_EN                 0x01000000
#define SESSION_TX_ITF_IPOA                     0x02000000
#define SESSION_TX_ITF_PPPOA                    0x04000000
#define SESSION_TX_ITF_IPOA_PPPOA_MASK          (SESSION_TX_ITF_IPOA | SESSION_TX_ITF_PPPOA)

#define PPA_IS_NAT_SESSION(flags)               ((flags) & (SESSION_VALID_NAT_IP | SESSION_VALID_NAT_PORT))
#endif
#if defined(ENABLE_SESSION_DEBUG_FLAGS) && ENABLE_SESSION_DEBUG_FLAGS
  #define SESSION_DBG_NOT_REACH_MIN_HITS        0x00000001
  #define SESSION_DBG_ALG_OR_WRONG_PROTO        0x00000002
  #define SESSION_DBG_ZERO_DST_MAC              0x00000004
  #define SESSION_DBG_RX_IF_TYPE_ERR            0x00000010
  #define SESSION_DBG_TX_IF_TYPE_ERR            0x00000020
  #define SESSION_DBG_DST_BRG_IF_NOT_IN_BRG_TBL 0x00000100
  #define SESSION_DBG_DST_IF_NOT_IN_IF_LIST     0x00000200
  #define SESSION_DBG_GET_DST_MAC_FAIL          0x00000400
  #define SESSION_DBG_ADD_PPPOE_ENTRY_FAIL      0x00001000
  #define SESSION_DBG_ADD_MTU_ENTRY_FAIL        0x00002000
  #define SESSION_DBG_ADD_MAC_ENTRY_FAIL        0x00004000
  #define SESSION_DBG_ADD_OUT_VLAN_ENTRY_FAIL   0x00008000
  #define SESSION_DBG_RX_PPPOE                  0x00010000
  #define SESSION_DBG_TX_PPPOE                  0x00020000
  #define SESSION_DBG_TX_BR2684_EOA             0x00040000
  #define SESSION_DBG_TX_BR2684_IPOA            0x00080000
  #define SESSION_DBG_TX_PPPOA                  0x00100000
  #define SESSION_DBG_RX_VLAN                   0x00200000
  #define SESSION_DBG_TX_VLAN                   0x00400000

  #define SET_DBG_FLAG(pitem, flag)             ((pitem)->debug_flags |= (flag))
  #define CLR_DBG_FLAG(pitem, flag)             ((pitem)->debug_flags &= ~(flag))
#else
  #define SET_DBG_FLAG(pitem, flag)             do { } while ( 0 )
  #define CLR_DBG_FLAG(pitem, flag)             do { } while ( 0 )
#endif

/*
 *  helper macro
 */
#define NUM_ENTITY(x)                           (sizeof(x) / sizeof(*(x)))



/*
 * ####################################
 *              Data Type
 * ####################################
 */

/*
 *  Private Data Structure
 */

struct netif_info {
    struct netif_info          *next;

    PPA_IFNAME                  name[PPA_IF_NAME_SIZE];
//    PPA_NETIF                  *phys_netif;
    PPA_IFNAME                  phys_netif_name[PPA_IF_NAME_SIZE];
    uint8_t                     mac[PPA_ETH_ALEN];
    uint32_t                    flags;  //  NETIF_VLAN, NETIF_BRIDGE, NETIF_PPPOE, NETIF_BR2684
                                        //  NETIF_LAN_IF, NETIF_WAN_IF, NETIF_PHY_IF_GOT
                                        //  NETIF_MAC_ENTRY_CREATED

    uint32_t                    mac_entry;

#ifdef CONFIG_IFX_PPA_A4
    uint32_t                    dslwan_qid;
#endif
};

/* Moved to ifx_ppa_api.h */
#if 0
struct session_list_item {
    struct session_list_item   *next;
//    uint32_t                    use_count;

    PPA_SESSION                *session;
    PPA_SESSION_EXTRA           session_extra;
    uint16_t                    ip_proto;
    uint32_t                    src_ip;
    uint16_t                    src_port;
    uint8_t                     src_mac[PPA_ETH_ALEN];
    uint32_t                    dst_ip;
    uint16_t                    dst_port;
    uint8_t                     dst_mac[PPA_ETH_ALEN];
    uint32_t                    nat_ip;         //  IP address to be replaced by NAT if NAT applies
    uint16_t                    nat_port;       //  Port to be replaced by NAT if NAT applies
//    uint8_t                     new_src_mac[PPA_ETH_ALEN];  //  src MAC address to be replaced
    PPA_NETIF                  *rx_if;
    PPA_NETIF                  *tx_if;
    uint32_t                    timeout;
    uint32_t                    last_hit_time;  //  Updated by bookkeeping thread
    uint32_t                    num_adds;       //  Number of attempts to add session
    uint32_t                    pppoe_session_id;
    uint32_t                    new_dscp;
    uint32_t                    new_vci;
    uint32_t                    mtu;
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    uint32_t                    mc_dest_list;
#endif
    uint32_t                    flags;          //  Internal flag : SESSION_IS_REPLY,
                                                //                  SESSION_ADDED_IN_HW, SESSION_CAN_NOT_ACCEL
                                                //                  SESSION_VALID_NAT_IP, SESSION_VALID_NAT_PORT,
                                                //                  SESSION_VALID_VLAN_INS, SESSION_VALID_VLAN_RM,
                                                //                  SESSION_VALID_OUT_VLAN_INS, SESSION_VALID_OUT_VLAN_RM,
                                                //                  SESSION_VALID_PPPOE, SESSION_VALID_SRC_MAC,
                                                //                  SESSION_VALID_MTU,
                                                //                  SESSION_VALID_NEW_DSCP,
                                                //                  SESSION_LAN_ENTRY, SESSION_WAN_ENTRY,
                                                //                  SESSION_TX_ITF_IPOA, SESSION_TX_ITF_PPPOA

    uint32_t                    routing_entry;
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    uint32_t                    mc_routing_entry;
#endif
    uint32_t                    pppoe_entry;
    uint32_t                    mtu_entry;
    uint32_t                    src_mac_entry;

#ifdef CONFIG_IFX_PPA_A4
    uint32_t                    dslwan_qid;
    uint32_t                    out_vlan_tag;
    uint32_t                    out_vlan_ix;
#endif

#if defined(ENABLE_SESSION_DEBUG_FLAGS) && ENABLE_SESSION_DEBUG_FLAGS
    uint32_t                    debug_flags;
#endif
};

#if defined(USE_NEW_MC_API) && USE_NEW_MC_API
  struct mc_group_list_item {
    struct mc_group_list_item  *next;

    IPADDR                      ip_mc_group;
    uint32_t                    num_ifs;
    PPA_NETIF                  *netif[MAX_MC_IFS_NUM];
    uint32_t                    ttl[MAX_MC_IFS_NUM];
    uint32_t                    if_mask;
    uint32_t                    new_vci;
    uint32_t                    flags;          //  Internal flag : SESSION_IS_REPLY,
                                                //                  SESSION_ADDED_IN_HW, SESSION_CAN_NOT_ACCEL
                                                //                  SESSION_VALID_NAT_IP, SESSION_VALID_NAT_PORT,
                                                //                  SESSION_VALID_VLAN_INS, SESSION_VALID_VLAN_RM,
                                                //                  SESSION_VALID_OUT_VLAN_INS, SESSION_VALID_OUT_VLAN_RM,
                                                //                  SESSION_VALID_PPPOE, SESSION_VALID_SRC_MAC,
                                                //                  SESSION_VALID_MTU,
                                                //                  SESSION_VALID_NEW_DSCP,
                                                //                  SESSION_LAN_ENTRY, SESSION_WAN_ENTRY

    uint32_t                    mc_entry;
    uint32_t                    src_mac_entry;

#ifdef CONFIG_IFX_PPA_A4
    uint32_t                    out_vlan_tag;
    uint32_t                    out_vlan_ix;
#endif

#if defined(ENABLE_SESSION_DEBUG_FLAGS) && ENABLE_SESSION_DEBUG_FLAGS
    uint32_t                    debug_flags;
#endif
  };
#endif

struct bridging_session_list_item {
    struct bridging_session_list_item
                               *next;
//    uint32_t                    use_count;

    uint8_t                     mac[PPA_ETH_ALEN];
    uint32_t                    vci;
    PPA_NETIF                  *netif;
    uint32_t                    new_vci;
    uint32_t                    timeout;
    uint32_t                    last_hit_time;  //  Updated by bookkeeping thread
    uint32_t                    flags;          //  Internal flag : SESSION_BRIDGING_VCI_CHECK
                                                //                  SESSION_ADDED_IN_HW, SESSION_CAN_NOT_ACCEL
                                                //                  SESSION_STATIC, SESSION_DROP
                                                //                  SESSION_VALID_VLAN_INS, SESSION_VALID_VLAN_RM
                                                //                  SESSION_SRC_MAC_DROP_EN
    uint32_t                    bridging_entry;

#ifdef CONFIG_IFX_PPA_A4
    uint32_t                    dslwan_qid;
#endif

#if defined(ENABLE_SESSION_DEBUG_FLAGS) && ENABLE_SESSION_DEBUG_FLAGS
    uint32_t                    debug_flags;
#endif
};

#endif //0

/*
 * ####################################
 *             Declaration
 * ####################################
 */

extern int danube_eth_download_firmware(int);
#ifdef CONFIG_IFX_PPA_D4
  extern void setup_vlan_queue_map(uint32_t, uint32_t);
#endif
#ifdef CONFIG_IFX_PPA_A4
  extern int get_dslwan_qid_with_vcc(PPA_VCC *);
#endif

static int ppa_dev_open(struct inode *, struct file *);
static int ppa_dev_release(struct inode *, struct file *);
static int ppa_dev_ioctl(struct inode *, struct file *, unsigned int, unsigned long);

static INLINE struct netif_info * alloc_netif_info(void);
static INLINE void free_netif_info(struct netif_info *);
static INLINE void lock_netif_info_list(void);
static INLINE void unlock_netif_info_list(void);
static INLINE void free_netif_info_list(void);
static INLINE void add_netif_info(struct netif_info *);
static INLINE int32_t netif_info_is_added(PPA_IFNAME[PPA_IF_NAME_SIZE], struct netif_info **);

static INLINE int32_t ppa_d4_init(PPA_INIT_INFO *, uint32_t);
static INLINE void ppa_d4_exit(void);
static INLINE void ppa_init_session_list_item(struct session_list_item *);
#if defined(USE_NEW_MC_API) && USE_NEW_MC_API
  static INLINE void ppa_init_mc_group_list_item(struct mc_group_list_item *);
#endif
static INLINE void ppa_bridging_init_session_list_item(struct bridging_session_list_item *);
static INLINE int32_t ppa_get_iftype(PPA_NETIF *);
static INLINE int32_t ppa_get_ifid(PPA_NETIF *);
 INLINE int32_t ppa_lookup_session(PPA_SESSION *, uint32_t, struct session_list_item **);
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
  static INLINE int32_t ppa_lookup_mc_session(uint32_t, struct session_list_item **);
#else
  static INLINE int32_t ppa_lookup_mc_group(IPADDR, struct mc_group_list_item **);
#endif
static INLINE int32_t ppa_check_if_session_can_accel(PPA_BUF *, PPA_SESSION *, struct session_list_item *);
static INLINE int32_t ppa_add_session(PPA_BUF *, PPA_SESSION *, struct session_list_item **, uint32_t);
static INLINE int32_t ppa_update_session(PPA_BUF *, struct session_list_item *, uint32_t);
static INLINE int32_t ppa_update_session_extra(PPA_SESSION_EXTRA *, struct session_list_item *, uint32_t);
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
  static INLINE int32_t ppa_add_mc_session(PPA_MC_SESSION *, struct session_list_item **, uint32_t);
  static INLINE int32_t ppa_update_mc_session_extra(PPA_SESSION_EXTRA *, struct session_list_item *, uint32_t);
#else
  static INLINE int32_t ppa_add_mc_group(PPA_MC_GROUP *, struct mc_group_list_item **, uint32_t);
#endif
static INLINE int32_t ppa_hw_add_session(struct session_list_item *);
static INLINE int32_t ppa_hw_update_session_extra(struct session_list_item *, uint32_t);
static INLINE void ppa_hw_del_session(struct session_list_item *);
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
  static INLINE int32_t ppa_hw_add_mc_session(struct session_list_item *);
  static INLINE int32_t ppa_hw_update_mc_session_extra(struct session_list_item *, uint32_t);
  static INLINE void ppa_hw_del_mc_session(struct session_list_item *);
#else
  static INLINE int32_t ppa_hw_add_mc_group(struct mc_group_list_item *);
  static INLINE void ppa_hw_del_mc_group(struct mc_group_list_item *);
#endif
static INLINE void ppa_insert_session_item(struct session_list_item *);
static INLINE void ppa_remove_session_item(struct session_list_item *);
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
  static INLINE void ppa_insert_mc_session_item(struct session_list_item *);
  static INLINE void ppa_remove_mc_session_item(struct session_list_item *);
#else
  static INLINE void ppa_insert_mc_group_item(struct mc_group_list_item *);
  static INLINE void ppa_remove_mc_group_item(struct mc_group_list_item *);
#endif
static INLINE uint32_t ppa_get_default_session_timeout(void);
static void ppa_check_hit_stat(unsigned long);

 INLINE int32_t ppa_bridging_lookup_session(uint8_t *, PPA_NETIF *, struct bridging_session_list_item **);
static INLINE int32_t ppa_bridging_add_session(uint8_t *, PPA_NETIF *, struct bridging_session_list_item **, uint32_t);
static INLINE int32_t ppa_bridging_hw_add_session(struct bridging_session_list_item *);
static INLINE void ppa_bridging_hw_del_session(struct bridging_session_list_item *);
static INLINE void ppa_bridging_insert_session_item(struct bridging_session_list_item *);
static INLINE void ppa_bridging_remove_session_item(struct bridging_session_list_item *);
static INLINE uint32_t ppa_bridging_get_default_session_timeout(void);
static void ppa_bridging_check_hit_stat(unsigned long);

/*
 *  Debug Function
 */
static INLINE void dump_list_item(struct session_list_item *, char *);
static INLINE void dump_bridging_list_item(struct bridging_session_list_item *, char *);

/*
 *  Proc File
 */
static INLINE void proc_file_create(void);
static INLINE void proc_file_delete(void);
#if defined(ENABLE_DBG_PROC) && ENABLE_DBG_PROC
  static int proc_read_dbg(char *, char **, off_t, int, int *, void *);
  static int proc_write_dbg(struct file *, const char *, unsigned long, void *);
#endif
static int proc_read_session(char *, char **, off_t, int, int *, void *);
static int proc_write_session(struct file *, const char *, unsigned long, void *);
static int proc_read_hook(char *, char **, off_t, int, int *, void *);
static int proc_write_hook(struct file *, const char *, unsigned long, void *);
static int proc_read_netif(char *, char **, off_t, int, int *, void *);
static int proc_write_netif(struct file *, const char *, unsigned long, void *);
  
#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
static int proc_read_directpath(char *, char **, off_t, int, int *, void *);
static int proc_write_directpath(struct file *, const char *, unsigned long, void *);
#endif

static INLINE int print_session(char *, int, struct session_list_item *);
static INLINE int print_bridging_session(char *, int, struct bridging_session_list_item *);
static INLINE int print_netif(char *, int, struct netif_info *);
#if defined(ENABLE_SESSION_DEBUG_FLAGS) && ENABLE_SESSION_DEBUG_FLAGS
  static INLINE int print_debug_flags(char *, const char *, uint32_t);
#endif
static INLINE int stricmp(const char *, const char *);
static INLINE int strincmp(const char *, const char *, int);



/*
 * ####################################
 *           Global Variable
 * ####################################
 */

#if defined(ENABLE_DBG_PROC) && ENABLE_DBG_PROC
  int g_dbg_enable = 0;
#endif

#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) 
int g_directpath_enable = 0;
#endif

static struct file_operations g_ppa_dev_ops = {
    owner:      THIS_MODULE,
    ioctl:      ppa_dev_ioctl,
    open:       ppa_dev_open,
    release:    ppa_dev_release
};

static int g_init = 0;
#ifdef CONFIG_IFX_PPA_A4
  static int g_fwcode = FWCODE_ROUTING_BRIDGING_ACC_A4;
#else
  static int g_fwcode = FWCODE_ROUTING_BRIDGING_ACC_D4;
#endif

static PPA_IFNAME br0_ifname[PPA_IF_NAME_SIZE] = "br0";
static PPA_IFNAME eth0_ifname[PPA_IF_NAME_SIZE] = "eth0";
static PPA_IFNAME eth1_ifname[PPA_IF_NAME_SIZE] = "eth1";
#ifdef CONFIG_IFX_PPA_D4
  static PPA_IFNAME ppp0_ifname[PPA_IF_NAME_SIZE] = "ppp0";
  static PPA_IFNAME ath0_ifname[PPA_IF_NAME_SIZE] = "ath0";
  static PPA_IFNAME eth1_vlan5_ifname[PPA_IF_NAME_SIZE] = "eth1.5";
#endif
#ifdef CONFIG_IFX_PPA_A4
  static PPA_IFNAME ppp0_ifname[PPA_IF_NAME_SIZE] = "ppp0";
  static PPA_IFNAME nas0_ifname[PPA_IF_NAME_SIZE] = "nas0";
  static PPA_IFNAME nas1_ifname[PPA_IF_NAME_SIZE] = "nas1";

//805101:<IFTW-fchang>.added
static PPA_IFNAME nas2_ifname[PPA_IF_NAME_SIZE] = "nas2";
static PPA_IFNAME nas3_ifname[PPA_IF_NAME_SIZE] = "nas3";
static PPA_IFNAME nas4_ifname[PPA_IF_NAME_SIZE] = "nas4";
static PPA_IFNAME nas5_ifname[PPA_IF_NAME_SIZE] = "nas5";
static PPA_IFNAME nas6_ifname[PPA_IF_NAME_SIZE] = "nas6";
static PPA_IFNAME nas7_ifname[PPA_IF_NAME_SIZE] = "nas7";

  static PPA_IFNAME eth0_vlan5_ifname[PPA_IF_NAME_SIZE] = "eth0.5";
  static PPA_IFNAME eth0_vlan25_ifname[PPA_IF_NAME_SIZE] = "eth0.25";
#endif
#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
static  PPA_IFNAME directpath_ifname[PPA_IF_NAME_SIZE] = "usb0"; /* usb0 */
#endif

static PPA_IFNAME *phys_ifnames[8] = {
    eth0_ifname,
#ifdef CONFIG_IFX_PPA_D4
    eth1_ifname,
#else
    NULL,
#endif
    NULL,
#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
    NULL,
#else
    NULL,
#endif
    NULL,
#ifdef CONFIG_IFX_PPA_D4
    ath0_ifname,
#else
    NULL,
#endif
    NULL,
    NULL,
};

int ppa_set_phys_ifname_for_ifid(uint32_t if_id, PPA_IFNAME *ifname)
{
	char *new_ifname=NULL;

	if (if_id >= sizeof(phys_ifnames)/sizeof(phys_ifnames[0])) {
		return -1;
	}

	if (ifname) {
		/* Set phys_ifnames slot for if_id index */
		new_ifname=ppa_malloc(PPA_IF_NAME_SIZE+1);
		if (!new_ifname) {
			return -1;
		}
		strncpy(new_ifname, ifname, PPA_IF_NAME_SIZE);
		new_ifname[PPA_IF_NAME_SIZE-1]='\0';
		if (phys_ifnames[if_id]) {
			#if 0
			ppa_free(phys_ifnames[if_id]);
			#endif
		}
		phys_ifnames[if_id]=new_ifname;
		
	} else {
		/* Unset phys_ifname slot for if_id index */
		if (phys_ifnames[if_id]) {
			ppa_free(phys_ifnames[if_id]);
		}
		phys_ifnames[if_id]=0;

	}
	return 0;
}

#ifdef CONFIG_IFX_PPA_A4
  static PPA_IFINFO g_lanif_info[] = {
    {br0_ifname,        0},
    {eth0_ifname,       0},
    {eth0_vlan25_ifname,0},
#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
  //  {NULL,       0},
#endif
  };
  static PPA_IFINFO g_wanif_info[] = {
    {eth0_ifname,       0},
    {eth0_vlan5_ifname, 0},
    {nas0_ifname,       0},
    {nas1_ifname,       0},
//805101:<IFTW-fchang>.added
    {nas2_ifname,       0},
    {nas3_ifname,       0},
    {nas4_ifname,       0},
    {nas5_ifname,       0},
    {nas6_ifname,       0},
    {nas7_ifname,       0},

    {ppp0_ifname,       0},

  };
#else
  static PPA_IFINFO g_lanif_info[] = {
    {br0_ifname,        0},
    {eth0_ifname,       0},
    {ath0_ifname,       0},
  };
  static PPA_IFINFO g_wanif_info[] = {
    {eth1_ifname,       0},
    {eth1_vlan5_ifname, 0},
    {ppp0_ifname,       0},
  };
#endif

static PPA_INIT_INFO g_ppa_init_info = {
    lan_rx_checks : {
        f_ip_verify         : 1,
        f_tcp_udp_verify    : 1,
        f_tcp_udp_err_drop  : 1,
        f_drop_on_no_hit    : 0,
        f_mc_drop_on_no_hit : 0
    },
    wan_rx_checks : {
        f_ip_verify         : 1,
        f_tcp_udp_verify    : 1,
        f_tcp_udp_err_drop  : 1,
        f_drop_on_no_hit    : 0,
       f_mc_drop_on_no_hit : 0
    },
    num_lanifs              : NUM_ENTITY(g_lanif_info),
    p_lanifs                : g_lanif_info,
    num_wanifs              : NUM_ENTITY(g_wanif_info),
    p_wanifs                : g_wanif_info,
#ifdef CONFIG_IFX_PPA_A4
#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
    max_lan_source_entries  : 44,   //  ver 0.55: 48, ver 0.60 44
    max_wan_source_entries  : 44,   //  ver 0.55: 47, ver 0.60 44
  #else
    max_lan_source_entries  : 48,   //  ver 0.55: 48, ver 0.60 44
    max_wan_source_entries  : 47,   //  ver 0.55: 47, ver 0.60 44
  #endif
    max_mc_entries          : 32,
    max_bridging_entries    : 128,
#else
    max_lan_source_entries  : 128,
    max_wan_source_entries  : 128,
    max_mc_entries          : 32,
    max_bridging_entries    : 256,
#endif
    add_requires_min_hits   : 1
};

static struct netif_info *g_netif = NULL;
static PPA_LOCK g_netif_lock;

static PPA_LOCK g_session_list_lock;
static struct session_list_item *g_session_list_hash_table[SESSION_LIST_HASH_TABLE_SIZE] = {0};
static uint32_t g_session_list_length = 0;

#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
  static PPA_LOCK g_mc_session_list_lock;
  static struct session_list_item *g_mc_session_list_hash_table[SESSION_LIST_MC_HASH_TABLE_SIZE] = {0};
//  static uint32_t g_mc_session_list_length = 0;
#else
  static PPA_LOCK g_mc_group_list_lock;
  static struct mc_group_list_item *g_mc_group_list_hash_table[SESSION_LIST_MC_HASH_TABLE_SIZE] = {0};
  static uint32_t g_mc_group_list_length = 0;
#endif

static PPA_MEM_CACHE *g_session_item_cache = NULL;
#if defined(USE_NEW_MC_API) && USE_NEW_MC_API
  static PPA_MEM_CACHE *g_mc_group_item_cache = NULL;
#endif

static PPA_LOCK g_bridging_session_list_lock;
static struct bridging_session_list_item *g_bridging_session_list_hash_table[BRIDGING_SESSION_LIST_HASH_TABLE_SIZE] = {0};

static PPA_MEM_CACHE *g_bridging_session_item_cache = NULL;

static PPA_TIMER g_hit_stat_timer;
static uint32_t g_hit_polling_time = DEFAULT_HIT_POLLING_TIME;

static PPA_TIMER g_bridging_hit_stat_timer;
static uint32_t g_bridging_hit_polling_time = DEFAULT_BRIDGING_HIT_POLLING_TIME;

/*
 *  PROC
 */
static int g_ppa_proc_dir_flag = 0;
static struct proc_dir_entry *g_ppa_proc_dir = NULL;
static struct proc_dir_entry *g_ppa_api_proc_dir = NULL;



/*
 * ####################################
 *           Extern Variable
 * ####################################
 */

#ifdef CONFIG_IFX_PPA_D4
  extern int (*indirect_xmit[8])(uint32_t, PPA_BUF *);
  extern int (*device_push[8])(uint32_t, PPA_BUF *);
  extern uint32_t device_push_dev[8];
#endif

#if 0
#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
/*
 * Called by PPE low-level driver to send pkt via Directpath OR Indirect fastpath thru' PPA module.
 * PPA does the de-multiplexing between Directpath OR Indirectpath fastpath. tx_if_mask provides info 
 * of list of outgoing interfaces (for ex: a multicast pkt can be destined to multiple interfaces.
 */
int32_t ifx_ppa_rx_pkt (PPA_BUF *ppa_buf, int32_t len, int32_t rx_if_id, uint32_t tx_if_mask, void *ppe_header)
{	
    // check if outgoing intf is a directpath interface
    // if (ifx_ppa_get_netif_for_ifid (tx_if_mask) != NULL)
    {
        dbg ("PPE --> PPA\n");
        ifx_ppa_directpath_rx (ppa_buf, len, rx_if_id, tx_if_mask, ppe_header);
    }
    return 0;
}


/* 
 * Transmit a packet to the PPE from CPU bound interface. tx_dma_qid is reserved.
 */

int32_t ppa_send_pkt_to_accel_hw (PPA_BUF *ppa_buf, uint32_t rx_if_id, uint32_t tx_dma_qid, int32_t len)
{
    int32_t ret = 0;

    dbg ("PPA --> PPE\n");

    ret = ppe_send_pkt_to_ppe_upperhalf (ppa_buf, rx_if_id, tx_dma_qid, len);
    return ret;
}
#endif // defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
#endif

/*
 * ####################################
 *            Local Function
 * ####################################
 */

/*
 *  file operation functions
 */

static int ppa_dev_open(struct inode *inode, struct file *file)
{
    MOD_INC_USE_COUNT;
    return 0;
}

static int ppa_dev_release(struct inode *inode, struct file *file)
{
    MOD_DEC_USE_COUNT;
    return 0;
}

static int ppa_dev_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    int ret = -1;
    uint32_t flags = 0;

    if ( _IOC_TYPE(cmd) != PPA_IOC_MAGIC || _IOC_NR(cmd) >= PPA_IOC_MAXNR )
    {
        printk("_IOC_TYPE(%08X - %d) != PPA_IOC_MAGIC(%d)\n", cmd, _IOC_TYPE(cmd), PPA_IOC_MAGIC);
        return -ENOTTY;
    }

    if ( ((_IOC_DIR(cmd) & _IOC_READ) && !access_ok(VERIFY_WRITE, arg, _IOC_SIZE(cmd)))
        || ((_IOC_DIR(cmd) & _IOC_WRITE) && !access_ok(VERIFY_READ, arg, _IOC_SIZE(cmd))) )
    {
        printk("access check: (%08X && %d) || (%08X && %d)\n", (_IOC_DIR(cmd) & _IOC_READ), (int)!access_ok(VERIFY_WRITE, arg, _IOC_SIZE(cmd)),
                                                               (_IOC_DIR(cmd) & _IOC_WRITE), (int)!access_ok(VERIFY_READ, arg, _IOC_SIZE(cmd)));
        return -EFAULT;
    }

    switch (cmd)
    {
        /* Add flags field in PPA_INIT_INFO structure */
    case PPA_CMD_INIT:
        {
            PPA_CMD_INIT_INFO cmd_info;
            PPA_IFINFO ifinfo[NUM_ENTITY(cmd_info.ifinfo)];
            int i;

            copy_from_user(&cmd_info, (void *)arg, sizeof(cmd_info));
            for ( i = 0; i < NUM_ENTITY(ifinfo); i++ )
            {
                ifinfo[i].ifname   = cmd_info.ifinfo[i].ifname;
                ifinfo[i].if_flags = cmd_info.ifinfo[i].if_flags;
            }
            cmd_info.init_info.p_lanifs = ifinfo + (int32_t)cmd_info.init_info.p_lanifs;
            cmd_info.init_info.p_wanifs = ifinfo + (int32_t)cmd_info.init_info.p_wanifs;

            ret = ifx_ppa_init(&cmd_info.init_info, cmd_info.flags);
        }
        break;

    case PPA_CMD_EXIT:
        {
            ifx_ppa_exit();
            ret = 0;
        }
        break;

    case PPA_CMD_ENABLE:
        {
            PPA_CMD_ENABLE_INFO cmd_info;

            copy_from_user(&cmd_info, (void *)arg, sizeof(cmd_info));

            ret = ifx_ppa_enable(cmd_info.lan_rx_ppa_enable, cmd_info.wan_rx_ppa_enable, cmd_info.flags);
        }
        break;

    case PPA_CMD_GET_STATUS:
        {
            PPA_CMD_GET_STATUS_INFO cmd_info;

            ret = ifx_ppa_get_status(&cmd_info.lan_rx_ppa_enable, &cmd_info.wan_rx_ppa_enable);

            if ( ret == 0 )
                copy_to_user((void *)arg, &cmd_info, sizeof(cmd_info));
        }
        break;

    case PPA_CMD_SET_IF_MAC:
        {
            PPA_CMD_IF_MAC_INFO cmd_info;

            copy_from_user(&cmd_info, (void *)arg, sizeof(cmd_info));

            ret = ifx_ppa_set_if_mac_address(cmd_info.ifname, cmd_info.mac);
        }
        break;

    case PPA_CMD_GET_IF_MAC:
        {
            PPA_CMD_IF_MAC_INFO cmd_info;

            copy_from_user(&cmd_info, (void *)arg, sizeof(cmd_info));

            ret = ifx_ppa_get_if_mac_address(cmd_info.ifname, cmd_info.mac);

            if ( ret == 0 )
                copy_to_user((void *)arg, &cmd_info, sizeof(cmd_info));
        }
        break;

    case PPA_CMD_SET_BR_VLAN:
        {
            PPA_CMD_BR_IF_VLAN_CFG br_if_vlan_cmd_info;

            copy_from_user(&br_if_vlan_cmd_info, (PPA_CMD_BR_IF_VLAN_CFG *)arg, sizeof(PPA_CMD_BR_IF_VLAN_CFG));
            dbg("PPA_CMD_SET_BR_VLAN: %s 0x%08X 0x%08X\n", br_if_vlan_cmd_info.if_name, (u32)&br_if_vlan_cmd_info.vlan_tag_ctrl, (u32)&br_if_vlan_cmd_info.vlan_cfg);
            ret = ifx_ppa_set_bridge_if_vlan_config(ppa_get_netif(br_if_vlan_cmd_info.if_name), &br_if_vlan_cmd_info.vlan_tag_ctrl, &br_if_vlan_cmd_info.vlan_cfg, 0);
        }
        break;

    case PPA_CMD_ADD_BR_ENTRY:
        {
            PPA_IOCTL_VLAN_MAP_ENTRY map_entry;
            copy_from_user(&map_entry, (PPA_IOCTL_VLAN_MAP_ENTRY *)arg, sizeof(PPA_IOCTL_VLAN_MAP_ENTRY));
            ret = add_vlan_map(map_entry.ig_criteria_type, map_entry.ig_criteria, map_entry.new_vci, map_entry.vlan_port_map);
        }
        break;

    case PPA_CMD_DEL_BR_ENTRY:
        {
            PPA_IOCTL_VLAN_MAP_ENTRY map_entry;
            copy_from_user(&map_entry, (PPA_IOCTL_VLAN_MAP_ENTRY *)arg, sizeof(PPA_IOCTL_VLAN_MAP_ENTRY));
            del_vlan_map(map_entry.ig_criteria_type, map_entry.ig_criteria);
            ret = 0;
        }
        break;

    case PPA_CMD_ERASE_ALL_BR_ENTRY:
        {
            del_all_vlan_map();
            ret = 0;
        }
        break;

    case PPA_CMD_SET_BR_VLAN_PRIO:
        {
#ifdef CONFIG_IFX_PPA_D4
            PPA_IOCTL_BR_VLAN_PRIO br_vlan_prio;
            copy_from_user(&br_vlan_prio, (PPA_IOCTL_BR_VLAN_PRIO *)arg, sizeof(PPA_IOCTL_BR_VLAN_PRIO));
            setup_vlan_queue_map(br_vlan_prio.vlan_p, br_vlan_prio.queue);
            ret = 0;
#endif
        }
    case PPA_CMD_ENABLE_DIRECTPATH:
	{
#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
	    ifx_ppa_directpath_init(flags);
#endif
	}	
        break;

    case PPA_CMD_DISABLE_DIRECTPATH:
	{
#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
	    ifx_ppa_directpath_exit(flags);
#endif
	}	
        break;

    default: /* error */
        break;
    }

    return ret;
}

/*
 *  support functions
 */

static INLINE struct netif_info * alloc_netif_info(void)
{
    struct netif_info *obj;

    obj = (struct netif_info *)ppa_malloc(sizeof(*obj));
    if ( obj )
    {
        ppa_memset(obj, 0, sizeof(*obj));
        obj->mac_entry = ~0;
    }
    return obj;
}

static INLINE void free_netif_info(struct netif_info *obj)
{
    ppa_free(obj);
}

static INLINE void lock_netif_info_list(void)
{
    ppa_lock_get(&g_netif_lock);
}

static INLINE void unlock_netif_info_list(void)
{
    ppa_lock_release(&g_netif_lock);
}

static INLINE void free_netif_info_list(void)
{
    struct netif_info *obj;

    lock_netif_info_list();

    while ( g_netif )
    {
        obj = g_netif;
        g_netif = g_netif->next;

        free_netif_info(obj);
    }

    unlock_netif_info_list();
}

static INLINE void add_netif_info(struct netif_info *obj)
{
    lock_netif_info_list();

    obj->next = g_netif;
    g_netif = obj;

    unlock_netif_info_list();
}

static INLINE int32_t netif_info_is_added(PPA_IFNAME ifname[PPA_IF_NAME_SIZE], struct netif_info **pp_info)
{
    int32_t ret = 0;
    struct netif_info *p;

    lock_netif_info_list();

    for ( p = g_netif; p; p = p->next )
        if ( strcmp(p->name, ifname) == 0 )
        {
            ret = 1;
            if ( pp_info )
                *pp_info = p;
            break;
        }

    unlock_netif_info_list();

    return ret;
}

static INLINE int32_t ppa_d4_init(PPA_INIT_INFO *p_info, uint32_t flags)
{
    uint32_t max_rt_entry_num, max_mc_entry_num;
    uint32_t max_brg_entry_num;
    uint32_t eth0_iftype, eth1_iftype;
    uint8_t mac[PPA_ETH_ALEN];
    uint8_t broadcast_mac[PPA_ETH_ALEN] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
//  uint8_t pauseframe_mac[PPA_ETH_ALEN] = {0x01, 0x80, 0xC2, 0x00, 0x00, 0x01};
    struct netif_info *p_netif_info;
    PPA_NETIF *netif;
    uint32_t i;

    if ( p_info )
        ppa_memcpy(&g_ppa_init_info, p_info, sizeof(g_ppa_init_info));

    get_max_route_entries(&max_rt_entry_num, &max_mc_entry_num);

    get_max_bridging_entries(&max_brg_entry_num);

    if ( g_ppa_init_info.max_lan_source_entries + g_ppa_init_info.max_wan_source_entries > max_rt_entry_num )
        goto MAX_SOURCE_ENTRIES_ERROR;

    if ( g_ppa_init_info.max_mc_entries > max_mc_entry_num )
        goto MAX_MC_ENTRIES_ERROR;

    if ( g_ppa_init_info.max_bridging_entries > max_brg_entry_num )
        goto MAX_BRG_ENTRIES_ERROR;

    //  disable accelation mode by default
    set_acc_mode(1, IFX_PPA_ACC_MODE_NONE);
    set_acc_mode(0, IFX_PPA_ACC_MODE_NONE);

    set_route_cfg(1,
                  g_ppa_init_info.max_lan_source_entries,
                  0,
                  g_ppa_init_info.lan_rx_checks.f_ip_verify,
                  g_ppa_init_info.lan_rx_checks.f_tcp_udp_verify,
                  g_ppa_init_info.lan_rx_checks.f_tcp_udp_err_drop,
                  g_ppa_init_info.lan_rx_checks.f_drop_on_no_hit,
                  0,
                  IFX_PPA_SET_ROUTE_CFG_ENTRY_NUM | IFX_PPA_SET_ROUTE_CFG_IP_VERIFY | IFX_PPA_SET_ROUTE_CFG_TCPUDP_VERIFY | IFX_PPA_SET_ROUTE_CFG_DROP_ON_NOT_HIT);

    set_route_cfg(0,
                  g_ppa_init_info.max_wan_source_entries,
                  g_ppa_init_info.max_mc_entries,
                  g_ppa_init_info.wan_rx_checks.f_ip_verify,
                  g_ppa_init_info.wan_rx_checks.f_tcp_udp_verify,
                  g_ppa_init_info.wan_rx_checks.f_tcp_udp_err_drop,
                  g_ppa_init_info.wan_rx_checks.f_drop_on_no_hit,
                  g_ppa_init_info.wan_rx_checks.f_mc_drop_on_no_hit,
                  IFX_PPA_SET_ROUTE_CFG_ENTRY_NUM | IFX_PPA_SET_ROUTE_CFG_MC_ENTRY_NUM | IFX_PPA_SET_ROUTE_CFG_IP_VERIFY | IFX_PPA_SET_ROUTE_CFG_TCPUDP_VERIFY | IFX_PPA_SET_ROUTE_CFG_DROP_ON_NOT_HIT | IFX_PPA_SET_ROUTE_CFG_MC_DROP_ON_NOT_HIT);

    set_bridging_cfg(g_ppa_init_info.max_bridging_entries,
                     0xFF,              //  br_to_src_port_mask
                     0x00,              //  br_to_src_port_en
                     0,                 //  f_dest_vlan_en
                     0,                 //  f_src_vlan_en
                     0,                 //  f_mac_change_drop
                     IFX_PPA_SET_BRIDGING_CFG_ENTRY_NUM | IFX_PPA_SET_BRIDGING_CFG_BR_TO_SRC_PORT_EN | IFX_PPA_SET_BRIDGING_CFG_DEST_VLAN_EN | IFX_PPA_SET_BRIDGING_CFG_SRC_VLAN_EN | IFX_PPA_SET_BRIDGING_CFG_MAC_CHANGE_DROP);

    eth0_iftype = eth1_iftype = 0;
    free_netif_info_list();
    for ( i = 0; i < g_ppa_init_info.num_lanifs; i++ )
    {
        if ( !netif_info_is_added(g_ppa_init_info.p_lanifs[i].ifname, &p_netif_info) )
        {
            p_netif_info = alloc_netif_info();
            if ( !p_netif_info )
                goto G_LANIF_IFINDEX_KMALLOC_FAIL;

            strcpy(p_netif_info->name, g_ppa_init_info.p_lanifs[i].ifname);

            if ( strcmp(g_ppa_init_info.p_lanifs[i].ifname, eth0_ifname) == 0 )
                eth0_iftype |= IFX_PPA_IF_TYPE_LAN;
            else if ( strcmp(g_ppa_init_info.p_lanifs[i].ifname, eth1_ifname) == 0 )
                eth1_iftype |= IFX_PPA_IF_TYPE_LAN;

            netif = ppa_get_netif(g_ppa_init_info.p_lanifs[i].ifname);
            if ( netif )
            {
                if (  ppa_get_physical_if(netif, NULL, p_netif_info->phys_netif_name) >= 0 )
                    p_netif_info->flags |= NETIF_PHY_IF_GOT;

                if ( ppa_check_is_ppp_netif(netif) && ppa_check_is_pppoe_netif(netif) )
                    p_netif_info->flags |= NETIF_PPPOE;
                else if ( ppa_if_is_vlan_if(netif, NULL) )
                    p_netif_info->flags |= NETIF_VLAN;

#ifdef CONFIG_IFX_PPA_A4
                if ( (p_netif_info->flags & NETIF_PHY_IF_GOT) && (netif = ppa_get_netif(p_netif_info->phys_netif_name)) )
                {
                    PPA_VCC *vcc;
                    uint32_t dslwan_qid;

                    if ( ppa_br2684_get_vcc(netif, &vcc) >= 0 && (dslwan_qid = get_dslwan_qid_with_vcc(vcc)) >= 0 )
                    {
                        p_netif_info->flags |= NETIF_BR2684;
                        p_netif_info->dslwan_qid =dslwan_qid;
                    }
                }
#endif
            }

            add_netif_info(p_netif_info);
        }

        p_netif_info->flags |= NETIF_LAN_IF;
    }
    for ( i = 0; i < g_ppa_init_info.num_wanifs; i++ )
    {
        if ( !netif_info_is_added(g_ppa_init_info.p_wanifs[i].ifname, &p_netif_info) )
        {
            p_netif_info = alloc_netif_info();
            if ( !p_netif_info )
                goto G_WANIF_IFINDEX_KMALLOC_FAIL;

            strcpy(p_netif_info->name, g_ppa_init_info.p_wanifs[i].ifname);

            if ( strcmp(g_ppa_init_info.p_wanifs[i].ifname, eth0_ifname) == 0 )
                eth0_iftype |= IFX_PPA_IF_TYPE_WAN;
            else if ( strcmp(g_ppa_init_info.p_wanifs[i].ifname, eth1_ifname) == 0 )
                eth1_iftype |= IFX_PPA_IF_TYPE_WAN;

            netif = ppa_get_netif(g_ppa_init_info.p_wanifs[i].ifname);
            if ( netif )
            {
                if (  ppa_get_physical_if(netif, NULL, p_netif_info->phys_netif_name) >= 0 )
                    p_netif_info->flags |= NETIF_PHY_IF_GOT;

                if ( ppa_check_is_ppp_netif(netif) && ppa_check_is_pppoe_netif(netif) )
                    p_netif_info->flags |= NETIF_PPPOE;
                else if ( ppa_if_is_vlan_if(netif, NULL) )
                    p_netif_info->flags |= NETIF_VLAN;

#ifdef CONFIG_IFX_PPA_A4
                if ( (p_netif_info->flags & NETIF_PHY_IF_GOT) && (netif = ppa_get_netif(p_netif_info->phys_netif_name)) )
                {
                    PPA_VCC *vcc;
                    uint32_t dslwan_qid;

                    if ( ppa_br2684_get_vcc(netif, &vcc) >= 0 && (dslwan_qid = get_dslwan_qid_with_vcc(vcc)) >= 0 )
                    {
                        p_netif_info->flags |= NETIF_BR2684;
                        p_netif_info->dslwan_qid =dslwan_qid;
                    }
                }
#endif
            }

            add_netif_info(p_netif_info);
        }

        p_netif_info->flags |= NETIF_WAN_IF;
    }

#ifdef CONFIG_IFX_PPA_A4
    set_if_type(eth0_iftype, 0);
    set_wan_vlan_id((0x010 << 16) | 0x000);         //  high 16 bits is upper of WAN VLAN ID, low 16 bits is min WAN VLAN ID
#else
    set_if_type(eth0_iftype, 0);
    set_if_type(eth1_iftype, 1);
#endif

//  *** WFQ is hidden and not to be changed after firmware running. ***
//  set_if_wfq(ETX_WFQ_D4, 1);
//  set_fastpath_wfq(FASTPATH_WFQ_D4);
//  set_dplus_wfq(DPLUS_WFQ_D4);

    for ( i = 0; i < 8; i++ )
        set_default_dest_list(IFX_PPA_DEST_LIST_CPU0, IFX_PPA_DEST_LIST_CPU0, i);

    //set_fast_mode(IFX_PPA_SET_FAST_MODE_CPU1_DIRECT | IFX_PPA_SET_FAST_MODE_ETH1_INDIRECT, IFX_PPA_SET_FAST_MODE_CPU1 | IFX_PPA_SET_FAST_MODE_ETH1);
    set_fast_mode(IFX_PPA_SET_FAST_MODE_CPU1_DIRECT | IFX_PPA_SET_FAST_MODE_ETH1_DIRECT, IFX_PPA_SET_FAST_MODE_CPU1 | IFX_PPA_SET_FAST_MODE_ETH1);

    lock_netif_info_list();
    for ( p_netif_info = g_netif; p_netif_info; p_netif_info = p_netif_info->next )
    {
        netif = ppa_get_netif(p_netif_info->name);
        if ( netif )
        {
            ppa_get_netif_hwaddr(netif, mac);
            ASSERT(*(uint32_t *)mac != 0 || *((uint16_t *)mac + 2) != 0, "MAC address is all zero - %s", p_netif_info->name);

            ppa_memcpy(p_netif_info->mac, mac, PPA_ETH_ALEN);

            if ( add_mac_entry(mac, &p_netif_info->mac_entry) == 0 )
                p_netif_info->flags |= NETIF_MAC_ENTRY_CREATED;
            else
                err("add_mac_entry failed");
        }
    }
    unlock_netif_info_list();

    //  TODO
    //set_rt_br_id_tbl

#ifdef CONFIG_IFX_PPA_A4
    add_bridging_entry(IFX_PPA_PORT_ANY,
                       broadcast_mac,
                       0,                       //  f_src_mac_drop
                       0,                       //  dslwan_qid
                       IFX_PPA_DEST_LIST_CPU0,
                       NULL);
#else
    add_bridging_entry(IFX_PPA_PORT_ANY,
                       broadcast_mac,
                       0,                       //  vci
                       0,                       //  f_src_mac_drop
                       0,                       //  f_vlan_ins_enable
                       0,                       //  new_vci
                       0,                       //  f_vlan_rm_enable
  #if 1
                       IFX_PPA_DEST_LIST_CPU0,
  #else
                       IFX_PPA_DEST_LIST_CPU0 | IFX_PPA_DEST_LIST_ETH0 | IFX_PPA_DEST_LIST_ETH1,
  #endif
                       NULL);

  #if 0
    add_bridging_entry(IFX_PPA_PORT_ANY,
                       pauseframe_mac,
                       0,                       //  vci
                       0,                       //  f_src_mac_drop
                       0,                       //  f_vlan_ins_enable
                       0,                       //  new_vci
                       0,                       //  f_vlan_rm_enable
                       IFX_PPA_DEST_LIST_ETH0,
                       NULL);
  #endif
#endif

    //  start timer
    ppa_timer_init(&g_hit_stat_timer, ppa_check_hit_stat);
    ppa_timer_add(&g_hit_stat_timer, g_hit_polling_time);
    ppa_timer_init(&g_bridging_hit_stat_timer, ppa_bridging_check_hit_stat);
    ppa_timer_add(&g_bridging_hit_stat_timer, g_bridging_hit_polling_time);

    ppa_hook_enable_fn              = ifx_ppa_enable;
    ppa_hook_get_status_fn          = ifx_ppa_get_status;

    ppa_hook_session_add_fn         = ifx_ppa_session_add;
    ppa_hook_session_del_fn         = ifx_ppa_session_delete;
    ppa_hook_session_modify_fn      = ifx_ppa_session_modify;
    ppa_hook_session_get_fn         = ifx_ppa_session_get;
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    ppa_hook_mc_session_add_fn      = ifx_ppa_mc_session_add;
    ppa_hook_mc_session_del_fn      = ifx_ppa_mc_session_delete;
#else
    ppa_hook_mc_group_update_fn     = ifx_ppa_mc_group_update;
    ppa_hook_mc_group_get_fn        = ifx_ppa_mc_group_get;
#endif
    ppa_hook_inactivity_status_fn   = ifx_ppa_inactivity_status;
    ppa_hook_set_inactivity_fn      = ifx_ppa_set_session_inactivity;
    ppa_hook_get_if_stats_fn        = ifx_ppa_get_if_stats;
    ppa_hook_get_accel_stats_fn     = ifx_ppa_get_accel_stats;
    ppa_hook_set_if_mac_address_fn  = ifx_ppa_set_if_mac_address;
    ppa_hook_get_if_mac_address_fn  = ifx_ppa_get_if_mac_address;

    ppa_hook_bridge_entry_add_fn               = ifx_ppa_bridge_entry_add;
    ppa_hook_bridge_entry_delete_fn            = ifx_ppa_bridge_entry_delete;
    ppa_hook_set_bridge_entry_timeout_fn       = ifx_ppa_set_bridge_entry_timeout;
    ppa_hook_bridge_entry_inactivity_status_fn = ifx_ppa_bridge_entry_inactivity_status;
    ppa_hook_bridge_entry_hit_time_fn          = ifx_ppa_bridge_entry_hit_time;

//804251:<IFTW-fchang>.added
#if defined(CONFIG_IFX_PPA_D4) || defined(CONFIG_DANUBE_PPA_D4) || defined(CONFIG_IFX_PPA_A4)

    ppa_hook_set_bridge_if_vlan_config  = ifx_ppa_set_bridge_if_vlan_config;
    ppa_hook_get_bridge_if_vlan_config  = ifx_ppa_get_bridge_if_vlan_config;
    ppa_hook_vlan_filter_del            = ifx_ppa_vlan_filter_del;
    ppa_hook_vlan_filter_add            = ifx_ppa_vlan_filter_add;

//804251:<IFTW-fchang>.added
#endif

    ppa_hook_add_itf                    = ifx_ppa_add_itf;

#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
    // ppa_hook_rx_pkt_fn 			= ifx_ppa_rx_pkt; 
#endif

    g_init = 1;
    
    //804282:<IFTW-fchang>.modified
    printk("ppa_a4 - init succeeded\n");
    
    return 0;

G_WANIF_IFINDEX_KMALLOC_FAIL:
G_LANIF_IFINDEX_KMALLOC_FAIL:
    err("failed in alloc memory for netif_info");
    free_netif_info_list();
MAX_BRG_ENTRIES_ERROR:
MAX_MC_ENTRIES_ERROR:
MAX_SOURCE_ENTRIES_ERROR:
    err("failed in PPA init");
    return -1;
}

static INLINE void ppa_d4_exit(void)
{
    struct session_list_item *p_item, *p_next;
#if defined(USE_NEW_MC_API) && USE_NEW_MC_API
    struct mc_group_list_item *p_mc_item, *p_mc_next;
#endif
    struct bridging_session_list_item *p_brg_item, *p_brg_next;
    int i;

    ppa_hook_enable_fn              = NULL;
    ppa_hook_get_status_fn          = NULL;

    ppa_hook_session_add_fn         = NULL;
    ppa_hook_session_del_fn         = NULL;
    ppa_hook_session_modify_fn      = NULL;
    ppa_hook_session_get_fn         = NULL;
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    ppa_hook_mc_session_add_fn      = NULL;
    ppa_hook_mc_session_del_fn      = NULL;
#else
    ppa_hook_mc_group_update_fn     = NULL;
    ppa_hook_mc_group_get_fn        = NULL;
#endif
    ppa_hook_inactivity_status_fn   = NULL;
    ppa_hook_set_inactivity_fn      = NULL;
    ppa_hook_get_if_stats_fn        = NULL;
    ppa_hook_get_accel_stats_fn     = NULL;
    ppa_hook_set_if_mac_address_fn  = NULL;
    ppa_hook_get_if_mac_address_fn  = NULL;

    ppa_hook_enable_fn                         = NULL;
    ppa_hook_get_status_fn                     = NULL;
    ppa_hook_bridge_entry_add_fn               = NULL;
    ppa_hook_bridge_entry_delete_fn            = NULL;
    ppa_hook_set_bridge_entry_timeout_fn       = NULL;
    ppa_hook_bridge_entry_inactivity_status_fn = NULL;
    ppa_hook_bridge_entry_hit_time_fn          = NULL;

//804251:<IFTW-fchang>.added
#if defined(CONFIG_IFX_PPA_D4) || defined(CONFIG_DANUBE_PPA_D4) ||  defined(CONFIG_IFX_PPA_A4)

    ppa_hook_set_bridge_if_vlan_config  = NULL;
    ppa_hook_get_bridge_if_vlan_config  = NULL;
    ppa_hook_vlan_filter_del            = NULL;
    ppa_hook_vlan_filter_add            = NULL;

//804251:<IFTW-fchang>.added
#endif

    ppa_hook_add_itf                    = NULL;

#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
    //ppa_hook_directpath_tx_complete_fn  = NULL;
#endif

    //  delete timer
    ppa_timer_del(&g_hit_stat_timer);
    ppa_timer_del(&g_bridging_hit_stat_timer);

    ppa_lock_get(&g_session_list_lock);
    for ( i = 0; i < NUM_ENTITY(g_session_list_hash_table); i++ )
    {
        for ( p_item = g_session_list_hash_table[i]; p_item; p_item = p_next )
        {
            p_next = p_item->next;

            ppa_hw_del_session(p_item);

            ppa_mem_cache_free(p_item, g_session_item_cache);
        }
        g_session_list_hash_table[i] = NULL;
    }
    ppa_lock_release(&g_session_list_lock);

#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    ppa_lock_get(&g_mc_session_list_lock);
    for ( i = 0; i < NUM_ENTITY(g_mc_session_list_hash_table); i++ )
    {
        for ( p_item = g_mc_session_list_hash_table[i]; p_item; p_item = p_next )
        {
            p_next = p_item->next;

            ppa_hw_del_mc_session(p_item);

            ppa_mem_cache_free(p_item, g_session_item_cache);
        }
        g_mc_session_list_hash_table[i] = NULL;
    }
    ppa_lock_release(&g_mc_session_list_lock);
#else
    ppa_lock_get(&g_mc_group_list_lock);
    for ( i = 0; i < NUM_ENTITY(g_mc_group_list_hash_table); i++ )
    {
        for ( p_mc_item = g_mc_group_list_hash_table[i]; p_mc_item; p_mc_item = p_mc_next )
        {
            p_mc_next = p_mc_item->next;

            ppa_hw_del_mc_group(p_mc_item);

            ppa_mem_cache_free(p_mc_item, g_mc_group_item_cache);
        }
        g_mc_group_list_hash_table[i] = NULL;
    }
    ppa_lock_release(&g_mc_group_list_lock);
#endif

    ppa_lock_get(&g_bridging_session_list_lock);
    for ( i = 0; i < NUM_ENTITY(g_bridging_session_list_hash_table); i++ )
    {
        for ( p_brg_item = g_bridging_session_list_hash_table[i]; p_brg_item; p_brg_item = p_brg_next )
        {
            p_brg_next = p_brg_item->next;

            ppa_bridging_hw_del_session(p_brg_item);

            ppa_mem_cache_free(p_brg_item, g_bridging_session_item_cache);
        }
        g_bridging_session_list_hash_table[i] = NULL;
    }
    ppa_lock_release(&g_bridging_session_list_lock);

    free_netif_info_list();

    kmem_cache_shrink(g_session_item_cache);
#if defined(USE_NEW_MC_API) && USE_NEW_MC_API
    kmem_cache_shrink(g_mc_group_item_cache);
#endif
    kmem_cache_shrink(g_bridging_session_item_cache);

    g_init = 0;
}

static INLINE void ppa_init_session_list_item(struct session_list_item *p_item)
{
    ppa_memset(p_item, 0, sizeof(*p_item));
    p_item->mtu = DEFAULT_MTU;
    p_item->routing_entry = ~0;
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    p_item->mc_routing_entry = ~0;
#endif
    p_item->pppoe_entry = ~0;
    p_item->mtu_entry = ~0;
    p_item->src_mac_entry = ~0;
#ifdef CONFIG_IFX_PPA_A4
    p_item->out_vlan_ix = ~0;
#endif
}

#if defined(USE_NEW_MC_API) && USE_NEW_MC_API
static INLINE void ppa_init_mc_group_list_item(struct mc_group_list_item *p_item)
{
    ppa_memset(p_item, 0, sizeof(*p_item));
    p_item->mc_entry = ~0;
    p_item->src_mac_entry = ~0;
  #ifdef CONFIG_IFX_PPA_A4
    p_item->out_vlan_ix = ~0;
  #endif
}
#endif

static INLINE void ppa_bridging_init_session_list_item(struct bridging_session_list_item *p_item)
{
    ppa_memset(p_item, 0, sizeof(*p_item));
    p_item->bridging_entry = ~0;
}

static INLINE int32_t ppa_get_iftype(PPA_NETIF *netif)
{
    int32_t type;
    struct netif_info *pinfo;

    type = IFX_PPA_IF_NOT_FOUND;

    lock_netif_info_list();
    for ( pinfo = g_netif; pinfo; pinfo = pinfo->next )
        if ( ppa_is_netif_name(netif, pinfo->name) )
        {
            if ( (pinfo->flags & NETIF_LAN_IF) )
                type |= IFX_PPA_IF_TYPE_LAN;
            if ( (pinfo->flags & NETIF_WAN_IF) )
                type |= IFX_PPA_IF_TYPE_WAN;
            break;
        }
    unlock_netif_info_list();

    return type;
}

//  return negative means error
static INLINE int32_t ppa_get_ifid(PPA_NETIF *netif)
{
    int ret;
    PPA_IFNAME phy_ifname[PPA_IF_NAME_SIZE];

    if ( ppa_get_physical_if(netif, NULL, phy_ifname) )
    {
        dbg("failed in getting phiscal interface");
        return -1;
    }

#ifdef CONFIG_IFX_PPA_A4
    if ( ppa_if_is_br2684(NULL, phy_ifname) || ppa_if_is_pppoa(NULL, phy_ifname) )
        return IFX_PPA_IF_ID_ATM;
#endif

    for ( ret = 0; ret < NUM_ENTITY(phys_ifnames); ret++ )
        if ( phys_ifnames[ret] && strcmp(phy_ifname, phys_ifnames[ret]) == 0 )
            return ret;

    return -1;
}

int32_t ppa_lookup_session(PPA_SESSION *p_session, uint32_t is_reply, struct session_list_item **pp_item)
{
    int32_t ret;
    struct session_list_item *p_prev, *p_cur;
    struct session_list_item **pp_header;

    ASSERT(pp_item != NULL, "pp_item == NULL");

    ret = IFX_PPA_SESSION_NOT_ADDED;

    pp_header = g_session_list_hash_table + SESSION_LIST_HASH_VALUE(p_session, is_reply);
    p_prev = NULL;
    ppa_lock_get(&g_session_list_lock);
    p_cur = *pp_header;
    while ( p_cur && !ppa_is_session_equal(p_session, p_cur->session) )
    {
        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( p_prev )
        {
            p_prev->next = p_cur->next;
            p_cur->next = *pp_header;
            *pp_header = p_cur;
        }

        p_cur->num_adds++;

        ret = IFX_PPA_SESSION_EXISTS;
    }
    ppa_lock_release(&g_session_list_lock);

    *pp_item = p_cur;

    return ret;
}

#if 0
static INLINE int32_t ppa_lookup_session_add(PPA_SESSION *p_session, struct session_list_item **pp_item, uint32_t src_ip)
{
    int32_t ret;
    struct session_list_item *p_prev, *p_cur;
    struct session_list_item **pp_header;

    ASSERT(pp_item != NULL, "pp_item == NULL");

    ret = IFX_PPA_SESSION_NOT_ADDED;

    pp_header = g_session_list_hash_table + SESSION_LIST_HASH_VALUE(p_session);
    p_prev = NULL;
    ppa_lock_get(&g_session_list_lock);
    p_cur = *pp_header;
    while ( p_cur && (!ppa_is_session_equal(p_session, p_cur->session) || p_cur->src_ip != src_ip) )
    {
//        printk("p_session = %08X, p_cur->session = %08X\n", (uint32_t)p_session, (uint32_t)p_cur->session);
        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( p_prev )
        {
            p_prev->next = p_cur->next;
            p_cur->next = *pp_header;
            *pp_header = p_cur;
        }

        p_cur->num_adds++;

        ret = IFX_PPA_SESSION_EXISTS;
    }
    ppa_lock_release(&g_session_list_lock);

    *pp_item = p_cur;

    return ret;
}
#endif

#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
int32_t ppa_lookup_mc_session(uint32_t ip_mc_group, struct session_list_item **pp_item)
{
    uint32_t ret;
    struct session_list_item *p_prev, *p_cur;
    struct session_list_item **pp_header;

    ASSERT(pp_item != NULL, "pp_item == NULL");

    ret = IFX_PPA_SESSION_NOT_ADDED;

    pp_header = g_mc_session_list_hash_table + SESSION_LIST_MC_HASH_VALUE(ip_mc_group);
    p_prev = NULL;
    ppa_lock_get(&g_mc_session_list_lock);
    p_cur = *pp_header;
    while ( p_cur )
    {
        if ( p_cur->dst_ip == ip_mc_group )
            break;

        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( p_prev )
        {
            p_prev->next = p_cur->next;
            p_cur->next = *pp_header;
            *pp_header = p_cur;
        }

        p_cur->num_adds++;

        ret = IFX_PPA_SESSION_EXISTS;
    }
    ppa_lock_release(&g_mc_session_list_lock);

    *pp_item = p_cur;

    return ret;
}
#else
static INLINE int32_t ppa_lookup_mc_group(IPADDR ip_mc_group, struct mc_group_list_item **pp_item)
{
    uint32_t ret;
    struct mc_group_list_item *p_prev, *p_cur;
    struct mc_group_list_item **pp_header;

    ASSERT(pp_item != NULL, "pp_item == NULL");

    ret = IFX_PPA_SESSION_NOT_ADDED;

    pp_header = g_mc_group_list_hash_table + SESSION_LIST_MC_HASH_VALUE(ip_mc_group);
    p_prev = NULL;
    ppa_lock_get(&g_mc_group_list_lock);
    p_cur = *pp_header;
    while ( p_cur )
    {
        if ( p_cur->ip_mc_group == ip_mc_group )
            break;

        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( p_prev )
        {
            p_prev->next = p_cur->next;
            p_cur->next = *pp_header;
            *pp_header = p_cur;
        }

        ret = IFX_PPA_SESSION_EXISTS;
    }
    ppa_lock_release(&g_mc_group_list_lock);

    *pp_item = p_cur;

    return ret;
}
#endif

static INLINE int32_t ppa_check_if_session_can_accel(PPA_BUF *ppa_buf, PPA_SESSION *p_session, struct session_list_item *p_item)
{
    uint32_t proto;

    if ( !p_session )
        p_session = ppa_get_session(ppa_buf);

    proto = ppa_get_pkt_ip_proto(ppa_buf);
    if ( proto != IPPROTO_UDP && proto != IPPROTO_TCP )
    {
        dbg("neither UDP nor TCP");
        return 0;
    }

    //  Check if session needs to be handled in MIPS for Conntrack handling
    if ( ppa_check_is_special_session(ppa_buf, p_session) )
        return 0;

    //  TODO: If Local session, i.e. Terminates or originates, skip

    return 1;
}

static INLINE int32_t ppa_add_session(PPA_BUF *ppa_buf, PPA_SESSION *p_session, struct session_list_item **pp_item, uint32_t flags)
{
    struct session_list_item *p_item;

    if ( !p_session )
    {
        p_session = ppa_get_session(ppa_buf);
        if ( !p_session )
            return -1;
    }

    p_item = ppa_mem_cache_alloc(g_session_item_cache);
    if ( !p_item )
    {
        err("failed in memory allocation");
        return -1;
    }

    ppa_init_session_list_item(p_item);

    dump_list_item(p_item, "ppa_add_session (after init)");

    dbg("ppa_get_session(ppa_buf) = %08X", (uint32_t)ppa_get_session(ppa_buf));

    p_item->session       = p_session;

    if ( (flags & PPA_F_SESSION_REPLY_DIR) )
        p_item->flags    |= SESSION_IS_REPLY;

    p_item->ip_proto      = ppa_get_pkt_ip_proto(ppa_buf);
    p_item->src_ip        = ppa_get_pkt_src_ip(ppa_buf);
    p_item->src_port      = ppa_get_pkt_src_port(ppa_buf);
    p_item->dst_ip        = ppa_get_pkt_dst_ip(ppa_buf);
    p_item->dst_port      = ppa_get_pkt_dst_port(ppa_buf);
    p_item->rx_if         = ppa_get_pkt_src_if(ppa_buf);
    p_item->timeout       = ppa_get_default_session_timeout();
    p_item->last_hit_time = ppa_get_time_in_sec();

#if 0
    if ( ppa_check_is_ppp_netif(p_item->rx_if ) && ppa_check_is_pppoe_netif(p_item->rx_if) )
    {
        p_item->pppoe_session_id = ppa_get_pkt_pppoe_session_id(ppa_buf);
        p_item->flags |= SESSION_VALID_PPPOE;
        dbg("pppoe src case session_id: %d", p_item->pppoe_session_id);
    }
#endif

    ppa_get_pkt_rx_src_mac_addr(ppa_buf, p_item->src_mac);

    ppa_insert_session_item(p_item);

    dump_list_item(p_item, "ppa_add_session (after setting)");

    *pp_item = p_item;

    return 0;
}

static INLINE int32_t ppa_update_session(PPA_BUF *ppa_buf, struct session_list_item *p_item, uint32_t flags)
{
    uint32_t ip, port;
    PPA_NETIF *netif;
#ifdef CONFIG_IFX_PPA_A4
    PPA_IFNAME phys_netif_name[PPA_IF_NAME_SIZE];
    //  added by xuliang: start
    int f_is_ipoa = 0;
    int f_is_pppoa = 0;
    //  added by xuliang: end
#endif

    //  XXX: only port change with same IP not supported here, not really useful
    ip = ppa_get_pkt_src_ip(ppa_buf);
    if ( ip != p_item->src_ip )
    {
        p_item->nat_ip = ip;
        p_item->flags |= SESSION_VALID_NAT_IP;

        port = ppa_get_pkt_src_port(ppa_buf);
        if ( port != p_item->src_port )
        {
            p_item->nat_port = port;
            p_item->flags |= SESSION_VALID_NAT_PORT | SESSION_VALID_NAT_SNAT;
        }
    }
    else
    {
        ip = ppa_get_pkt_dst_ip(ppa_buf);
        if ( ip != p_item->dst_ip )
        {
            p_item->nat_ip = ip;
            p_item->flags |= SESSION_VALID_NAT_IP;

            port = ppa_get_pkt_dst_port(ppa_buf);
            if ( port != p_item->dst_port )
            {
                p_item->nat_port = port;
                p_item->flags |= SESSION_VALID_NAT_PORT;
            }
       }
    }

    p_item->tx_if = ppa_get_pkt_dst_if(ppa_buf);

#if 1
    netif = p_item->rx_if;
    if ( (ppa_get_iftype(netif) &  IFX_PPA_IF_TYPE_WAN) && ppa_check_is_ppp_netif(netif) && ppa_check_is_pppoe_netif(netif) )
    {
        //  src interface is WAN and PPPoE
        p_item->pppoe_session_id = ppa_get_pkt_pppoe_session_id(ppa_buf);
        p_item->flags |= SESSION_VALID_PPPOE;
        SET_DBG_FLAG(p_item, SESSION_DBG_RX_PPPOE);
    }

    //  if destination interface is PPPoE, it covers the previous setting
    netif = p_item->tx_if;
    if ( (ppa_get_iftype(netif) &  IFX_PPA_IF_TYPE_WAN) && ppa_check_is_ppp_netif(netif) && ppa_check_is_pppoe_netif(netif) )
    {
        ASSERT(!(p_item->flags & SESSION_VALID_PPPOE), "both interfaces are WAN PPPoE interface, not possible");
        p_item->pppoe_session_id = ppa_get_pkt_pppoe_session_id(ppa_buf);
        p_item->flags |= SESSION_VALID_PPPOE;
        SET_DBG_FLAG(p_item, SESSION_DBG_TX_PPPOE);
    }
#else
    netif = p_item->tx_if;
    dbg("pppoe %s : %d", netif->name, ppa_check_is_ppp_netif(netif));
    if ( ppa_check_is_ppp_netif(netif) && ppa_check_is_pppoe_netif(netif) )
    {
        if ( (p_item->flags & SESSION_VALID_PPPOE) )
        {
            dbg("can not be as pppoe interface both LAN and WAN");
            return -1;
        }
        p_item->pppoe_session_id = ppa_get_pkt_pppoe_session_id(ppa_buf);
        p_item->flags |= SESSION_VALID_PPPOE;
        dbg("pppoe dst case session_id: %d", p_item->pppoe_session_id );
    }
#endif

#ifdef CONFIG_IFX_PPA_A4
    if ( ppa_get_physical_if(p_item->tx_if, NULL, phys_netif_name) >= 0 && (netif = ppa_get_netif(phys_netif_name)) )
    {
        PPA_VCC *vcc;
        uint32_t dslwan_qid;

        if ( ppa_br2684_get_vcc(netif, &vcc) >= 0 && (dslwan_qid = get_dslwan_qid_with_vcc(vcc)) >= 0 )
        //  following one line added by xuliang
        {
            SET_DBG_FLAG(p_item, SESSION_DBG_TX_BR2684_EOA);
            p_item->dslwan_qid = dslwan_qid;
        //  added by xuliang: start
            if ( ppa_if_is_ipoa(netif, NULL) )
            {
                SET_DBG_FLAG(p_item, SESSION_DBG_TX_BR2684_IPOA);
                f_is_ipoa = 1;
                p_item->flags |= SESSION_TX_ITF_IPOA;
            }
        }
        else if ( ppa_pppoa_get_vcc(netif, &vcc) >= 0 && (dslwan_qid = get_dslwan_qid_with_vcc(vcc)) >= 0 )
        {
            SET_DBG_FLAG(p_item, SESSION_DBG_TX_PPPOA);
            p_item->dslwan_qid = dslwan_qid;
            f_is_pppoa = 1;
            p_item->flags |= SESSION_TX_ITF_PPPOA;
        }
        //  added by xuliang: end
    }

    //  Testing Purpose - VLAN
    {
        u32 rx_vid = ~0, tx_vid = ~0;

        if ( ppa_if_is_vlan_if(p_item->rx_if, NULL) )
        {
            SET_DBG_FLAG(p_item, SESSION_DBG_RX_VLAN);
            rx_vid = ppa_get_vlan_id(p_item->rx_if);
        }
        if ( ppa_if_is_vlan_if(p_item->tx_if, NULL) )
        {
            SET_DBG_FLAG(p_item, SESSION_DBG_TX_VLAN);
            tx_vid = ppa_get_vlan_id(p_item->tx_if);
        }
        if ( rx_vid != ~0 && tx_vid != ~0 )
        {
            if ( rx_vid != tx_vid )
            {
                p_item->flags |= SESSION_VALID_VLAN_INS | SESSION_VALID_VLAN_RM;
                p_item->new_vci = tx_vid;   //  ignore prio and cfi
            }
        }
        else if ( rx_vid != ~0 )
            p_item->flags |= SESSION_VALID_VLAN_RM;
        else if ( tx_vid != ~0 )
        {
            p_item->flags |= SESSION_VALID_VLAN_INS;
            p_item->new_vci = tx_vid;
        }
    }
  #if 0
    {
        u32 rx_vid = ~0, tx_vid = ~0;

        if ( ppa_if_is_vlan_if(p_item->rx_if, NULL) )
            rx_vid = ppa_get_vlan_id(p_item->rx_if);
        if ( ppa_if_is_vlan_if(p_item->tx_if, NULL) )
            tx_vid = ppa_get_vlan_id(p_item->tx_if);
        if ( rx_vid != ~0 && tx_vid != ~0 )
        {
            if ( rx_vid != tx_vid )
            {
                p_item->flags |= SESSION_VALID_OUT_VLAN_INS | SESSION_VALID_OUT_VLAN_RM;
                p_item->out_vlan_tag = 0x81000000 | tx_vid; //  ignore prio and cfi
            }
        }
        else if ( rx_vid != ~0 )
            p_item->flags |= SESSION_VALID_OUT_VLAN_RM;
        else if ( tx_vid != ~0 )
        {
            p_item->flags |= SESSION_VALID_OUT_VLAN_INS;
            p_item->out_vlan_tag = 0x81000000 | tx_vid;
        }
    }
  #endif

    //  added by xuliang: start
    if ( f_is_ipoa || f_is_pppoa )
        return 0;
    //  added by xuliang: end
#endif

    if ( ppa_get_dst_mac(ppa_buf, p_item->session, p_item->dst_mac) != RET_E_OK )
    {
        dbg("session:%x can not get dst mac!", (u32)ppa_get_session(ppa_buf));
        SET_DBG_FLAG(p_item, SESSION_DBG_GET_DST_MAC_FAIL);
        return -1;
    }

//    printk("dst_mac = %02x:%02x:%02x:%02x:%02x:%02x\n", (uint32_t)p_item->dst_mac[0], (uint32_t)p_item->dst_mac[1], (uint32_t)p_item->dst_mac[2], (uint32_t)p_item->dst_mac[3], (uint32_t)p_item->dst_mac[4], (uint32_t)p_item->dst_mac[5]);

//    if ( PPA_IF_NETIF_NAME_START(p_item->tx_if, "ppp", 3) )
//    {
//        //  TODO: call ppp
//    }

    return 0;
}

static INLINE int32_t ppa_update_session_extra(PPA_SESSION_EXTRA *p_extra, struct session_list_item *p_item, uint32_t flags)
{
    p_item->session_extra = *p_extra;

    if ( (flags & PPA_F_SESSION_NEW_DSCP) )
    {
        if ( p_extra->dscp_remark )
        {
            p_item->flags |= SESSION_VALID_NEW_DSCP;
            p_item->new_dscp = p_extra->new_dscp;
        }
        else
            p_item->flags &= ~SESSION_VALID_NEW_DSCP;
    }

    if ( (flags & PPA_F_SESSION_VLAN) )
    {
        if ( p_extra->vlan_insert )
        {
            p_item->flags |= SESSION_VALID_VLAN_INS;
            p_item->new_vci = ((uint32_t)p_extra->vlan_prio << 13) | ((uint32_t)p_extra->vlan_cfi << 12) | p_extra->vlan_id;
        }
        else
        {
            p_item->flags &= ~SESSION_VALID_VLAN_INS;
            p_item->new_vci = 0;
        }

        if ( p_extra->vlan_remove )
            p_item->flags |= SESSION_VALID_VLAN_RM;
        else
            p_item->flags &= ~SESSION_VALID_VLAN_RM;
    }

    if ( (flags & PPA_F_MTU) )
    {
        p_item->flags |= SESSION_VALID_MTU;
        p_item->mtu = p_extra->mtu;
    }

#ifdef CONFIG_IFX_PPA_A4
    if ( (flags & PPA_F_SESSION_OUT_VLAN) )
    {
        if ( p_extra->out_vlan_insert )
        {
            p_item->flags |= SESSION_VALID_OUT_VLAN_INS;
            p_item->out_vlan_tag = p_extra->out_vlan_tag;
        }
        else
        {
            p_item->flags &= ~SESSION_VALID_OUT_VLAN_INS;
            p_item->out_vlan_tag = 0;
        }

        if ( p_extra->out_vlan_remove )
            p_item->flags |= SESSION_VALID_OUT_VLAN_RM;
        else
            p_item->flags &= ~SESSION_VALID_OUT_VLAN_RM;
    }
#endif

    return 0;
}

#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API

static INLINE int32_t ppa_add_mc_session(PPA_MC_SESSION *p_mc_session, struct session_list_item **pp_item, uint32_t flags)
{
    struct session_list_item *p_item;
    uint32_t i;

    p_item = ppa_mem_cache_alloc(g_session_item_cache);
    if ( !p_item )
        return -1;

    ppa_init_session_list_item(p_item);

    //  broadcast address
    p_item->dst_ip = p_mc_session->ip_mc_group;

  #if 0
    if ( p_mc_session->replace_src_mac )
    {
        p_item->flags |= SESSION_VALID_SRC_MAC;
        ppa_memcpy(p_item->src_mac, p_mc_session->src_mac, PPA_ETH_ALEN);
    }
  #else
    p_item->flags |= SESSION_VALID_SRC_MAC;
  #endif

    for ( i = 0; i < p_mc_session->num_ifs; i++ )
    {
        if ( strcmp(p_mc_session->array_mem_ifs[i].ifname, eth0_ifname) == 0 )
            p_item->mc_dest_list |= IFX_PPA_DEST_LIST_ETH0;
        if ( strcmp(p_mc_session->array_mem_ifs[i].ifname, eth1_ifname) == 0 )
            p_item->mc_dest_list |= IFX_PPA_DEST_LIST_ETH1;
    }

    //  TODO: similar process as ppa_add_session to support VLAN

    ppa_insert_mc_session_item(p_item);

    *pp_item = p_item;

    return 0;
}

static INLINE int32_t ppa_update_mc_session_extra(PPA_SESSION_EXTRA *p_extra, struct session_list_item *p_item, uint32_t flags)
{
    p_item->session_extra = *p_extra;

    if ( (flags & PPA_F_SESSION_VLAN) )
    {
        if ( p_extra->vlan_insert )
        {
            p_item->flags |= SESSION_VALID_VLAN_INS;
            p_item->new_vci = ((uint32_t)p_extra->vlan_prio << 13) | ((uint32_t)p_extra->vlan_cfi << 12) | p_extra->vlan_id;
        }
        else
            p_item->flags &= ~SESSION_VALID_VLAN_INS;

        if ( p_extra->vlan_remove )
            p_item->flags |= SESSION_VALID_VLAN_RM;
        else
            p_item->flags &= ~SESSION_VALID_VLAN_RM;
    }

  #ifdef CONFIG_IFX_PPA_A4
    if ( (flags & PPA_F_SESSION_OUT_VLAN) )
    {
        if ( p_extra->out_vlan_insert )
        {
            p_item->flags |= SESSION_VALID_OUT_VLAN_INS;
            p_item->out_vlan_tag = p_extra->out_vlan_tag;
        }
        else
        {
            p_item->flags &= ~SESSION_VALID_OUT_VLAN_INS;
            p_item->out_vlan_tag = 0;
        }

        if ( p_extra->out_vlan_remove )
            p_item->flags |= SESSION_VALID_OUT_VLAN_RM;
        else
            p_item->flags &= ~SESSION_VALID_OUT_VLAN_RM;
    }
  #endif

    return 0;
}

#else

static INLINE int32_t ppa_add_mc_group(PPA_MC_GROUP *p_mc_group, struct mc_group_list_item **pp_item, uint32_t flags)
{
    struct mc_group_list_item *p_item;
    struct netif_info *p_netif_info;
    PPA_NETIF *netif;
    uint32_t bit;
    uint32_t idx;
    uint32_t i;

    p_item = ppa_mem_cache_alloc(g_mc_group_item_cache);
    if ( !p_item )
        return -1;

    ppa_init_mc_group_list_item(p_item);

    //  braodcast address
    p_item->ip_mc_group = p_mc_group->ip_mc_group;

    if ( (get_if_type(0) & IFX_PPA_IF_TYPE_LAN) && netif_info_is_added(eth0_ifname, &p_netif_info) && (p_netif_info->flags & NETIF_MAC_ENTRY_CREATED) )
    {
        //  if ETH0 is one of destinations, use its MAC, otherwise try ETH1
        p_item->src_mac_entry = p_netif_info->mac_entry;
        p_item->flags |= SESSION_VALID_SRC_MAC;
    }
    else if ( (get_if_type(1) & IFX_PPA_IF_TYPE_LAN) && netif_info_is_added(eth1_ifname, &p_netif_info) && (p_netif_info->flags & NETIF_MAC_ENTRY_CREATED) )
    {
        p_item->src_mac_entry = p_netif_info->mac_entry;
        p_item->flags |= SESSION_VALID_SRC_MAC;
    }

    for ( i = 0, bit = 1, idx = 0; i < MAX_MC_IFS_NUM && idx < p_mc_group->num_ifs; i++, bit <<= 1 )
        if ( (p_mc_group->if_mask & bit) )
        {
            netif = ppa_get_netif(p_mc_group->array_mem_ifs[i].ifname);
            if ( netif && (ppa_get_iftype(netif) &  IFX_PPA_IF_TYPE_WAN) )
            {
                p_item->netif[idx] = netif;
                p_item->ttl[idx]   = p_mc_group->array_mem_ifs[i].ttl;
                idx++;

                //  PPPoE
                if ( ppa_check_is_ppp_netif(netif) && ppa_check_is_pppoe_netif(netif) )
                    p_item->flags |= SESSION_VALID_PPPOE;

                //  VLAN
                //  TODO: no method to detect VLAN on LAN side
                if ( ppa_if_is_vlan_if(netif, NULL) )
                    p_item->flags |= SESSION_VALID_VLAN_RM;
            }
        }

    p_item->num_ifs = idx;
    p_item->if_mask = (1 << idx) - 1;

  #ifdef CONFIG_IFX_PPA_A4
    if ( (flags & PPA_F_SESSION_VLAN) )
    {
        if ( p_mc_group->vlan_insert )
        {
            p_item->flags |= SESSION_VALID_VLAN_INS;
            p_item->new_vci = ((uint32_t)p_mc_group->vlan_prio << 13) | ((uint32_t)p_mc_group->vlan_cfi << 12) | (uint32_t)p_mc_group->vlan_id;
        }
        else
        {
            p_item->flags &= ~SESSION_VALID_VLAN_INS;
            p_item->new_vci = 0;
        }

        if ( p_mc_group->vlan_remove )
            p_item->flags |= SESSION_VALID_VLAN_RM;
        else
            p_item->flags &= ~SESSION_VALID_VLAN_RM;
    }

    if ( (flags & PPA_F_SESSION_OUT_VLAN) )
    {
        if ( p_mc_group->out_vlan_insert )
        {
            p_item->flags |= SESSION_VALID_OUT_VLAN_INS;
            p_item->out_vlan_tag = p_mc_group->out_vlan_tag;
        }
        else
        {
            p_item->flags &= ~SESSION_VALID_OUT_VLAN_INS;
            p_item->out_vlan_tag = 0;
        }

        if ( p_mc_group->out_vlan_remove )
            p_item->flags |= SESSION_VALID_OUT_VLAN_RM;
        else
            p_item->flags &= ~SESSION_VALID_OUT_VLAN_RM;
    }
  #endif

    ppa_insert_mc_group_item(p_item);

    *pp_item = p_item;

    return 0;
}

#endif

static INLINE int32_t ppa_hw_add_session(struct session_list_item *p_item)
{
    int32_t dst_ifid;
    uint8_t mac[PPA_ETH_ALEN];
    uint32_t routing_entry;
    uint32_t route_type;
    uint32_t src_mac_ix = ~0, mtu_ix = ~0, pppoe_ix = ~0;
#ifdef CONFIG_IFX_PPA_A4
    uint32_t out_vlan_ix = ~0;
#endif
    uint32_t new_dscp = 0;
    uint32_t new_ip = 0;
    uint32_t new_port = 0;
    uint32_t dest_list = 0;
    struct bridging_session_list_item *p_br_item;

    //  Only add session in H/w when the called from the post-NAT hook

    switch ( ppa_get_iftype(p_item->rx_if) )
    {
    case IFX_PPA_IF_NOT_FOUND:
        dbg("rx_if %s not found", p_item->rx_if->name);
        SET_DBG_FLAG(p_item, SESSION_DBG_RX_IF_TYPE_ERR);
        return -1;
    case IFX_PPA_IF_TYPE_LAN:
        p_item->flags |= SESSION_LAN_ENTRY;
        break;
    case IFX_PPA_IF_TYPE_WAN:
        p_item->flags |= SESSION_WAN_ENTRY;
        break;
    default:
        switch ( ppa_get_iftype(p_item->tx_if) )
        {
        case IFX_PPA_IF_NOT_FOUND:
            dbg("rx_if %s is mix type, tx_if %s is not found", p_item->rx_if->name, p_item->tx_if->name);
            SET_DBG_FLAG(p_item, SESSION_DBG_TX_IF_TYPE_ERR);
            return -1;
        case IFX_PPA_IF_TYPE_LAN:
            p_item->flags |= SESSION_WAN_ENTRY; //  dest is LAN, so entry is WAN
            break;
        case IFX_PPA_IF_TYPE_WAN:
            p_item->flags |= SESSION_LAN_ENTRY; //  dest is WAN, so entry is LAN
            break;
        default:
            dbg("both rx/tx netif is mix type, not supported yet");
            SET_DBG_FLAG(p_item, SESSION_DBG_RX_IF_TYPE_ERR | SESSION_DBG_TX_IF_TYPE_ERR);
            return -1;
        }
    }

    if ( ppa_if_is_br_if(p_item->tx_if, NULL) )
    {
#if defined(ENABLE_ASSERT) && ENABLE_ASSERT
        if ( in_irq() )
            err("semaphore down in irq");
#endif

        //  Warning: not given the dest interface
        if ( ppa_bridging_lookup_session(p_item->dst_mac, NULL, &p_br_item) != IFX_PPA_SESSION_EXISTS)
        {
            SET_DBG_FLAG(p_item, SESSION_DBG_DST_BRG_IF_NOT_IN_BRG_TBL);
            return -1;
        }

        if ( (dst_ifid = ppa_get_ifid(p_br_item->netif)) < 0 )
        {
            SET_DBG_FLAG(p_item, SESSION_DBG_DST_IF_NOT_IN_IF_LIST);
            return -1;
        }
    }
    else
    {
        if ( (dst_ifid = ppa_get_ifid(p_item->tx_if)) < 0 )
        {
            SET_DBG_FLAG(p_item, SESSION_DBG_DST_IF_NOT_IN_IF_LIST);
            return -1;
        }
    }

    route_type = (p_item->flags & SESSION_VALID_NAT_IP) ? ((p_item->flags & SESSION_VALID_NAT_PORT) ? IFX_PPA_ROUTE_TYPE_NAPT : IFX_PPA_ROUTE_TYPE_NAT) : IFX_PPA_ROUTE_TYPE_IPV4;
    if ( (p_item->flags & SESSION_VALID_NAT_IP) )
        new_ip = p_item->nat_ip;
    if ( (p_item->flags & SESSION_VALID_NAT_PORT) )
        new_port = p_item->nat_port;

    if ( (p_item->flags & SESSION_VALID_PPPOE) )
    {
        if ( add_pppoe_entry(p_item->pppoe_session_id, &pppoe_ix) == 0 )
            p_item->pppoe_entry = pppoe_ix;
        else
        {
            dbg("add pppoe_entry error");
            SET_DBG_FLAG(p_item, SESSION_DBG_ADD_PPPOE_ENTRY_FAIL);
            goto SESSION_VALID_PPPOE_ERROR;
        }
    }

    if ( add_mtu_entry(p_item->mtu, &mtu_ix) == 0 )
    {
        p_item->mtu_entry = mtu_ix;
        p_item->flags |= SESSION_VALID_MTU;
    }
    else
    {
        SET_DBG_FLAG(p_item, SESSION_DBG_ADD_MTU_ENTRY_FAIL);
        goto MTU_ERROR;
    }

    if ( !(p_item->flags & SESSION_TX_ITF_IPOA_PPPOA_MASK) )
    {
        ppa_get_netif_hwaddr(p_item->tx_if, mac);
        if ( add_mac_entry(mac, &src_mac_ix) == 0 )
        {
            p_item->src_mac_entry = src_mac_ix;
            p_item->flags |= IFX_PPA_UPDATE_ROUTING_ENTRY_NEW_SRC_MAC_IX;
        }
        else
        {
            SET_DBG_FLAG(p_item, SESSION_DBG_ADD_MAC_ENTRY_FAIL);
            goto NEW_SRC_MAC_ERROR;
        }
    }

#ifdef CONFIG_IFX_PPA_A4
    if ( (p_item->flags & SESSION_VALID_OUT_VLAN_INS) )
    {
        if ( add_outer_vlan_entry(p_item->out_vlan_tag, &out_vlan_ix) == 0 )
            p_item->out_vlan_ix = out_vlan_ix;
        else
        {
            dbg("add out_vlan_ix error");
            SET_DBG_FLAG(p_item, SESSION_DBG_ADD_OUT_VLAN_ENTRY_FAIL);
            goto OUT_VLAN_ERROR;
        }
    }
#endif

    if ( (p_item->flags & SESSION_VALID_NEW_DSCP) )
        new_dscp = p_item->new_dscp;

#if 1
    dest_list = 1 << dst_ifid;
    dbg("dest_list = %02X", dest_list);
#else
    if ( (p_item->flags & SESSION_LAN_ENTRY) )
    {
        dbg("LAN: get_if_type(0) = %d", get_if_type(0));
        dbg("LAN: get_if_type(1) = %d", get_if_type(1));

        //  LAN
        if ( (get_if_type(0) & IFX_PPA_IF_TYPE_WAN) )
            dest_list |= IFX_PPA_DEST_LIST_ETH0;
        if ( (get_if_type(1) & IFX_PPA_IF_TYPE_WAN) )
            dest_list |= IFX_PPA_DEST_LIST_ETH1;

        dbg("dest_list = %02X", dest_list);
    }
    else
    {
        dbg("WAN: get_if_type(0) = %d", get_if_type(0));
        dbg("WAN: get_if_type(1) = %d", get_if_type(1));

        //  WAN
        if ( (get_if_type(0) & IFX_PPA_IF_TYPE_LAN) )
            dest_list |= IFX_PPA_DEST_LIST_ETH0;
        if ( (get_if_type(1) & IFX_PPA_IF_TYPE_LAN) )
            dest_list |= IFX_PPA_DEST_LIST_ETH1;

        dbg("dest_list = %02X", dest_list);
    }
#endif

    if ( add_routing_entry(p_item->flags & SESSION_LAN_ENTRY,
                           p_item->src_ip,
                           p_item->src_port,
                           p_item->dst_ip,
                           p_item->dst_port,
                           route_type,
                           new_ip,
                           new_port,
                           p_item->dst_mac,
                           src_mac_ix,
                           mtu_ix,
                           p_item->flags & SESSION_VALID_NEW_DSCP,
                           new_dscp,
                           p_item->flags & SESSION_VALID_VLAN_INS,
                           p_item->new_vci,
                           p_item->flags & SESSION_VALID_VLAN_RM,
                           p_item->flags & SESSION_VALID_PPPOE,
                           pppoe_ix,
#ifdef CONFIG_IFX_PPA_A4
                           p_item->flags & SESSION_VALID_OUT_VLAN_INS,
                           out_vlan_ix,
                           p_item->flags & SESSION_VALID_OUT_VLAN_RM,
                           p_item->dslwan_qid,
#endif
                           dest_list,
                           &routing_entry) == 0 )
    {
        p_item->routing_entry = routing_entry;
        p_item->flags |= SESSION_ADDED_IN_HW;
        return 0;
    }

    //  fail in add_routing_entry
    dbg("fail in add_routing_entry");
#ifdef CONFIG_IFX_PPA_A4
    p_item->out_vlan_ix = ~0;
    del_outer_vlan_entry(out_vlan_ix);
OUT_VLAN_ERROR:
#endif
    p_item->src_mac_entry = ~0;
    del_mac_entry(src_mac_ix);
NEW_SRC_MAC_ERROR:
    p_item->mtu_entry = ~0;
    del_mtu_entry(mtu_ix);
MTU_ERROR:
    p_item->pppoe_entry = ~0;
    del_pppoe_entry(pppoe_ix);
SESSION_VALID_PPPOE_ERROR:
    return -1;
}

static INLINE int32_t ppa_hw_update_session_extra(struct session_list_item *p_item, uint32_t flags)
{
    uint32_t update_flags = 0;
    uint32_t mtu;
    uint32_t mtu_ix = ~0;
#ifdef CONFIG_IFX_PPA_A4
    uint32_t out_vlan_tag;
    uint32_t out_vlan_ix = ~0;
#endif

    if ( (flags & PPA_F_SESSION_NEW_DSCP) )
        update_flags |= IFX_PPA_UPDATE_ROUTING_ENTRY_NEW_DSCP_EN | IFX_PPA_UPDATE_ROUTING_ENTRY_NEW_DSCP;

    if ( (flags & PPA_F_SESSION_VLAN) )
        update_flags |= IFX_PPA_UPDATE_ROUTING_ENTRY_VLAN_INS_EN | IFX_PPA_UPDATE_ROUTING_ENTRY_NEW_VCI | IFX_PPA_UPDATE_ROUTING_ENTRY_VLAN_RM_EN;

    if ( (flags & PPA_F_MTU) )
    {
        if ( get_mtu_entry(p_item->mtu_entry, &mtu) == 0 )
        {
            if ( mtu == p_item->mtu )
            {
                //  entry not changed
                mtu_ix = p_item->mtu_entry;
                goto PPA_HW_UPDATE_SESSION_EXTRA_MTU_GOON;
            }
            else
            {
                //  entry changed, so delete old first and create new one later
                del_mtu_entry(p_item->mtu_entry);
                p_item->mtu_entry = ~0;
            }
        }
        //  create new MTU entry
        if ( add_mtu_entry(p_item->mtu, &mtu_ix) == 0 )
        {
            //  success
            p_item->mtu_entry = mtu_ix;
            update_flags |= IFX_PPA_UPDATE_ROUTING_ENTRY_MTU_IX;
        }
        else
            return -1;
    }
PPA_HW_UPDATE_SESSION_EXTRA_MTU_GOON:

#ifdef CONFIG_IFX_PPA_A4
    if ( (flags & PPA_F_SESSION_OUT_VLAN) )
    {
        if ( get_outer_vlan_entry(p_item->out_vlan_ix, &out_vlan_tag) == 0 )
        {
            if ( out_vlan_tag == p_item->out_vlan_tag )
            {
                //  entry not changed
                out_vlan_ix = p_item->out_vlan_ix;
                goto PPA_HW_UPDATE_SESSION_EXTRA_OUT_VLAN_GOON;
            }
            else
            {
                //  entry changed, so delete old first and create new one later
                del_outer_vlan_entry(p_item->out_vlan_ix);
                p_item->out_vlan_ix = ~0;
            }
        }
        //  create new OUT VLAN entry
        if ( add_outer_vlan_entry(p_item->out_vlan_tag, &out_vlan_ix) == 0 )
        {
            p_item->out_vlan_ix = out_vlan_ix;
            update_flags |= IFX_PPA_UPDATE_ROUTING_ENTRY_OUT_VLAN_IX;
        }

        update_flags |= IFX_PPA_UPDATE_ROUTING_ENTRY_OUT_VLAN_INS_EN | IFX_PPA_UPDATE_ROUTING_ENTRY_OUT_VLAN_RM_EN;
    }
PPA_HW_UPDATE_SESSION_EXTRA_OUT_VLAN_GOON:
#endif

    update_routing_entry(p_item->routing_entry,
                         0,
                         0,
                         0,
                         NULL,
                         0,     //  src_mac_ix
                         mtu_ix,
                         p_item->flags & SESSION_VALID_NEW_DSCP,
                         p_item->new_dscp,
                         p_item->flags & SESSION_VALID_VLAN_INS,
                         p_item->new_vci,
                         p_item->flags & SESSION_VALID_VLAN_RM,
                         0,
                         0,
#ifdef CONFIG_IFX_PPA_A4
                         p_item->flags & SESSION_VALID_OUT_VLAN_INS,    //  f_out_vlan_ins_enable
                         p_item->out_vlan_ix,                           //  out_vlan_ix
                         p_item->flags & SESSION_VALID_OUT_VLAN_RM,     //  f_out_vlan_rm_enable
                         0,
#endif
                         0,
                         update_flags);

    return 0;
}

static INLINE void ppa_hw_del_session(struct session_list_item *p_item)
{
    //  when init, these entry values are ~0, the max the number which can be detected by these functions
    del_routing_entry(p_item->routing_entry);
    p_item->routing_entry = ~0;

#ifdef CONFIG_IFX_PPA_A4
    del_outer_vlan_entry(p_item->out_vlan_ix);
    p_item->out_vlan_ix = ~0;
#endif

#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    del_wan_mc_entry(p_item->mc_routing_entry);
    p_item->mc_routing_entry = ~0;
#endif

    del_pppoe_entry(p_item->pppoe_entry);
    p_item->pppoe_entry = ~0;

    del_mtu_entry(p_item->mtu_entry);
    p_item->mtu_entry = ~0;

    del_mac_entry(p_item->src_mac_entry);
    p_item->src_mac_entry = ~0;
}

#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API

static INLINE int32_t ppa_hw_add_mc_session(struct session_list_item *p_item)
{
    uint32_t mc_routing_entry;

    if ( add_wan_mc_entry(p_item->dst_ip,
                          p_item->flags & SESSION_VALID_VLAN_INS,
                          p_item->new_vci,
                          p_item->flags & SESSION_VALID_VLAN_RM,
                          p_item->flags & SESSION_VALID_SRC_MAC,
                          0,    //  TODO: src_mac_ix
                          0,    //  pppoe_mode
  #ifdef CONFIG_IFX_PPA_A4
                          0,
                          0,
                          0,    //  TODO out vlan
  #endif
                          p_item->mc_dest_list,
                          &mc_routing_entry) == 0 )
    {
        p_item->mc_routing_entry = mc_routing_entry;
        p_item->flags |= SESSION_ADDED_IN_HW;
        return 0;
    }
    else
        return -1;
}

static INLINE int32_t ppa_hw_update_mc_session_extra(struct session_list_item *p_item, uint32_t flags)
{
//    uint32_t ret = 0;
    uint32_t update_flags = 0;
    uint32_t new_vci;
    uint32_t vlan_ix = ~0;

    if ( (flags & PPA_F_SESSION_VLAN) )
    {
        if ( !(p_item->flags & SESSION_VALID_VLAN_INS) )
        {
            //  disble VLAN insert, delete existing VLAN entry
            del_vlan_entry(p_item->vlan_entry);
            p_item->vlan_entry = ~0;
        }
        else
        {
            if ( get_vlan_entry(p_item->vlan_entry, &new_vci) == 0 )
            {
                //  old VLAN entry exists
                if ( new_vci == p_item->new_vci )
                {
                    //  entry not changed
                    vlan_ix = p_item->vlan_entry;
                    goto PPA_HW_UPDATE_MC_SESSION_EXTRA_VLAN_GOON;
                }
                else
                {
                    //  entry changed, so delete old first and create new one later
                    del_vlan_entry(p_item->vlan_entry);
                    p_item->vlan_entry = ~0;
                }
            }
            //  create new VLAN entry
            if ( add_vlan_entry(p_item->new_vci, &vlan_ix) == 0 )
                //  success
                p_item->vlan_entry = vlan_ix;
            else
            {
                //  failure, disable VLAN insert
                p_item->session_extra.dscp_remark = 0;
                p_item->flags &= ~SESSION_VALID_VLAN_INS;
//                ret = -1;
                return -1;
            }
        }

PPA_HW_UPDATE_MC_SESSION_EXTRA_VLAN_GOON:
        update_flags |= IFX_PPA_UPDATE_ROUTING_ENTRY_VLAN_INS_EN | IFX_PPA_UPDATE_ROUTING_ENTRY_NEW_VCI | IFX_PPA_UPDATE_ROUTING_ENTRY_VLAN_RM_EN;
    }

    update_wan_mc_entry(p_item->mc_routing_entry,
                        p_item->flags & SESSION_VALID_VLAN_INS,
                        vlan_ix,
                        p_item->flags & SESSION_VALID_VLAN_RM,
                        0,
                        0,
                        0,
                        0,
                        update_flags);

//    return ret;
    return 0;
}

static INLINE void ppa_hw_del_mc_session(struct session_list_item *p_item)
{
    //  XXX: same process as normal session
    ppa_hw_del_session(p_item);
}

#else

static INLINE int32_t ppa_hw_add_mc_group(struct mc_group_list_item *p_item)
{
    uint32_t entry;
    uint32_t dest_list = 0;
  #ifdef CONFIG_IFX_PPA_A4
    uint32_t out_vlan_ix = ~0;
  #endif

    //  must be LAN port
    if ( (get_if_type(0) & IFX_PPA_IF_TYPE_LAN) )
        dest_list |= IFX_PPA_DEST_LIST_ETH0;
    if ( (get_if_type(1) & IFX_PPA_IF_TYPE_LAN) )
        dest_list |= IFX_PPA_DEST_LIST_ETH1;

  #ifdef CONFIG_IFX_PPA_A4
    if ( (p_item->flags & SESSION_VALID_OUT_VLAN_INS) )
    {
        if ( add_outer_vlan_entry(p_item->out_vlan_tag, &out_vlan_ix) == 0 )
            p_item->out_vlan_ix = out_vlan_ix;
        else
        {
            dbg("add out_vlan_ix error");
            goto OUT_VLAN_ERROR;
        }
    }
  #endif

    if ( add_wan_mc_entry(p_item->ip_mc_group,
                          p_item->flags & SESSION_VALID_VLAN_INS,
                          p_item->new_vci,
                          p_item->flags & SESSION_VALID_VLAN_RM,
                          p_item->flags & SESSION_VALID_SRC_MAC,
                          p_item->src_mac_entry,
                          p_item->flags & SESSION_VALID_PPPOE,
  #ifdef CONFIG_IFX_PPA_A4
                          p_item->flags & SESSION_VALID_OUT_VLAN_INS,
                          p_item->out_vlan_ix,
                          p_item->flags & SESSION_VALID_OUT_VLAN_RM,
  #endif
                          dest_list,
                          &entry) == 0 )
    {
        p_item->mc_entry = entry;
        p_item->flags |= SESSION_ADDED_IN_HW;
        return 0;
    }

  #ifdef CONFIG_IFX_PPA_A4
    p_item->out_vlan_ix = ~0;
    del_outer_vlan_entry(out_vlan_ix);
OUT_VLAN_ERROR:
  #endif
    return -1;
}

  #if 0
static INLINE int32_t ppa_hw_update_mc_group(struct mc_group_list_item *p_item, uint32_t flags)
{
    uint32_t update_flags = 0;

    if ( (flags & PPA_F_SESSION_VLAN) )
        update_flags |= IFX_PPA_UPDATE_ROUTING_ENTRY_VLAN_INS_EN | IFX_PPA_UPDATE_ROUTING_ENTRY_NEW_VCI | IFX_PPA_UPDATE_ROUTING_ENTRY_VLAN_RM_EN;

    update_wan_mc_entry(p_item->mc_entry,
                        p_item->flags & SESSION_VALID_VLAN_INS,
                        p_item->new_vci,
                        p_item->flags & SESSION_VALID_VLAN_RM,
                        0,
                        0,
                        0,
                        0,
                        0,
                        update_flags);

    return 0;
}
  #endif

static INLINE void ppa_hw_del_mc_group(struct mc_group_list_item *p_item)
{
    //  when init, these entry values are ~0, the max the number which can be detected by these functions
    del_wan_mc_entry(p_item->mc_entry);
    p_item->mc_entry = ~0;

    //  taken from netif_info, so don't need to be removed from MAC table
    //del_mac_entry(p_item->src_mac_entry);
    p_item->src_mac_entry = ~0;
}

#endif

static INLINE void ppa_insert_session_item(struct session_list_item *p_item)
{
    uint32_t idx;

    idx = SESSION_LIST_HASH_VALUE(p_item->session, p_item->flags & SESSION_IS_REPLY);
    ppa_lock_get(&g_session_list_lock);
    p_item->next = g_session_list_hash_table[idx];
    g_session_list_hash_table[idx] = p_item;
    g_session_list_length++;
    ppa_lock_release(&g_session_list_lock);
}

static INLINE void ppa_remove_session_item(struct session_list_item *p_item)
{
    uint32_t idx;
    struct session_list_item *p_prev, *p_cur;

    idx = SESSION_LIST_HASH_VALUE(p_item->session, p_item->flags & SESSION_IS_REPLY);
    p_prev = NULL;
    ppa_lock_get(&g_session_list_lock);
    p_cur = g_session_list_hash_table[idx];
    while ( p_cur && p_cur != p_item )
    {
        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( !p_prev )
            g_session_list_hash_table[idx] = p_cur->next;
        else
            p_prev->next = p_cur->next;

        p_cur->next = NULL;

        g_session_list_length--;
    }
    ppa_lock_release(&g_session_list_lock);
}

#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API

static INLINE void ppa_insert_mc_session_item(struct session_list_item *p_item)
{
    uint32_t idx;

    idx = SESSION_LIST_MC_HASH_VALUE(p_item->dst_ip);
    ppa_lock_get(&g_mc_session_list_lock);
    p_item->next = g_mc_session_list_hash_table[idx];
    g_mc_session_list_hash_table[idx] = p_item;
//    g_mc_session_list_length++;
    ppa_lock_release(&g_mc_session_list_lock);
}

static INLINE void ppa_remove_mc_session_item(struct session_list_item *p_item)
{
    uint32_t idx;
    struct session_list_item *p_prev, *p_cur;

    idx = SESSION_LIST_MC_HASH_VALUE(p_item->dst_ip);
    p_prev = NULL;
    ppa_lock_get(&g_mc_session_list_lock);
    p_cur = g_mc_session_list_hash_table[idx];
    while ( p_cur && p_cur != p_item )
    {
        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( !p_prev )
            g_mc_session_list_hash_table[idx] = p_cur->next;
        else
            p_prev->next = p_cur->next;

        p_cur->next = NULL;

//        g_mc_session_list_length--;
    }
    ppa_lock_release(&g_mc_session_list_lock);
}

#else

static INLINE void ppa_insert_mc_group_item(struct mc_group_list_item *p_item)
{
    uint32_t idx;

    idx = SESSION_LIST_MC_HASH_VALUE(p_item->ip_mc_group);
    ppa_lock_get(&g_mc_group_list_lock);
    p_item->next = g_mc_group_list_hash_table[idx];
    g_mc_group_list_hash_table[idx] = p_item;
    g_mc_group_list_length++;
    ppa_lock_release(&g_mc_group_list_lock);
}

static INLINE void ppa_remove_mc_group_item(struct mc_group_list_item *p_item)
{
    uint32_t idx;
    struct mc_group_list_item *p_prev, *p_cur;

    idx = SESSION_LIST_MC_HASH_VALUE(p_item->ip_mc_group);
    p_prev = NULL;
    ppa_lock_get(&g_mc_group_list_lock);
    p_cur = g_mc_group_list_hash_table[idx];
    while ( p_cur && p_cur != p_item )
    {
        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( !p_prev )
            g_mc_group_list_hash_table[idx] = p_cur->next;
        else
            p_prev->next = p_cur->next;

        p_cur->next = NULL;

        g_mc_group_list_length--;
    }
    ppa_lock_release(&g_mc_group_list_lock);
}

#endif

static INLINE uint32_t ppa_get_default_session_timeout(void)
{
    return DEFAULT_TIMEOUT_IN_SEC;
}

static void ppa_check_hit_stat(unsigned long dummy)
{
    struct session_list_item *p_item;
    uint32_t i;

    //  XXX: not lock the list, expect removing item won't affect this function
    ppa_lock_get(&g_session_list_lock);
    for ( i = 0; i < NUM_ENTITY(g_session_list_hash_table); i++ )
        for ( p_item = g_session_list_hash_table[i]; p_item; p_item = p_item->next )
            if ( test_and_clear_hit_stat(p_item->routing_entry) )
                p_item->last_hit_time = ppa_get_time_in_sec();
    ppa_lock_release(&g_session_list_lock);

#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    ppa_lock_get(&g_mc_session_list_lock);
    for ( i = 0; i < NUM_ENTITY(g_mc_session_list_hash_table); i++ )
        for ( p_item = g_mc_session_list_hash_table[i]; p_item; p_item = p_item->next )
            if ( test_and_clear_mc_hit_stat(p_item->mc_routing_entry) )
                p_item->last_hit_time = ppa_get_time_in_sec();
    ppa_lock_release(&g_mc_session_list_lock);
#endif

    ppa_timer_add(&g_hit_stat_timer, g_hit_polling_time);
}

int32_t ppa_bridging_lookup_session(uint8_t *mac, PPA_NETIF *netif, struct bridging_session_list_item **pp_item)
{
    int32_t ret;
    struct bridging_session_list_item *p_prev, *p_cur;
    struct bridging_session_list_item **pp_header;

    ASSERT(pp_item != NULL, "pp_item == NULL");

    ret = IFX_PPA_SESSION_NOT_ADDED;

    pp_header = g_bridging_session_list_hash_table + BRIDGING_SESSION_LIST_HASH_VALUE(mac);
    p_prev = NULL;
    ppa_lock_get(&g_bridging_session_list_lock);
    p_cur = *pp_header;
//    while ( p_cur && !(*(uint32_t *)mac == *(uint32_t *)p_cur->mac && *((uint16_t *)mac + 2) == *((uint16_t *)p_cur->mac + 2) && ppa_is_netif_equal(netif, p_cur->netif)) )
    while ( p_cur && !(*(uint32_t *)mac == *(uint32_t *)p_cur->mac && *((uint16_t *)mac + 2) == *((uint16_t *)p_cur->mac + 2)) )
    {
        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( p_prev )
        {
            p_prev->next = p_cur->next;
            p_cur->next = *pp_header;
            *pp_header = p_cur;
        }

        ret = IFX_PPA_SESSION_EXISTS;
    }
    ppa_lock_release(&g_bridging_session_list_lock);

    *pp_item = p_cur;

    return ret;
}

static INLINE int32_t ppa_bridging_add_session(uint8_t *mac, PPA_NETIF *netif, struct bridging_session_list_item **pp_item, uint32_t flags)
{
   struct bridging_session_list_item *p_item;
#ifdef CONFIG_IFX_PPA_A4
    PPA_IFNAME phys_netif_name[PPA_IF_NAME_SIZE];
#endif

    p_item = ppa_mem_cache_alloc(g_bridging_session_item_cache);
    if ( !p_item )
        return -1;

    ppa_bridging_init_session_list_item(p_item);

    dump_bridging_list_item(p_item, "ppa_bridging_add_session (after init)");

    ppa_memcpy(p_item->mac, mac, PPA_ETH_ALEN);
    p_item->netif         = netif;
    p_item->timeout       = ppa_bridging_get_default_session_timeout();
    p_item->last_hit_time = ppa_get_time_in_sec();

    //  TODO: vlan related fields

#ifdef CONFIG_IFX_PPA_A4
    if ( ppa_get_physical_if(netif, NULL, phys_netif_name) >= 0 && (netif = ppa_get_netif(phys_netif_name)) )
    {
        PPA_VCC *vcc;
        uint32_t dslwan_qid;

        if ( ppa_br2684_get_vcc(netif, &vcc) >= 0 && (dslwan_qid = get_dslwan_qid_with_vcc(vcc)) >= 0 )
            p_item->dslwan_qid = dslwan_qid;
    }
#endif

    if ( (flags & PPA_F_STATIC_ENTRY) )
    {
        p_item->flags    |= SESSION_STATIC;
        p_item->timeout   = ~0; //  max timeout
    }

    if ( (flags & PPA_F_DROP_PACKET) )
        p_item->flags    |= SESSION_DROP;

    ppa_bridging_insert_session_item(p_item);

    dump_bridging_list_item(p_item, "ppa_bridging_add_session (after setting)");

    *pp_item = p_item;

    return 0;
}

static INLINE int32_t ppa_bridging_hw_add_session(struct bridging_session_list_item *p_item)
{
    uint32_t bridging_entry;
    uint32_t port;
    uint32_t dest_list;
#if 0
    uint32_t if_type;
#endif

#if 1
    switch ( ppa_get_iftype(p_item->netif) )
    {
    case IFX_PPA_IF_NOT_FOUND:
        dbg("netif not found");
        return -1;
    case IFX_PPA_IF_TYPE_LAN:
    case IFX_PPA_IF_TYPE_WAN:
        break;
    default:
        if ( get_if_type(0) == IFX_PPA_IF_TYPE_MIX )
            break;
        dbg("mix type netif is deprecated");
        //return -1;
    }
    if ( (port = ppa_get_ifid(p_item->netif)) < 0 )
        return -1;
    dest_list = 1 << port;
    dbg("dest_list = %02X", dest_list);
#else
    switch ( (if_type = ppa_get_iftype(p_item->netif)) )
    {
    case IFX_PPA_IF_NOT_FOUND:
        dbg("netif not found");
        return -1;
    case IFX_PPA_IF_TYPE_LAN:
    case IFX_PPA_IF_TYPE_WAN:
        if ( (get_if_type(0) & if_type) )
        {
            port = IFX_PPA_PORT_ETH0;
            dest_list = IFX_PPA_DEST_LIST_ETH0;
            break;
        }
        else if ( (get_if_type(1) & if_type) )
        {
            port = IFX_PPA_PORT_ETH1;
            dest_list = IFX_PPA_DEST_LIST_ETH1;
            break;
        }
    default:
        dbg("mix type netif is not supported");
        return -1;
    }
#endif
    if ( (p_item->flags & SESSION_DROP) )
        dest_list = 0;  //  no dest list, dropped

#if !defined(ENABLE_MULTI_QUEUE_PER_PVC) || !ENABLE_MULTI_QUEUE_PER_PVC //Modified by Henry Hsu for PPA4 multiQ test case.

#ifdef CONFIG_IFX_PPA_A4
    if ( add_bridging_entry(port,
                            p_item->mac,
                            p_item->flags & SESSION_SRC_MAC_DROP_EN,
                            p_item->dslwan_qid,
                            dest_list,
                            &bridging_entry) == 0 )
#else
    if ( add_bridging_entry(port,
                            p_item->mac,
                            (p_item->flags & SESSION_BRIDGING_VCI_CHECK) ? p_item->vci : 0,
                            p_item->flags & SESSION_SRC_MAC_DROP_EN,
                            p_item->flags & SESSION_VALID_VLAN_INS,
                            p_item->new_vci,
                            p_item->flags & SESSION_VALID_VLAN_RM,
                            dest_list,
                            &bridging_entry) == 0 )
#endif
    {
        p_item->bridging_entry = bridging_entry;
        p_item->flags |= SESSION_ADDED_IN_HW;
        return 0;
    }


#else


if (    (uint32_t)p_item->mac[0]==0x00 && (uint32_t)p_item->mac[1]==0x33 && (uint32_t)p_item->mac[2]==0x33 && 
	 (uint32_t)p_item->mac[3]==0x33 && (uint32_t)p_item->mac[4]==0x33 && (uint32_t)p_item->mac[5]==0x33)
{
       //803284:fchang.removed printk("henry_debug:%s,%d\n",__func__,__LINE__);
	p_item->dslwan_qid=0;			
	    if ( add_bridging_entry(port,
                            p_item->mac,
                            p_item->flags & SESSION_SRC_MAC_DROP_EN,
                            p_item->dslwan_qid,
                            dest_list,
                            &bridging_entry) == 0 )
	    	{
        		p_item->bridging_entry = bridging_entry;
        		p_item->flags |= SESSION_ADDED_IN_HW;
        		return 0;
	    	}
}
else if (    (uint32_t)p_item->mac[0]==0x00 && (uint32_t)p_item->mac[1]==0x55 && (uint32_t)p_item->mac[2]==0x55 && 
	 (uint32_t)p_item->mac[3]==0x55 && (uint32_t)p_item->mac[4]==0x55 && (uint32_t)p_item->mac[5]==0x55)
{
	//803284:fchang.removed printk("henry_debug:%s,%d\n",__func__,__LINE__);
	p_item->dslwan_qid=1;			
	    if ( add_bridging_entry(port,
                            p_item->mac,
                            p_item->flags & SESSION_SRC_MAC_DROP_EN,
                            p_item->dslwan_qid,
                            dest_list,
                            &bridging_entry) == 0 )
	    	{
        		p_item->bridging_entry = bridging_entry;
        		p_item->flags |= SESSION_ADDED_IN_HW;
        		return 0;
	    	}
		
}
else
{

	//803284:fchang.removed printk("henry_debug:%s,%d\n",__func__,__LINE__);
	p_item->dslwan_qid=3;			
	    if ( add_bridging_entry(port,
                            p_item->mac,
                            p_item->flags & SESSION_SRC_MAC_DROP_EN,
                            p_item->dslwan_qid,
                            dest_list,
                            &bridging_entry) == 0 )
	    	{
        		p_item->bridging_entry = bridging_entry;
        		p_item->flags |= SESSION_ADDED_IN_HW;
        		return 0;
	    	}
}
	
#endif
//803284:fchang.removed printk("henry_debug:%s,%d\n",__func__,__LINE__);
    return -1;
}

static INLINE void ppa_bridging_hw_del_session(struct bridging_session_list_item *p_item)
{
    //  when init, these entry values are ~0, the max the number which can be detected by these functions
    del_bridging_entry(p_item->bridging_entry);
    p_item->bridging_entry = ~0;
}

static INLINE void ppa_bridging_insert_session_item(struct bridging_session_list_item *p_item)
{
    uint32_t idx;

    idx = BRIDGING_SESSION_LIST_HASH_VALUE(p_item->mac);
    ppa_lock_get(&g_bridging_session_list_lock);
    p_item->next = g_bridging_session_list_hash_table[idx];
    g_bridging_session_list_hash_table[idx] = p_item;
    ppa_lock_release(&g_bridging_session_list_lock);
}

static INLINE void ppa_bridging_remove_session_item(struct bridging_session_list_item *p_item)
{
    uint32_t idx;
    struct bridging_session_list_item *p_prev, *p_cur;

    idx = BRIDGING_SESSION_LIST_HASH_VALUE(p_item->mac);
    p_prev = NULL;
    ppa_lock_get(&g_bridging_session_list_lock);
    p_cur = g_bridging_session_list_hash_table[idx];
    while ( p_cur && p_cur != p_item )
    {
        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( !p_prev )
            g_bridging_session_list_hash_table[idx] = p_cur->next;
        else
            p_prev->next = p_cur->next;

        p_cur->next = NULL;
    }
    ppa_lock_release(&g_bridging_session_list_lock);
}

static INLINE uint32_t ppa_bridging_get_default_session_timeout(void)
{
    return DEFAULT_BRIDGING_TIMEOUT_IN_SEC;
}

static void ppa_bridging_check_hit_stat(unsigned long dummy)
{
    struct bridging_session_list_item *p_item;
    uint32_t i;

    ppa_lock_get(&g_bridging_session_list_lock);
    for ( i = 0; i < NUM_ENTITY(g_bridging_session_list_hash_table); i++ )
        for ( p_item = g_bridging_session_list_hash_table[i]; p_item; p_item = p_item->next )
            if ( !(p_item->flags & SESSION_STATIC) && test_and_clear_bridging_hit_stat(p_item->bridging_entry) )
                p_item->last_hit_time = ppa_get_time_in_sec();
    ppa_lock_release(&g_bridging_session_list_lock);

    ppa_timer_add(&g_bridging_hit_stat_timer, g_bridging_hit_polling_time);
}

static INLINE void dump_list_item(struct session_list_item *p_item, char *comment)
{
#if defined(DEBUG_DUMP_LIST_ITEM) && DEBUG_DUMP_LIST_ITEM

    if ( comment )
        printk("dump_list_item - %s\n", comment);
    else
        printk("dump_list_item\n");
    printk("  next             = %08X\n", (uint32_t)p_item->next);
    printk("  session          = %08X\n", (uint32_t)p_item->session);
    printk("  session_extra:\n");
    printk("    new_dscp       = %d\n", (uint32_t)p_item->session_extra.new_dscp);
    printk("    dscp_remark    = %d\n", (uint32_t)p_item->session_extra.dscp_remark);
    printk("    vlan_insert    = %d\n", (uint32_t)p_item->session_extra.vlan_insert);
    printk("    vlan_remove    = %d\n", (uint32_t)p_item->session_extra.vlan_remove);
    printk("    vlan_prio      = %d\n", (uint32_t)p_item->session_extra.vlan_prio);
    printk("    vlan_cfi       = %d\n", (uint32_t)p_item->session_extra.vlan_cfi);
    printk("    vlan_id        = %08X\n", (uint32_t)p_item->session_extra.vlan_id);
    printk("    mtu            = %d\n", (uint32_t)p_item->session_extra.mtu);
    printk("    session_flags  = %08X\n",  (uint32_t)p_item->session_extra.session_flags);
    printk("  ip_proto         = %08X\n", p_item->ip_proto);
    printk("  src_ip           = %d.%d.%d.%d\n", p_item->src_ip >> 24, (p_item->src_ip >> 16) & 0xFF, (p_item->src_ip >> 8) & 0xFF, p_item->src_ip & 0xFF);
    printk("  src_port         = %d\n",   p_item->src_port);
    printk("  src_mac[6]       = %02x:%02x:%02x:%02x:%02x:%02x\n", p_item->src_mac[0], p_item->src_mac[1], p_item->src_mac[2], p_item->src_mac[3], p_item->src_mac[4], p_item->src_mac[5]);
    printk("  dst_ip           = %d.%d.%d.%d\n", p_item->dst_ip >> 24, (p_item->dst_ip >> 16) & 0xFF, (p_item->dst_ip >> 8) & 0xFF, p_item->dst_ip & 0xFF);
    printk("  dst_port         = %d\n",   p_item->dst_port);
    printk("  dst_mac[6]       = %02x:%02x:%02x:%02x:%02x:%02x\n", p_item->dst_mac[0], p_item->dst_mac[1], p_item->dst_mac[2], p_item->dst_mac[3], p_item->dst_mac[4], p_item->dst_mac[5]);
    printk("  nat_ip           = %d.%d.%d.%d\n", p_item->nat_ip >> 24, (p_item->nat_ip >> 16) & 0xFF, (p_item->nat_ip >> 8) & 0xFF, p_item->nat_ip & 0xFF);
    printk("  nat_port         = %d\n",   p_item->nat_port);
    printk("  rx_if            = %08X\n", (uint32_t)p_item->rx_if);
    printk("  tx_if            = %08X\n", (uint32_t)p_item->tx_if);
    printk("  timeout          = %d\n",   p_item->timeout);
    printk("  last_hit_time    = %d\n",   p_item->last_hit_time);
    printk("  num_adds         = %d\n",   p_item->num_adds);
    printk("  pppoe_session_id = %d\n",   p_item->pppoe_session_id);
    printk("  new_dscp         = %d\n",   p_item->new_dscp);
    printk("  new_vci          = %08X\n",  p_item->new_vci);
    printk("  mtu              = %d\n",   p_item->mtu);
  #if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    printk("  mc_dest_list     = %02X\n", p_item->mc_dest_list);
  #endif
    printk("  flags            = %08X\n", p_item->flags);
    printk("  routing_entry    = %08X\n", p_item->routing_entry);
  #if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    printk("  mc_routing_entry = %08X\n", p_item->mc_routing_entry);
  #endif
    printk("  pppoe_entry      = %08X\n", p_item->pppoe_entry);
    printk("  mtu_entry        = %08X\n", p_item->mtu_entry);
    printk("  src_mac_entry    = %08X\n", p_item->src_mac_entry);

#endif
}

static INLINE void dump_bridging_list_item(struct bridging_session_list_item *p_item, char *comment)
{
#if defined(DEBUG_DUMP_LIST_ITEM) && DEBUG_DUMP_LIST_ITEM

    if ( comment )
        printk("dump_bridging_list_item - %s\n", comment);
    else
        printk("dump_bridging_list_item\n");
    printk("  next             = %08X\n", (uint32_t)p_item->next);
    printk("  mac[6]           = %02x:%02x:%02x:%02x:%02x:%02x\n", p_item->mac[0], p_item->mac[1], p_item->mac[2], p_item->mac[3], p_item->mac[4], p_item->mac[5]);
    printk("  netif            = %08X\n", (uint32_t)p_item->netif);
    printk("  timeout          = %d\n",   p_item->timeout);
    printk("  last_hit_time    = %d\n",   p_item->last_hit_time);
    printk("  flags            = %08X\n", p_item->flags);
    printk("  bridging_entry   = %08X\n", p_item->bridging_entry);

#endif
}

static INLINE void proc_file_create(void)
{
    struct proc_dir_entry *res;

    for ( res = proc_root.subdir; res; res = res->next )
        if ( res->namelen == 3
            && res->name[0] == 'p'
            && res->name[1] == 'p'
            && res->name[2] == 'a' ) //  "ppa"
        {
            g_ppa_proc_dir = res;
            break;
        }
    if ( !res )
    {
        g_ppa_proc_dir = proc_mkdir("ppa", NULL);
        g_ppa_proc_dir_flag = 1;
    }

    g_ppa_api_proc_dir = proc_mkdir("api", g_ppa_proc_dir);

#if defined(ENABLE_DBG_PROC) && ENABLE_DBG_PROC
    res = create_proc_read_entry("dbg",
                                  0,
                                  g_ppa_api_proc_dir,
                                  proc_read_dbg,
                                  NULL);
    if ( res )
        res->write_proc = proc_write_dbg;
#endif

    res = create_proc_read_entry("session",
                                  0,
                                  g_ppa_api_proc_dir,
                                  proc_read_session,
                                  NULL);
    if ( res )
        res->write_proc = proc_write_session;

    res = create_proc_read_entry("hook",
                                  0,
                                  g_ppa_api_proc_dir,
                                  proc_read_hook,
                                  NULL);
    if ( res )
        res->write_proc = proc_write_hook;

    res = create_proc_read_entry("netif",
                                  0,
                                  g_ppa_api_proc_dir,
                                  proc_read_netif,
                                  NULL);
    if ( res )
        res->write_proc = proc_write_netif;

#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
    res = create_proc_read_entry("directpath",
                                  0,
                                  g_ppa_api_proc_dir,
                                  proc_read_directpath,
                                  NULL);
    if ( res )
        res->write_proc = proc_write_directpath;
#endif


}

static INLINE void proc_file_delete(void)
{
    remove_proc_entry("netif",
                      g_ppa_api_proc_dir);

    remove_proc_entry("hook",
                      g_ppa_api_proc_dir);

    remove_proc_entry("session",
                      g_ppa_api_proc_dir);

#if defined(ENABLE_DBG_PROC) && ENABLE_DBG_PROC
    remove_proc_entry("dbg",
                      g_ppa_api_proc_dir);
#endif

    remove_proc_entry("api",
                      g_ppa_proc_dir);
    g_ppa_api_proc_dir = NULL;

    if ( g_ppa_proc_dir_flag )
    {
        remove_proc_entry("ppa", NULL);
        g_ppa_proc_dir = NULL;
        g_ppa_proc_dir_flag = 0;
    }
}

#if defined(ENABLE_DBG_PROC) && ENABLE_DBG_PROC
static int proc_read_dbg(char *page, char **start, off_t off, int count, int *eof, void *data)
{
    int len = 0;

    MOD_INC_USE_COUNT;

    if ( g_dbg_enable )
        len += sprintf(page + off + len, "debug enabled (g_dbg_enable = %08X)\n", g_dbg_enable);
    else
        len += sprintf(page + off + len, "debug disabled\n");

    MOD_DEC_USE_COUNT;

    *eof = 1;

    return len;
}

static int proc_write_dbg(struct file *file, const char *buf, unsigned long count, void *data)
{
    char str[64];
    char *p;

    int len, rlen;

    MOD_INC_USE_COUNT;

    len = count < sizeof(str) ? count : sizeof(str) - 1;
    rlen = len - copy_from_user(str, buf, len);
    while ( rlen && str[rlen - 1] <= ' ' )
        rlen--;
    str[rlen] = 0;
    for ( p = str; *p && *p <= ' '; p++, rlen-- );
    if ( !*p )
    {
        MOD_DEC_USE_COUNT;
        return 0;
    }

    if ( stricmp(str, "enable") == 0 )
        g_dbg_enable = 1;
    else
        g_dbg_enable = 0;

    MOD_DEC_USE_COUNT;

    return count;
}
#endif

static int proc_read_session(char *page, char **start, off_t off, int count, int *eof, void *data)
{
    int len = 0;
    int len_max = off + count;
    char *pstr;
    char str[2048];
    int llen;
    int no;
    int i;

    MOD_INC_USE_COUNT;

    pstr = *start = page;

    if ( !g_init )
        len += sprintf(page + off + len, "not init\n");
    else
    {
        llen = sprintf(str, "Session List (%s)\n", g_fwcode == FWCODE_ROUTING_ACC_D2 ? "routing" : (g_fwcode == FWCODE_BRIDGING_ACC_D3 ? "bridging" : "unknown"));
        if ( len <= off && len + llen > off )
        {
            memcpy(pstr, str + off - len, len + llen - off);
            pstr += len + llen - off;
        }
        else if ( len > off )
        {
            memcpy(pstr, str, llen);
            pstr += llen;
        }
        len += llen;

        no = 0;

        switch ( g_fwcode )
        {
        case FWCODE_ROUTING_ACC_D2:
            {
                struct session_list_item *p_cur;

//                ppa_lock_get(&g_session_list_lock);

                for ( i = 0; i < NUM_ENTITY(g_session_list_hash_table); i++ )
                    for ( p_cur = g_session_list_hash_table[i]; p_cur; p_cur = p_cur->next )
                    {
                        llen = print_session(str, ++no, p_cur);
                        if ( len <= off && len + llen > off )
                        {
                            memcpy(pstr, str + off - len, len + llen - off);
                            pstr += len + llen - off;
                        }
                        else if ( len > off )
                        {
                            memcpy(pstr, str, llen);
                            pstr += llen;
                        }
                        len += llen;
                        if ( len >= len_max )
                        {
                            //ppa_lock_release(&g_session_list_lock);
                            goto PROC_READ_SESSION_OVERRUN_END;
                        }
                    }

//                ppa_lock_release(&g_session_list_lock);
            }
            break;

        case FWCODE_BRIDGING_ACC_D3:
            {
                struct bridging_session_list_item *p_cur;

//                ppa_lock_get(&g_bridging_session_list_lock);

                for ( i = 0; i < NUM_ENTITY(g_bridging_session_list_hash_table); i++ )
                    for ( p_cur = g_bridging_session_list_hash_table[i]; p_cur; p_cur = p_cur->next )
                    {
                        llen = print_bridging_session(str, ++no, p_cur);
                        if ( len <= off && len + llen > off )
                        {
                            memcpy(pstr, str + off - len, len + llen - off);
                            pstr += len + llen - off;
                       }
                        else if ( len > off )
                        {
                            memcpy(pstr, str, llen);
                            pstr += llen;
                        }
                        len += llen;
                        if ( len >= len_max )
                        {
//                            ppa_lock_release(&g_bridging_session_list_lock);
                            goto PROC_READ_SESSION_OVERRUN_END;
                        }
                    }

//                ppa_lock_release(&g_bridging_session_list_lock);
            }
            break;

        case FWCODE_ROUTING_BRIDGING_ACC_D4:
        case FWCODE_ROUTING_BRIDGING_ACC_A4:
            {
                struct session_list_item *p_cur;

//                ppa_lock_get(&g_session_list_lock);

                for ( i = 0; i < NUM_ENTITY(g_session_list_hash_table); i++ )
                    for ( p_cur = g_session_list_hash_table[i]; p_cur; p_cur = p_cur->next )
                    {
                        llen = print_session(str, ++no, p_cur);
                        if ( len <= off && len + llen > off )
                        {
                            memcpy(pstr, str + off - len, len + llen - off);
                            pstr += len + llen - off;
                        }
                        else if ( len > off )
                        {
                            memcpy(pstr, str, llen);
                            pstr += llen;
                        }
                        len += llen;
                        if ( len >= len_max )
                        {
                            //ppa_lock_release(&g_session_list_lock);
                            goto PROC_READ_SESSION_OVERRUN_END;
                        }
                    }

//                ppa_lock_release(&g_session_list_lock);
            }
            {
                struct bridging_session_list_item *p_cur;

//                ppa_lock_get(&g_bridging_session_list_lock);

                for ( i = 0; i < NUM_ENTITY(g_bridging_session_list_hash_table); i++ )
                    for ( p_cur = g_bridging_session_list_hash_table[i]; p_cur; p_cur = p_cur->next )
                    {
                        llen = print_bridging_session(str, ++no, p_cur);
                        if ( len <= off && len + llen > off )
                        {
                            memcpy(pstr, str + off - len, len + llen - off);
                            pstr += len + llen - off;
                       }
                        else if ( len > off )
                        {
                            memcpy(pstr, str, llen);
                            pstr += llen;
                        }
                        len += llen;
                        if ( len >= len_max )
                        {
//                            ppa_lock_release(&g_bridging_session_list_lock);
                            goto PROC_READ_SESSION_OVERRUN_END;
                        }
                    }

//                ppa_lock_release(&g_bridging_session_list_lock);
            }
            break;
        }
    }

    *eof = 1;

    MOD_DEC_USE_COUNT;

    return len - off;

PROC_READ_SESSION_OVERRUN_END:

    MOD_DEC_USE_COUNT;

    return len - llen - off;
}

static int proc_write_session(struct file *file, const char *buf, unsigned long count, void *data)
{
    return count;
}

static int proc_read_hook(char *page, char **start, off_t off, int count, int *eof, void *data)
{
    char *ptitle;
    int len = 0;

    switch ( g_fwcode )
    {
    case FWCODE_ROUTING_ACC_D2:
        ptitle = "routing"; break;
    case FWCODE_BRIDGING_ACC_D3:
        ptitle = "bridging"; break;
    case FWCODE_ROUTING_BRIDGING_ACC_D4:
        ptitle = "routing/bridging - d4"; break;
    case FWCODE_ROUTING_BRIDGING_ACC_A4:
        ptitle = "routing/bridging - a4"; break;
    default:
        ptitle = "unknown";
    }

    MOD_INC_USE_COUNT;

    if ( !g_init )
        len += sprintf(page + off + len, "not init (%s)\n", ptitle);
    else
    {
        switch ( g_fwcode )
        {
        case FWCODE_ROUTING_ACC_D2:
            len += sprintf(page + off + len, "%s\n", ptitle);
            len += sprintf(page + off + len, "  ifx_ppa_session_add            - %s\n", ppa_hook_session_add_fn ?       "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_session_delete         - %s\n", ppa_hook_session_del_fn ?       "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_inactivity_status      - %s\n", ppa_hook_inactivity_status_fn ? "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_set_session_inactivity - %s\n", ppa_hook_set_inactivity_fn ?    "hooked" : "not hooked");
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
            len += sprintf(page + off + len, "  ifx_ppa_mc_session_add         - %s\n", ppa_hook_mc_session_add_fn ?    "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_mc_session_delete      - %s\n", ppa_hook_mc_session_del_fn ?    "hooked" : "not hooked");
#else
            len += sprintf(page + off + len, "  ifx_ppa_mc_group_update        - %s\n", ppa_hook_mc_group_update_fn ?   "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_mc_group_get           - %s\n", ppa_hook_mc_group_get_fn ?      "hooked" : "not hooked");
#endif
            break;
        case FWCODE_BRIDGING_ACC_D3:
            len += sprintf(page + off + len, "%s\n", ptitle);
            len += sprintf(page + off + len, "  ifx_ppa_bridge_entry_add               - %s\n", ppa_hook_bridge_entry_add_fn ?               "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_bridge_entry_delete            - %s\n", ppa_hook_bridge_entry_delete_fn ?            "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_set_bridge_entry_timeout       - %s\n", ppa_hook_set_bridge_entry_timeout_fn ?       "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_bridge_entry_inactivity_status - %s\n", ppa_hook_bridge_entry_inactivity_status_fn ? "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_bridge_entry_hit_time          - %s\n", ppa_hook_bridge_entry_hit_time_fn ?          "hooked" : "not hooked");
            break;
        case FWCODE_ROUTING_BRIDGING_ACC_D4:
        case FWCODE_ROUTING_BRIDGING_ACC_A4:
            len += sprintf(page + off + len, "%s\n", "routing");
            len += sprintf(page + off + len, "  ifx_ppa_session_add            - %s\n", ppa_hook_session_add_fn ?       "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_session_delete         - %s\n", ppa_hook_session_del_fn ?       "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_inactivity_status      - %s\n", ppa_hook_inactivity_status_fn ? "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_set_session_inactivity - %s\n", ppa_hook_set_inactivity_fn ?    "hooked" : "not hooked");
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
            len += sprintf(page + off + len, "  ifx_ppa_mc_session_add         - %s\n", ppa_hook_mc_session_add_fn ?    "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_mc_session_delete      - %s\n", ppa_hook_mc_session_del_fn ?    "hooked" : "not hooked");
#else
            len += sprintf(page + off + len, "  ifx_ppa_mc_group_update        - %s\n", ppa_hook_mc_group_update_fn ?   "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_mc_group_get           - %s\n", ppa_hook_mc_group_get_fn ?      "hooked" : "not hooked");
#endif

            len += sprintf(page + off + len, "%s\n", "bridging");
            len += sprintf(page + off + len, "  ifx_ppa_bridge_entry_add               - %s\n", ppa_hook_bridge_entry_add_fn ?               "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_bridge_entry_delete            - %s\n", ppa_hook_bridge_entry_delete_fn ?            "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_set_bridge_entry_timeout       - %s\n", ppa_hook_set_bridge_entry_timeout_fn ?       "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_bridge_entry_inactivity_status - %s\n", ppa_hook_bridge_entry_inactivity_status_fn ? "hooked" : "not hooked");
            len += sprintf(page + off + len, "  ifx_ppa_bridge_entry_hit_time          - %s\n", ppa_hook_bridge_entry_hit_time_fn ?          "hooked" : "not hooked");
            break;
        default:
            len += sprintf(page + off + len, "unknown firmware code\n");
        }
    }

    MOD_DEC_USE_COUNT;

    *eof = 1;

    return len;
}

static int proc_write_hook(struct file *file, const char *buf, unsigned long count, void *data)
{
    char str[64];
    char *p;
    int len, rlen;
    int flag_del = 0;
    int flag_mc_del = 0;
    int flag_bridging_del = 0;

    MOD_INC_USE_COUNT;

    len = count < sizeof(str) ? count : sizeof(str) - 1;
    rlen = len - copy_from_user(str, buf, len);
    while ( rlen && str[rlen - 1] <= ' ' )
        rlen--;
    str[rlen] = 0;
    for ( p = str; *p && *p <= ' '; p++, rlen-- );
    if ( !*p )
    {
        MOD_DEC_USE_COUNT;
        return count;
    }

    if ( stricmp(p, "enable routing") == 0 )
    {
        ifx_ppa_exit();
        ifx_ppa_init(NULL, 0);
        ifx_ppa_enable(1, 1, 0);

        MOD_DEC_USE_COUNT;
        return count;
    }
    else if ( stricmp(p, "enable bridging") == 0 )
    {
        ifx_ppa_exit();
        ifx_ppa_init(NULL, PPA_F_BRIDGE_ACCEL_MODE);
        ifx_ppa_enable(1, 1, 0);

        MOD_DEC_USE_COUNT;
        return count;
    }

    switch ( g_fwcode )
    {
    case FWCODE_ROUTING_ACC_D2:
        if ( stricmp(p, "enable") == 0 )
        {
            if ( ppa_hook_session_add_fn == NULL
                && ppa_hook_session_del_fn == NULL
                && ppa_hook_inactivity_status_fn == NULL
                && ppa_hook_set_inactivity_fn == NULL
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
                && ppa_hook_mc_session_add_fn == NULL
                && ppa_hook_mc_session_del_fn == NULL
#else
                && ppa_hook_mc_group_update_fn == NULL
                && ppa_hook_mc_group_get_fn == NULL
#endif
               )
            {
                ifx_ppa_exit();
                ifx_ppa_init(NULL, 0);
                ifx_ppa_enable(1, 1, 0);
            }
            else
            {
                ppa_hook_session_add_fn       = ifx_ppa_session_add;
                ppa_hook_session_del_fn       = ifx_ppa_session_delete;
                ppa_hook_inactivity_status_fn = ifx_ppa_inactivity_status;
                ppa_hook_set_inactivity_fn    = ifx_ppa_set_session_inactivity;
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
                ppa_hook_mc_session_add_fn    = ifx_ppa_mc_session_add;
                ppa_hook_mc_session_del_fn    = ifx_ppa_mc_session_delete;
#else
                ppa_hook_mc_group_update_fn   = ifx_ppa_mc_group_update;
                ppa_hook_mc_group_get_fn      = ifx_ppa_mc_group_get;
#endif
            }
        }
        else if ( stricmp(p, "enable add") == 0 )
            ppa_hook_session_add_fn       = ifx_ppa_session_add;
        else if ( stricmp(p, "disable add") == 0 )
            ppa_hook_session_add_fn       = NULL;
        else if ( stricmp(p, "enable del") == 0 )
            ppa_hook_session_del_fn       = ifx_ppa_session_delete;
        else if ( stricmp(p, "disable del") == 0 )
        {
            ppa_hook_session_del_fn       = NULL;
            flag_del = 1;
        }
#if 1
        //  don't remove entries
        else if ( stricmp(p, "del off") == 0 )
            ppa_hook_session_del_fn       = NULL;
#endif
        else if ( stricmp(p, "enable stat") == 0 )
            ppa_hook_inactivity_status_fn = ifx_ppa_inactivity_status;
        else if ( stricmp(p, "disable stat") == 0 )
            ppa_hook_inactivity_status_fn = NULL;
        else if ( stricmp(p, "enable timeout") == 0 )
            ppa_hook_set_inactivity_fn    = ifx_ppa_set_session_inactivity;
        else if ( stricmp(p, "disable timeout") == 0 )
            ppa_hook_set_inactivity_fn    = NULL;
        else
        {
            //ppa_hook_session_add_fn       = NULL;
            //ppa_hook_session_del_fn       = NULL;
            //ppa_hook_inactivity_status_fn = NULL;
            //ppa_hook_set_inactivity_fn    = NULL;
            //ppa_hook_mc_session_add_fn    = NULL;
            //ppa_hook_mc_session_del_fn    = NULL;

            //flag_del = 1;
            //flag_mc_del = 1;
            ifx_ppa_exit();
        }
        break;

    case FWCODE_BRIDGING_ACC_D3:
        if ( stricmp(p, "enable") == 0 )
        {
            if ( ppa_hook_bridge_entry_add_fn == NULL
                && ppa_hook_bridge_entry_delete_fn == NULL
                && ppa_hook_set_bridge_entry_timeout_fn == NULL
                && ppa_hook_bridge_entry_inactivity_status_fn == NULL
                && ppa_hook_bridge_entry_hit_time_fn == NULL )
            {
                ifx_ppa_exit();
                ifx_ppa_init(NULL, PPA_F_BRIDGE_ACCEL_MODE);
                ifx_ppa_enable(1, 1, 0);
            }
            else
            {
                ppa_hook_bridge_entry_add_fn               = ifx_ppa_bridge_entry_add;
                ppa_hook_bridge_entry_delete_fn            = ifx_ppa_bridge_entry_delete;
                ppa_hook_set_bridge_entry_timeout_fn       = ifx_ppa_set_bridge_entry_timeout;
                ppa_hook_bridge_entry_inactivity_status_fn = ifx_ppa_bridge_entry_inactivity_status;
                ppa_hook_bridge_entry_hit_time_fn          = ifx_ppa_bridge_entry_hit_time;
            }
        }
        else if ( stricmp(p, "enable add") == 0 )
            ppa_hook_bridge_entry_add_fn               = ifx_ppa_bridge_entry_add;
        else if ( stricmp(p, "disable add") == 0 )
            ppa_hook_bridge_entry_add_fn               = NULL;
        else if ( stricmp(p, "enable del") == 0 )
            ppa_hook_bridge_entry_delete_fn            = ifx_ppa_bridge_entry_delete;
        else if ( stricmp(p, "disable del") == 0 )
        {
            ppa_hook_bridge_entry_delete_fn            = NULL;
            flag_bridging_del = 1;
        }
#if 1
        //  don't remove entries
        else if ( stricmp(p, "del off") == 0 )
            ppa_hook_bridge_entry_delete_fn            = NULL;
#endif
        else if ( stricmp(p, "enable stat") == 0 )
            ppa_hook_bridge_entry_inactivity_status_fn = ifx_ppa_bridge_entry_inactivity_status;
        else if ( stricmp(p, "disable stat") == 0 )
            ppa_hook_bridge_entry_inactivity_status_fn = NULL;
        else if ( stricmp(p, "enable timeout") == 0 )
            ppa_hook_set_bridge_entry_timeout_fn       = ifx_ppa_set_bridge_entry_timeout;
        else if ( stricmp(p, "disable timeout") == 0 )
            ppa_hook_set_bridge_entry_timeout_fn       = NULL;
        else if ( stricmp(p, "enable hittime") == 0 )
            ppa_hook_bridge_entry_hit_time_fn          = ifx_ppa_bridge_entry_hit_time;
        else if ( stricmp(p, "disable hittime") == 0 )
            ppa_hook_bridge_entry_hit_time_fn          = NULL;
        else
            ifx_ppa_exit();
        break;

    case FWCODE_ROUTING_BRIDGING_ACC_D4:
    case FWCODE_ROUTING_BRIDGING_ACC_A4:    //  reuse D4 code
        if ( stricmp(p, "enable") == 0 || stricmp(p, "enable indirect") == 0 )
        {
            if ( ppa_hook_session_add_fn == NULL
                && ppa_hook_session_del_fn == NULL
                && ppa_hook_inactivity_status_fn == NULL
                && ppa_hook_set_inactivity_fn == NULL
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
                && ppa_hook_mc_session_add_fn == NULL
                && ppa_hook_mc_session_del_fn == NULL
#else
                && ppa_hook_mc_group_update_fn == NULL
                && ppa_hook_mc_group_get_fn == NULL
#endif
                && ppa_hook_bridge_entry_add_fn == NULL
                && ppa_hook_bridge_entry_delete_fn == NULL
                && ppa_hook_set_bridge_entry_timeout_fn == NULL
                && ppa_hook_bridge_entry_inactivity_status_fn == NULL
                && ppa_hook_bridge_entry_hit_time_fn == NULL
               )
            {
                ifx_ppa_exit();
                ifx_ppa_init(NULL, 0);
                if ( strlen(p) > 6 )
                    set_fast_mode(IFX_PPA_SET_FAST_MODE_CPU1_DIRECT | IFX_PPA_SET_FAST_MODE_ETH1_INDIRECT, IFX_PPA_SET_FAST_MODE_CPU1 | IFX_PPA_SET_FAST_MODE_ETH1);
                ifx_ppa_enable(1, 1, 0);
            }
            else
            {
                ppa_hook_session_add_fn       = ifx_ppa_session_add;
                ppa_hook_session_del_fn       = ifx_ppa_session_delete;
                ppa_hook_inactivity_status_fn = ifx_ppa_inactivity_status;
                ppa_hook_set_inactivity_fn    = ifx_ppa_set_session_inactivity;
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
                ppa_hook_mc_session_add_fn    = ifx_ppa_mc_session_add;
                ppa_hook_mc_session_del_fn    = ifx_ppa_mc_session_delete;
#else
                ppa_hook_mc_group_update_fn   = ifx_ppa_mc_group_update;
                ppa_hook_mc_group_get_fn      = ifx_ppa_mc_group_get;
#endif

                ppa_hook_bridge_entry_add_fn               = ifx_ppa_bridge_entry_add;
                ppa_hook_bridge_entry_delete_fn            = ifx_ppa_bridge_entry_delete;
                ppa_hook_set_bridge_entry_timeout_fn       = ifx_ppa_set_bridge_entry_timeout;
                ppa_hook_bridge_entry_inactivity_status_fn = ifx_ppa_bridge_entry_inactivity_status;
                ppa_hook_bridge_entry_hit_time_fn          = ifx_ppa_bridge_entry_hit_time;
            }
        }
        else if ( stricmp(p, "enable add") == 0 )
        {
            ppa_hook_session_add_fn       = ifx_ppa_session_add;
            ppa_hook_bridge_entry_add_fn               = ifx_ppa_bridge_entry_add;
        }
        else if ( stricmp(p, "disable add") == 0 )
        {
            ppa_hook_session_add_fn       = NULL;
            ppa_hook_bridge_entry_add_fn               = NULL;
        }
        else if ( stricmp(p, "enable del") == 0 )
        {
            ppa_hook_session_del_fn       = ifx_ppa_session_delete;
            ppa_hook_bridge_entry_delete_fn            = ifx_ppa_bridge_entry_delete;
        }
        else if ( stricmp(p, "disable del") == 0 )
        {
            ppa_hook_session_del_fn       = NULL;
            flag_del = 1;
            ppa_hook_bridge_entry_delete_fn            = NULL;
            flag_bridging_del = 1;
        }
#if 1
        //  don't remove entries
        else if ( stricmp(p, "del off") == 0 )
        {
            ppa_hook_session_del_fn       = NULL;
            ppa_hook_bridge_entry_delete_fn            = NULL;
        }
#endif
        else if ( stricmp(p, "enable stat") == 0 )
        {
            ppa_hook_inactivity_status_fn = ifx_ppa_inactivity_status;
            ppa_hook_bridge_entry_inactivity_status_fn = ifx_ppa_bridge_entry_inactivity_status;
        }
        else if ( stricmp(p, "disable stat") == 0 )
        {
            ppa_hook_inactivity_status_fn = NULL;
            ppa_hook_bridge_entry_inactivity_status_fn = NULL;
        }
        else if ( stricmp(p, "enable timeout") == 0 )
        {
            ppa_hook_set_inactivity_fn    = ifx_ppa_set_session_inactivity;
            ppa_hook_set_bridge_entry_timeout_fn       = ifx_ppa_set_bridge_entry_timeout;
        }
        else if ( stricmp(p, "disable timeout") == 0 )
        {
            ppa_hook_set_inactivity_fn    = NULL;
            ppa_hook_set_bridge_entry_timeout_fn       = NULL;
        }
        else if ( stricmp(p, "enable hittime") == 0 )
            ppa_hook_bridge_entry_hit_time_fn          = ifx_ppa_bridge_entry_hit_time;
        else if ( stricmp(p, "disable hittime") == 0 )
            ppa_hook_bridge_entry_hit_time_fn          = NULL;
        else
            ifx_ppa_exit();
        break;
    }

    MOD_DEC_USE_COUNT;

    if ( flag_del )
    {
        struct session_list_item *p_item, *p_next;
        int i;

        for ( i = 0; i < NUM_ENTITY(g_session_list_hash_table); i++ )
        {
            ppa_lock_get(&g_session_list_lock);
            p_item = g_session_list_hash_table[i];
            g_session_list_hash_table[i] = NULL;
            ppa_lock_release(&g_session_list_lock);

            for ( ; p_item; p_item = p_next )
            {
                p_next = p_item->next;

                ppa_hw_del_session(p_item);
                ppa_mem_cache_free(p_item, g_session_item_cache);
            }
        }
    }

#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    if ( flag_mc_del )
    {
        struct session_list_item *p_item, *p_next;
        int i;

        for ( i = 0; i < NUM_ENTITY(g_mc_session_list_hash_table); i++ )
        {
            ppa_lock_get(&g_mc_session_list_lock);
            p_item = g_mc_session_list_hash_table[i];
            g_mc_session_list_hash_table[i] = NULL;
            ppa_lock_release(&g_mc_session_list_lock);

            for ( ; p_item; p_item = p_next )
            {
                p_next = p_item->next;

                ppa_hw_del_mc_session(p_item);
                ppa_mem_cache_free(p_item, g_session_item_cache);
            }
        }
    }
#else
    if ( flag_mc_del )
    {
        struct mc_group_list_item *p_item, *p_next;
        int i;

        for ( i = 0; i < NUM_ENTITY(g_mc_group_list_hash_table); i++ )
        {
            ppa_lock_get(&g_mc_group_list_lock);
            p_item = g_mc_group_list_hash_table[i];
            g_mc_group_list_hash_table[i] = NULL;
            ppa_lock_release(&g_mc_group_list_lock);

            for ( ; p_item; p_item = p_next )
            {
                p_next = p_item->next;

                ppa_hw_del_mc_group(p_item);
                ppa_mem_cache_free(p_item, g_mc_group_item_cache);
            }
        }
    }
#endif

    if ( flag_bridging_del )
    {
        struct bridging_session_list_item *p_item, *p_next;
        int i;

        for ( i = 0; i < NUM_ENTITY(g_bridging_session_list_hash_table); i++ )
        {
            ppa_lock_get(&g_bridging_session_list_lock);
            p_item = g_bridging_session_list_hash_table[i];
            g_bridging_session_list_hash_table[i] = NULL;
            ppa_lock_release(&g_bridging_session_list_lock);

            for ( ; p_item; p_item = p_next )
            {
                p_next = p_item->next;

                ppa_bridging_hw_del_session(p_item);
                ppa_mem_cache_free(p_item, g_bridging_session_item_cache);
            }
        }
    }

    return count;
}

static int proc_read_netif(char *page, char **start, off_t off, int count, int *eof, void *data)
{
    int len = 0;
    int len_max = off + count;
    char *pstr;
    char str[2048];
    int llen;
    int no;
    struct netif_info *p_cur;

    MOD_INC_USE_COUNT;

    pstr = *start = page;

    llen = sprintf(str, "NetIF list\n");
    if ( len <= off && len + llen > off )
    {
        memcpy(pstr, str + off - len, len + llen - off);
        pstr += len + llen - off;
    }
    else if ( len > off )
    {
        memcpy(pstr, str, llen);
        pstr += llen;
    }
    len += llen;

    no = 0;

    for ( p_cur = g_netif; p_cur; p_cur = p_cur->next )
    {
        llen = print_netif(str, ++no, p_cur);
        if ( len <= off && len + llen > off )
        {
            memcpy(pstr, str + off - len, len + llen - off);
            pstr += len + llen - off;
        }
        else if ( len > off )
        {
            memcpy(pstr, str, llen);
            pstr += llen;
        }
        len += llen;
        if ( len >= len_max )
            goto PROC_READ_NETIF_OVERRUN_END;
    }

    *eof = 1;

    MOD_DEC_USE_COUNT;

    return len - off;

PROC_READ_NETIF_OVERRUN_END:

    MOD_DEC_USE_COUNT;

    return len - llen - off;
}

static int proc_write_netif(struct file *file, const char *buf, unsigned long count, void *data)
{
    char str[64];
    char *p;

    int len, rlen;

    PPA_IFINFO ifinfo;
    int f_is_lanif = -1;

    MOD_INC_USE_COUNT;

    len = count < sizeof(str) ? count : sizeof(str) - 1;
    rlen = len - copy_from_user(str, buf, len);
    while ( rlen && str[rlen - 1] <= ' ' )
        rlen--;
    str[rlen] = 0;
    for ( p = str; *p && *p <= ' '; p++, rlen-- );
    if ( !*p )
    {
        MOD_DEC_USE_COUNT;
        return 0;
    }

    if ( strincmp(p, "add ", 4) == 0 )
    {
        p += 4;

        while ( *p && *p <= ' ' )
            p++;

        if ( strincmp(p, "lan ", 4) == 0 )
            f_is_lanif = 1;
        else if ( strincmp(p, "wan ",4) == 0 )
            f_is_lanif = 0;

        if ( f_is_lanif >= 0 )
        {
            p += 4;

            while ( *p && *p <= ' ' )
                p++;

            ifinfo.ifname = p;
            ifinfo.if_flags = 0;
            ifx_ppa_add_itf(&ifinfo, f_is_lanif);
        }
    }
    
    MOD_DEC_USE_COUNT;

    return count;
}

#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
static int proc_read_directpath(char *page, char **start, off_t off, int count, int *eof, void *data)
                {
    int len = 0;

    MOD_INC_USE_COUNT;

    if (g_directpath_enable)
        len += sprintf(page + off + len, "directpath enabled\n");
                else
        len += sprintf(page + off + len, "directpath disabled\n");

    MOD_DEC_USE_COUNT;

    *eof = 1;

    return len;
            }

static int proc_write_directpath(struct file *file, const char *buf, unsigned long count, void *data)
            {
    char str[64];
    char *p;

    int len, rlen; 
    unsigned int flags = 0;

    MOD_INC_USE_COUNT;

    len = count < sizeof(str) ? count : sizeof(str) - 1;
    rlen = len - copy_from_user(str, buf, len);
    while ( rlen && str[rlen - 1] <= ' ' )
        rlen--;
    str[rlen] = 0;
    for ( p = str; *p && *p <= ' '; p++, rlen-- );
    if ( !*p )
    {
        MOD_DEC_USE_COUNT;
        return 0;
            }

    if ( stricmp(str, "enable") == 0 ) {
	ifx_ppa_directpath_init(flags);
	g_directpath_enable = 1;
        }
    else {
	ifx_ppa_directpath_exit(flags);
	g_directpath_enable = 0;
    }

    MOD_DEC_USE_COUNT;

    return count;
}
#endif

static INLINE int print_session(char *buf, int no, struct session_list_item *p_item)
{
    static const char *str_flag[] = {
        "IS_REPLY",         //  0x00000001
        "BRIDGING_VCI_CHECK",
        "INVALID",
        "INVALID",
        "ADDED_IN_HW",      //  0x00000010
        "CAN_NOT_ACCEL",
        "STATIC",
        "DROP",
        "VALID_NAT_IP",     //  0x00000100
        "VALID_NAT_PORT",
        "VALID_NAT_SNAT",
        "INVALID",
        "VALID_VLAN_INS",   //  0x00001000
        "VALID_VLAN_RM",
        "SESSION_VALID_OUT_VLAN_INS",
        "SESSION_VALID_OUT_VLAN_RM",
        "VALID_PPPOE",      //  0x00010000
        "VALID_NEW_SRC_MAC",
        "VALID_MTU",
        "VALID_NEW_DSCP",
        "LAN_ENTRY",        //  0x00100000
        "WAN_ENTRY",
        "INVALID",
        "INVALID",
        "SRC_MAC_DROP_EN",  //  0x01000000
        "TX_ITF_IPOA",
        "TX_ITF_PPPOA",
        "INVALID",
        "INVALID",          //  0x10000000
        "INVALID",
        "INVALID",
        "INVALID"
    };
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    static const char *str_dest[] = {
        "ETH0",
        "ETH1",
        "CPU0",
        "EXT_INT1",
        "EXT_INT2",
        "EXT_INT3",
        "EXT_INT4",
        "EXT_INT5"
    };
#endif

    int len = 0;
    int flag;
    unsigned long bit;
    int i;

    len += sprintf(buf + len,     "  no. %d\n", no);
    len += sprintf(buf + len,     "    next             = %08X\n", (uint32_t)p_item->next);
    len += sprintf(buf + len,     "    session          = %08X\n", (uint32_t)p_item->session);
    len += sprintf(buf + len,     "    session_extra:\n");
    len += sprintf(buf + len,     "      new_dscp       = %d\n", (uint32_t)p_item->session_extra.new_dscp);
    len += sprintf(buf + len,     "      dscp_remark    = %d\n", (uint32_t)p_item->session_extra.dscp_remark);
    len += sprintf(buf + len,     "      vlan_insert    = %d\n", (uint32_t)p_item->session_extra.vlan_insert);
    len += sprintf(buf + len,     "      vlan_remove    = %d\n", (uint32_t)p_item->session_extra.vlan_remove);
    len += sprintf(buf + len,     "      vlan_prio      = %d\n", (uint32_t)p_item->session_extra.vlan_prio);
    len += sprintf(buf + len,     "      vlan_cfi       = %d\n", (uint32_t)p_item->session_extra.vlan_cfi);
    len += sprintf(buf + len,     "      vlan_id        = %08X\n", (uint32_t)p_item->session_extra.vlan_id);
    len += sprintf(buf + len,     "      mtu            = %d\n", (uint32_t)p_item->session_extra.mtu);
    len += sprintf(buf + len,     "      session_flags  = %08X\n",  (uint32_t)p_item->session_extra.session_flags);
    len += sprintf(buf + len,     "    ip_proto         = %d\n", p_item->ip_proto);
    len += sprintf(buf + len,     "    src_ip           = %d.%d.%d.%d\n", p_item->src_ip >> 24, (p_item->src_ip >> 16) & 0xFF, (p_item->src_ip >> 8) & 0xFF, p_item->src_ip & 0xFF);
    len += sprintf(buf + len,     "    src_port         = %d\n",   p_item->src_port);
    len += sprintf(buf + len,     "    src_mac[6]       = %02x:%02x:%02x:%02x:%02x:%02x\n", p_item->src_mac[0], p_item->src_mac[1], p_item->src_mac[2], p_item->src_mac[3], p_item->src_mac[4], p_item->src_mac[5]);
    len += sprintf(buf + len,     "    dst_ip           = %d.%d.%d.%d\n", p_item->dst_ip >> 24, (p_item->dst_ip >> 16) & 0xFF, (p_item->dst_ip >> 8) & 0xFF, p_item->dst_ip & 0xFF);
    len += sprintf(buf + len,     "    dst_port         = %d\n",   p_item->dst_port);
    len += sprintf(buf + len,     "    dst_mac[6]       = %02x:%02x:%02x:%02x:%02x:%02x\n", p_item->dst_mac[0], p_item->dst_mac[1], p_item->dst_mac[2], p_item->dst_mac[3], p_item->dst_mac[4], p_item->dst_mac[5]);
    len += sprintf(buf + len,     "    nat_ip           = %d.%d.%d.%d\n", p_item->nat_ip >> 24, (p_item->nat_ip >> 16) & 0xFF, (p_item->nat_ip >> 8) & 0xFF, p_item->nat_ip & 0xFF);
    len += sprintf(buf + len,     "    nat_port         = %d\n",   p_item->nat_port);
    len += sprintf(buf + len,     "    rx_if            = %08X\n", (uint32_t)p_item->rx_if);
    len += sprintf(buf + len,     "    tx_if            = %08X\n", (uint32_t)p_item->tx_if);
    len += sprintf(buf + len,     "    timeout          = %d\n",   p_item->timeout);
    len += sprintf(buf + len,     "    last_hit_time    = %d\n",   p_item->last_hit_time);
    len += sprintf(buf + len,     "    num_adds         = %d\n",   p_item->num_adds);
    len += sprintf(buf + len,     "    pppoe_session_id = %d\n",   p_item->pppoe_session_id);
    len += sprintf(buf + len,     "    new_dscp         = %d\n",   p_item->new_dscp);
    len += sprintf(buf + len,     "    new_vci          = %08X\n", p_item->new_vci);
    len += sprintf(buf + len,     "    mtu              = %d\n",   p_item->mtu);

#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    len += sprintf(buf + len,     "    mc_dest_list     = ");
    flag = 0;
    for ( bit = 1, i = 0; i < NUM_ENTITY(str_dest); bit <<= 1, i++ )
        if ( (p_item->mc_dest_list & bit) )
        {
            if ( flag++ )
                len += sprintf(buf + len, "| ");
            len += sprintf(buf + len, str_dest[i]);
            len += sprintf(buf + len, " ");
        }
    if ( flag )
        len += sprintf(buf + len, "\n");
    else
        len += sprintf(buf + len, "NULL\n");
#endif

    len += sprintf(buf + len,     "    flags            = ");
    flag = 0;
    for ( bit = 1, i = 0; bit; bit <<= 1, i++ )
        if ( (p_item->flags & bit) )
        {
            if ( flag++ )
                len += sprintf(buf + len, "| ");
            len += sprintf(buf + len, str_flag[i]);
            len += sprintf(buf + len, " ");
        }
    if ( flag )
        len += sprintf(buf + len, "\n");
    else
        len += sprintf(buf + len, "NULL\n");

    if ( p_item->routing_entry == ~0 )
        len += sprintf(buf + len, "    routing_entry    = N/A\n");
    else
        len += sprintf(buf + len, "    routing_entry    = %d\n", p_item->routing_entry);
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    if ( p_item->mc_routing_entry == ~0 )
        len += sprintf(buf + len, "    mc_routing_entry = N/A\n");
    else
        len += sprintf(buf + len, "    mc_routing_entry = %d\n", p_item->mc_routing_entry);
#endif
    if ( p_item->pppoe_entry == ~0 )
        len += sprintf(buf + len, "    pppoe_entry      = N/A\n");
    else
        len += sprintf(buf + len, "    pppoe_entry      = %d\n", p_item->pppoe_entry);
    if ( p_item->mtu_entry == ~0 )
        len += sprintf(buf + len, "    mtu_entry        = N/A\n");
    else
        len += sprintf(buf + len, "    mtu_entry        = %d\n", p_item->mtu_entry);
    if ( p_item->src_mac_entry == ~0 )
        len += sprintf(buf + len, "    src_mac_entry    = N/A\n");
    else
        len += sprintf(buf + len, "    src_mac_entry    = %d\n", p_item->src_mac_entry);

#ifdef CONFIG_IFX_PPA_A4
    len += sprintf(buf + len,     "    dslwan_qid       = %d\n",   p_item->dslwan_qid);
    len += sprintf(buf + len,     "    out_vlan_tag     = 0x%08X\n", p_item->out_vlan_tag);
    if ( p_item->out_vlan_ix == ~0 )
        len += sprintf(buf + len, "    out_vlan_ix      = N/A\n");
    else
        len += sprintf(buf + len, "    out_vlan_ix      = %d\n",   p_item->out_vlan_ix);
#endif

#if defined(ENABLE_SESSION_DEBUG_FLAGS) && ENABLE_SESSION_DEBUG_FLAGS
    len += print_debug_flags(buf + len, "    debug_flags      = %s\n", p_item->debug_flags);
#endif

    return len;
}

static INLINE int print_bridging_session(char *buf, int no, struct bridging_session_list_item *p_item)
{
    static const char *str_flag[] = {
        "IS_REPLY",         //  0x00000001
        "BRIDGING_VCI_CHECK",
        "INVALID",
        "INVALID",
        "ADDED_IN_HW",      //  0x00000010
        "CAN_NOT_ACCEL",
        "STATIC",
        "DROP",
        "VALID_NAT_IP",     //  0x00000100
        "VALID_NAT_PORT",
        "VALID_NAT_SNAT",
        "INVALID",
        "VALID_VLAN_INS",   //  0x00001000
        "VALID_VLAN_RM",
        "SESSION_VALID_OUT_VLAN_INS",
        "SESSION_VALID_OUT_VLAN_RM",
        "VALID_PPPOE",      //  0x00010000
        "VALID_NEW_SRC_MAC",
        "VALID_MTU",
        "VALID_NEW_DSCP",
        "LAN_ENTRY",        //  0x00100000
        "WAN_ENTRY",
        "INVALID",
        "INVALID",
        "SRC_MAC_DROP_EN",  //  0x01000000
        "INVALID",
        "INVALID",
        "INVALID",
        "INVALID",          //  0x10000000
        "INVALID",
        "INVALID",
        "INVALID"
    };

    int len = 0;
    int flag;
    unsigned long bit;
    int i;

    len += sprintf(buf + len,     "  no. %d\n", no);
    len += sprintf(buf + len,     "    next             = %08X\n", (uint32_t)p_item->next);
    len += sprintf(buf + len,     "    mac[6]           = %02x:%02x:%02x:%02x:%02x:%02x\n", p_item->mac[0], p_item->mac[1], p_item->mac[2], p_item->mac[3], p_item->mac[4], p_item->mac[5]);
    len += sprintf(buf + len,     "    netif            = %08X\n", (uint32_t)p_item->netif);
    len += sprintf(buf + len,     "    timeout          = %d\n",   p_item->timeout);
    len += sprintf(buf + len,     "    last_hit_time    = %d\n",   p_item->last_hit_time);

    len += sprintf(buf + len,     "    flags            = ");
    flag = 0;
    for ( bit = 1, i = 0; bit; bit <<= 1, i++ )
        if ( (p_item->flags & bit) )
        {
            if ( flag++ )
                len += sprintf(buf + len, "| ");
            len += sprintf(buf + len, str_flag[i]);
            len += sprintf(buf + len, " ");
        }
    if ( flag )
        len += sprintf(buf + len, "\n");
    else
        len += sprintf(buf + len, "NULL\n");

    if ( p_item->bridging_entry == ~0 )
        len += sprintf(buf + len, "    bridging_entry   = N/A\n");
    else
        len += sprintf(buf + len, "    bridging_entry   = %08X\n", p_item->bridging_entry);

#ifdef CONFIG_IFX_PPA_A4
    len += sprintf(buf + len,     "    dslwan_qid       = %d\n",   p_item->dslwan_qid);
#endif

#if defined(ENABLE_SESSION_DEBUG_FLAGS) && ENABLE_SESSION_DEBUG_FLAGS
    len += print_debug_flags(buf + len, "    debug_flags      = %s\n", p_item->debug_flags);
#endif

    return len;
}

static INLINE int print_netif(char *buf, int no, struct netif_info *p_item)
{
    static const char *str_flag[] = {
        "NETIF_VLAN",               //  0x00000001
        "NETIF_BRIDGE",
        "NETIF_PPPOE",
        "NETIF_BR2684",
        "INVALID",                  //  0x00000010
        "INVALID",
        "INVALID",
        "INVALID",
        "NETIF_LAN_IF",             //  0x00000100
        "NETIF_WAN_IF",
        "INVALID",
        "NETIF_PHY_IF_GOT",
        "NETIF_MAC_ENTRY_CREATED",  //  0x00001000
        "INVALID",
        "INVALID",
        "INVALID",
    };

    int len = 0;
    int flag;
    unsigned long bit;
    int i;

    len += sprintf(buf + len,     "  no. %d\n", no);
    len += sprintf(buf + len,     "    next             = %08X\n", (uint32_t)p_item->next);
    len += sprintf(buf + len,     "    name             = %s\n", p_item->name);
    if ( p_item->phys_netif_name[0] )
        len += sprintf(buf + len, "    phys_netif_name  = %s\n", p_item->phys_netif_name);
    else
        len += sprintf(buf + len, "    phys_netif_name  = NULL\n");
    len += sprintf(buf + len,     "    mac              = %02x:%02x:%02x:%02x:%02x:%02x\n", (u32)p_item->mac[0], (u32)p_item->mac[1], (u32)p_item->mac[2], (u32)p_item->mac[3], (u32)p_item->mac[4], (u32)p_item->mac[5]);
#if CONFIG_IFX_PPA_A4
    if ( (p_item->flags & NETIF_BR2684) )
        len += sprintf(buf + len, "    dslwan_qid       = %d\n", p_item->dslwan_qid);
#endif

    len += sprintf(buf + len,     "    flags            = ");
    flag = 0;
    for ( bit = 1, i = 0; i < sizeof(str_flag) / sizeof(*str_flag); bit <<= 1, i++ )
        if ( (p_item->flags & bit) )
        {
            if ( flag++ )
                len += sprintf(buf + len, "| ");
            len += sprintf(buf + len, str_flag[i]);
            len += sprintf(buf + len, " ");
        }
    if ( flag )
        len += sprintf(buf + len, "\n");
    else
        len += sprintf(buf + len, "NULL\n");

    if ( p_item->mac_entry == ~0 )
        len += sprintf(buf + len, "    mac_entry        = N/A\n");
    else
        len += sprintf(buf + len, "    mac_entry        = %08X\n", p_item->mac_entry);

    return len;
}

#if defined(ENABLE_SESSION_DEBUG_FLAGS) && ENABLE_SESSION_DEBUG_FLAGS
static INLINE int print_debug_flags(char *buf, const char *format, uint32_t flags)
{
    static const char *str_flag[] = {
        "NOT_REACH_MIN_HITS",           //  0x00000001
        "ALG_OR_WRONG_PROTO",
        "ZERO_DST_MAC",
        "INVALID",
        "RX_IF_TYPE_ERR",               //  0x00000010
        "TX_IF_TYPE_ERR",
        "INVALID",
        "INVALID",
        "DST_BRG_IF_NOT_IN_BRG_TBL",    //  0x00000100
        "DST_IF_NOT_IN_IF_LIST",
        "GET_DST_MAC_FAIL",
        "INVALID",
        "ADD_PPPOE_ENTRY_FAIL",         //  0x00001000
        "ADD_MTU_ENTRY_FAIL",
        "ADD_MAC_ENTRY_FAIL",
        "ADD_OUT_VLAN_ENTRY_FAIL",
        "RX_PPPOE",                     //  0x00010000
        "TX_PPPOE",
        "TX_BR2684_EOA",
        "TX_BR2684_IPOA",
        "TX_PPPOA",                     //  0x00100000
        "RX_VLAN",
        "TX_VLAN",
        "INVALID",
  #if 0 //  reserved
        "INVALID",                      //  0x01000000
        "INVALID",
        "INVALID",
        "INVALID",
        "INVALID",                      //  0x10000000
        "INVALID",
        "INVALID",
        "INVALID",
  #endif
    };
    
    int len = 0;
    char tmpbuf[512];
    int flag;
    unsigned long bit;
    int i;
    
    flag = 0;
    for ( bit = 1, i = 0; i < sizeof(str_flag) / sizeof(*str_flag); bit <<= 1, i++ )
        if ( (flags & bit) )
        {
            if ( flag++ )
                len += sprintf(tmpbuf + len, "| ");
            len += sprintf(tmpbuf + len, str_flag[i]);
            len += sprintf(tmpbuf + len, " ");
        }
    if ( flag )
        len += sprintf(tmpbuf + len, "\n");
    else
        len += sprintf(tmpbuf + len, "NULL\n");
        
    return sprintf(buf, format, tmpbuf);
}
#endif

static INLINE int stricmp(const char *p1, const char *p2)
{
    int c1, c2;

    while ( *p1 && *p2 )
    {
        c1 = *p1 >= 'A' && *p1 <= 'Z' ? *p1 + 'a' - 'A' : *p1;
        c2 = *p2 >= 'A' && *p2 <= 'Z' ? *p2 + 'a' - 'A' : *p2;
        if ( (c1 -= c2) )
            return c1;
        p1++;
        p2++;
    }

    return *p1 - *p2;
}

static INLINE int strincmp(const char *p1, const char *p2, int n)
{
    int c1 = 0, c2;

    while ( n && *p1 && *p2 )
    {
        c1 = *p1 >= 'A' && *p1 <= 'Z' ? *p1 + 'a' - 'A' : *p1;
        c2 = *p2 >= 'A' && *p2 <= 'Z' ? *p2 + 'a' - 'A' : *p2;
        if ( (c1 -= c2) )
            return c1;
        p1++;
        p2++;
        n--;
    }

    return n ? *p1 - *p2 : c1;
}



/*
 * ####################################
 *           Global Function
 * ####################################
 */

int32_t ifx_ppa_init(PPA_INIT_INFO *p_info, uint32_t flags)
{
#ifdef CONFIG_IFX_PPA_A4
    if ( g_init && g_fwcode == FWCODE_ROUTING_BRIDGING_ACC_A4 )
        return 0;
    ifx_ppa_exit();
    g_fwcode = FWCODE_ROUTING_BRIDGING_ACC_A4;
    if ( danube_eth_download_firmware(g_fwcode) )
        return -1;
    return ppa_d4_init(p_info, flags);
#else
    if ( g_init && g_fwcode == FWCODE_ROUTING_BRIDGING_ACC_D4 )
        return 0;
    ifx_ppa_exit();
    g_fwcode = FWCODE_ROUTING_BRIDGING_ACC_D4;
    if ( danube_eth_download_firmware(g_fwcode) )
        return -1;
    return ppa_d4_init(p_info, flags);
#endif
}

void ifx_ppa_exit(void)
{
    if ( g_init )
        switch ( g_fwcode )
        {
        case FWCODE_ROUTING_BRIDGING_ACC_D4:
            ppa_d4_exit();
            break;
        case FWCODE_ROUTING_BRIDGING_ACC_A4:
            ppa_d4_exit();  //  reuse D4 function
            break;
        }
}

//  lan_rx_ppa_enable -> eth0
//  wan_rx_ppa_enable -> eth1
int32_t ifx_ppa_enable(uint32_t lan_rx_ppa_enable, uint32_t wan_rx_ppa_enable, uint32_t flags)
{
    u32 sys_flag;

    if ( g_init )
        switch ( g_fwcode )
        {
        case FWCODE_ROUTING_BRIDGING_ACC_D4:
        case FWCODE_ROUTING_BRIDGING_ACC_A4:    //  reuse D4 code
            lan_rx_ppa_enable = lan_rx_ppa_enable ? IFX_PPA_ACC_MODE_ROUTING : IFX_PPA_ACC_MODE_NONE;
            wan_rx_ppa_enable = wan_rx_ppa_enable ? IFX_PPA_ACC_MODE_ROUTING : IFX_PPA_ACC_MODE_NONE;
            local_irq_save(sys_flag);
            set_acc_mode(1, lan_rx_ppa_enable);
            set_acc_mode(0, wan_rx_ppa_enable);
            local_irq_restore(sys_flag);
            return 0;
        }

    return -1;
}

int32_t ifx_ppa_get_status(uint32_t *lan_rx_ppa_enable, uint32_t *wan_rx_ppa_enable)
{
    get_acc_mode(1, lan_rx_ppa_enable);
    get_acc_mode(0, wan_rx_ppa_enable);

    return 0;
}

#ifdef CONFIG_IFX_PPA_D4

int32_t ifx_ppe_directpath_register_dev(uint32_t if_id, PPA_NETIF *netif, uint32_t flags)
{
    if ( if_id < IFX_PPA_IF_ID_CPU0_EXT_IF0 || if_id > IFX_PPA_IF_ID_CPU1_EXT_IF2 )
    {
        dbg("error if_id (%d)", if_id);
        return -1;
    }

    if ( (flags & PPA_F_DIRECTPATH_REGISTER) )
    {
        if ( device_push_dev[if_id] )
        {
            dbg("register device for %d which is not NULL (%08x)", if_id, (uint32_t)device_push_dev[if_id]);
            return -1;
        }
        else
            device_push_dev[if_id] = (uint32_t)netif;
    }
    else
        device_push_dev[if_id] = (u32)NULL;

    return 0;
}

int32_t ifx_ppe_directpath_register_rx_callback(uint32_t if_id, PPE_FP_RX_FN rx_fn, uint32_t flags)
{
    if ( if_id < IFX_PPA_IF_ID_CPU0_EXT_IF0 || if_id > IFX_PPA_IF_ID_CPU1_EXT_IF2 )
    {
        dbg("error if_id (%d)", if_id);
        return -1;
    }

    if ( (flags & PPA_F_DIRECTPATH_REGISTER) )
    {
        if ( device_push[if_id] )
        {
            dbg("register rx callback for %d which is not NULL (%08x)", if_id, (uint32_t)device_push[if_id]);
            return -1;
        }
        else
            device_push[if_id] = rx_fn;
    }
    else
        device_push[if_id] = NULL;

    return 0;
}

#endif


int32_t ifx_ppe_directpath_send(PPA_NETIF *netif, PPA_BUF *ppa_buf, uint32_t flags)
{
    err("not implemented");
    return -1;
}

int32_t ifx_ppa_session_add(PPA_BUF *ppa_buf, PPA_SESSION *p_session, uint32_t flags)
{
    struct session_list_item *p_item;

    /* XXX: Remove packets output by the device */
#if 0
    if ( ppa_buf->pkt_type == PACKET_OUTGOING )
    {
        dbg("sent out packet (pkt_type == PACKET_OUTGOING)");
        return IFX_PPA_SESSION_NOT_ADDED;
    }
    if ( ppa_get_pkt_src_if(ppa_buf) == NULL )
    {
        dbg("sent out packet (src_if == NULL)");
        return IFX_PPA_SESSION_NOT_ADDED;
    }
#else
    if ( ppa_is_pkt_host_output(ppa_buf) )
    {
        dbg("ppa_is_pkt_host_output");
        return IFX_PPA_SESSION_NOT_ADDED;
    }
#endif

    /* XXX: Remove broadcast and multicast packet */
#if 0
    if ( ppa_buf->pkt_type == PACKET_BROADCAST )
    {
        dbg("sent out packet (pkt_type == PACKET_BROADCAST)");
        return IFX_PPA_SESSION_NOT_ADDED;
    }
    if ( ppa_buf->pkt_type == PACKET_MULTICAST )
    {
        dbg("sent out packet (pkt_type == PACKET_MULTICAST)");
       return IFX_PPA_SESSION_NOT_ADDED;
    }
    //  XXX: ignore broadcast
    if ( (ppa_get_pkt_dst_ip(ppa_buf) & htonl(0x000000FF)) == htonl(0x000000FF) )
    {
        dbg("mac broad cast");
        return IFX_PPA_SESSION_NOT_ADDED;
    }
#else
    if ( ppa_is_pkt_broadcast(ppa_buf) || ppa_is_pkt_multicast(ppa_buf) )
    {
        dbg("%s", ppa_is_pkt_broadcast(ppa_buf) ? "ppa_is_pkt_broadcast" : "ppa_is_pkt_multicast");
        return IFX_PPA_SESSION_NOT_ADDED;
    }
#endif

    /* XXX: Remove loopback packet */
#if 0
    if ( ppa_buf->pkt_type == PACKET_LOOPBACK )
    {
        dbg("sent out packet (pkt_type == PACKET_LOOPBACK)");
        return IFX_PPA_SESSION_NOT_ADDED;
    }
#else
    if ( ppa_is_pkt_loopback(ppa_buf) )
    {
        dbg("ppa_is_pkt_loopback");
        return IFX_PPA_SESSION_NOT_ADDED;
    }
#endif

    //  get session
    if ( !p_session )
    {
        p_session = ppa_get_session(ppa_buf);
        if ( !p_session )
        {
            dbg("no conntrack");
            return IFX_PPA_SESSION_NOT_ADDED;
        }
    }

    //  ignore protocol other than TCP/UDP, since ICMP packet can't be handled safe in this arch
    if ( ppa_get_pkt_ip_proto(ppa_buf) != IPPROTO_UDP && ppa_get_pkt_ip_proto(ppa_buf) != IPPROTO_TCP )
    {
        dbg("protocol: %d", (uint32_t)ppa_get_pkt_ip_proto(ppa_buf));
        return IFX_PPA_SESSION_NOT_ADDED;
    }

    //  ignore special broadcast type
    if ( ppa_get_pkt_src_port(ppa_buf) == 137 || ppa_get_pkt_dst_port(ppa_buf) == 137
        || ppa_get_pkt_src_port(ppa_buf) == 138 || ppa_get_pkt_dst_port(ppa_buf) == 138 )
    {
        dbg("src port = %d, dst port = %d", ppa_get_pkt_src_port(ppa_buf), ppa_get_pkt_dst_port(ppa_buf));
        return IFX_PPA_SESSION_NOT_ADDED;
    }

#if 0
    //  prevent packet without port such as ICMP, can be removed later if protocol
    //  check approved to be working
    if ( ppa_get_pkt_src_port(ppa_buf) == 0 || ppa_get_pkt_dst_port(ppa_buf) == 0 )
    {
        uint32_t src_ip, dst_ip, src_port, dst_port;

        src_ip = ppa_get_pkt_src_ip(ppa_buf);
        dst_ip = ppa_get_pkt_dst_ip(ppa_buf);
        src_port = ppa_get_pkt_src_port(ppa_buf);
        dst_port = ppa_get_pkt_dst_port(ppa_buf);
        dbg("port error: src %d.%d.%d.%d:%d -> dst %d.%d.%d.%d:%d",
            src_ip >> 24,
            (src_ip >> 16) & 0xFF,
            (src_ip >> 8) & 0xFF,
            src_ip & 0xFF,
            src_port,
            dst_ip >> 24,
            (dst_ip >> 16) & 0xFF,
            (dst_ip >> 8) & 0xFF,
            dst_ip & 0xFF,
            dst_port);
        return IFX_PPA_SESSION_NOT_ADDED;
    }
#endif

    //  XXX: ignore fragment packet
    if ( ppa_is_pkt_fragment(ppa_buf) )
    {
        dbg("fragment");
        return IFX_PPA_SESSION_NOT_ADDED;
    }

    if ( (flags & PPA_F_BEFORE_NAT_TRANSFORM) )
    {
        if ( ppa_lookup_session(p_session, flags & PPA_F_SESSION_REPLY_DIR, &p_item) == IFX_PPA_SESSION_EXISTS )
        {
            dbg("lookup exist");
            return IFX_PPA_SESSION_ADDED;
        }
        else if ( ppa_add_session(ppa_buf, p_session, &p_item, flags) != 0 )
        {
            dbg("ppa_add_session failed");
            return IFX_PPA_SESSION_NOT_ADDED;
        }

        //dump_list_item(p_item, "after ppa_add_session");

        return IFX_PPA_SESSION_ADDED;
    }
    else
    {
        //  don't handle local packet or multicast packet
        if ( !ppa_is_pkt_routing(ppa_buf) )
        {
            dbg("not routing");
            return IFX_PPA_SESSION_NOT_ADDED;
        }

        if ( ppa_lookup_session(p_session, flags & PPA_F_SESSION_REPLY_DIR, &p_item) != IFX_PPA_SESSION_EXISTS )
        {
            dbg("post routing point but no session exist: dst_ip = %d.%d.%d.%d", ppa_get_pkt_dst_ip(ppa_buf) >> 24, (ppa_get_pkt_dst_ip(ppa_buf) >> 16) & 0xFF, (ppa_get_pkt_dst_ip(ppa_buf) >> 8) & 0xFF, ppa_get_pkt_dst_ip(ppa_buf) & 0xFF);
            return IFX_PPA_SESSION_NOT_ADDED;
        }

        //printk("post-routing hook: %d.%d.%d.%d:%d -> %d.%d.%d.%d:%d\n",
        //                           p_item->src_ip >> 24, (p_item->src_ip >> 16) & 0xFF, (p_item->src_ip >> 8) & 0xFF, p_item->src_ip & 0xFF, p_item->src_port,
        //                           p_item->dst_ip >> 24, (p_item->dst_ip >> 16) & 0xFF, (p_item->dst_ip >> 8) & 0xFF, p_item->dst_ip & 0xFF, p_item->dst_port);

        //  not enough hit
        if ( p_item->num_adds < MIN_HITS )
        {
            dbg("p_item->num_adds (%d) < MIN_HITS (%d)", p_item->num_adds, MIN_HITS);
            SET_DBG_FLAG(p_item, SESSION_DBG_NOT_REACH_MIN_HITS);
            return IFX_PPA_SESSION_NOT_ADDED;
        }

        //  can not be accelerated
        if ( (p_item->flags & SESSION_CAN_NOT_ACCEL) )
        {
            dbg("can not accelerate (pretested)");
            return IFX_PPA_SESSION_NOT_ADDED;
        }

        //  added in hardware
        if ( (p_item->flags & SESSION_ADDED_IN_HW) )
        {
            dbg("session added in firmware table already");
            return IFX_PPA_SESSION_EXISTS;
        }

        //  can not accelerate
        if ( !ppa_check_if_session_can_accel(ppa_buf, p_item->session, p_item) )
        {
            dbg("can not accelerate: %08X", (uint32_t)p_session);
            p_item->flags |= SESSION_CAN_NOT_ACCEL;
            SET_DBG_FLAG(p_item, SESSION_DBG_ALG_OR_WRONG_PROTO);
            return IFX_PPA_SESSION_NOT_ADDED;
        }

#if 0
        //  UDP or TCP replied
//        if ( ppa_get_pkt_ip_proto(ppa_buf) != IPPROTO_UDP && !(ppa_get_pkt_ip_proto(ppa_buf) == IPPROTO_TCP && (PPA_GET_TCP_CONNECT_STAUS(ppa_buf) & IPS_ASSURED)) )
//            return IFX_PPA_SESSION_NOT_ADDED;

        if ( ppa_get_pkt_proto(ppa_buf) == IPPROTO_TCP )
        {
//            if ( !(PPA_GET_TCP_CONNECT_STAUS(ppa_buf) & IPS_ASSURED_BIT) )
            if ( !(PPA_GET_TCP_CONNECT_STAUS(ppa_buf) & IPS_ASSURED) )
            {
                dbg("PPA_GET_TCP_CONNECT_STAUS(ppa_buf) with no flag - IPS_ASSURED");
                return IFX_PPA_SESSION_NOT_ADDED;
            }
        }
#endif

        if ( ppa_update_session(ppa_buf, p_item, flags) != 0 )
        {
            dbg("update session fail");
            return IFX_PPA_SESSION_NOT_ADDED;
        }

        dump_list_item(p_item, "after ppa_update_session");

        //  protect before ARP
        if ( !(p_item->flags & SESSION_TX_ITF_IPOA_PPPOA_MASK) && *(uint32_t *)p_item->dst_mac == 0 && *((uint16_t *)p_item->dst_mac + 2) == 0 )
        {
            SET_DBG_FLAG(p_item, SESSION_DBG_ZERO_DST_MAC);
            return IFX_PPA_SESSION_NOT_ADDED;
        }

        if ( ppa_hw_add_session(p_item) != 0 )
        {
            dbg("ppa_hw_add_session(p_item) fail");
            return IFX_PPA_SESSION_NOT_ADDED;
        }

        dbg("hardware added");
//        printk("--- hardware added: %08X->%04x ---\n", (u32)ppa_buf->nh.iph, (u32)ppa_buf->nh.iph->frag_off);

#if 0
        {
            PPA_SESSION_EXTRA extra = {0};

            extra.vlan_prio = 0;
            extra.vlan_cfi  = 0;
            extra.vlan_id   = 5;

            if ( (ppa_get_iftype(p_item->rx_if) & IFX_PPA_IF_TYPE_WAN) )
            {
                //printk("%s - WAN\n", p_item->rx_if->name);
                extra.vlan_remove = 1;
            }
            else
            {
                //printk("%s - LAN\n", p_item->rx_if->name);
                extra.vlan_insert = 1;
            }
            ifx_ppa_session_modify(p_session, &extra, PPA_F_SESSION_VLAN | (flags & PPA_F_SESSION_REPLY_DIR));
        }
#endif

        return IFX_PPA_SESSION_ADDED;
    }
}

int32_t ifx_ppa_session_delete(PPA_SESSION *p_session, uint32_t flags)
{
    int32_t ret = -1;
    uint32_t flag_template[2] = {PPA_F_SESSION_ORG_DIR, PPA_F_SESSION_REPLY_DIR};
    struct session_list_item *p_item;
    uint32_t i;

    for ( i = 0; i < NUM_ENTITY(flag_template); i++ )
    {
        if ( !(flags & flag_template[i]) )
            continue;

        //  i = 0, org dir, i = 1, reply dir
        if ( ppa_lookup_session(p_session, i, &p_item) != IFX_PPA_SESSION_EXISTS )
        {
            //printk("try to delete none existing session\n");
            continue;
        }

        dump_list_item(p_item, "ifx_ppa_session_delete");

        ppa_hw_del_session(p_item);

        ppa_remove_session_item(p_item);

        ppa_mem_cache_free(p_item, g_session_item_cache);

        ret = 0;
    }

    return ret;
}

int32_t ifx_ppa_session_modify(PPA_SESSION *p_session, PPA_SESSION_EXTRA *p_extra, uint32_t flags)
{
    struct session_list_item *p_item;

    if ( ppa_lookup_session(p_session, flags & PPA_F_SESSION_REPLY_DIR, &p_item) == IFX_PPA_SESSION_EXISTS )
    {
        ppa_update_session_extra(p_extra, p_item, flags);
        if ( (p_item->flags & SESSION_ADDED_IN_HW) )
        {
            if ( ppa_hw_update_session_extra(p_item, flags) )
            {
                //  update failed
                ppa_hw_del_session(p_item);
                p_item->flags &= ~SESSION_ADDED_IN_HW;
                return -1;
            }
        }
        return 0;
    }
    else
        return -1;
}

int32_t ifx_ppa_session_get(PPA_SESSION ***pp_sessions, PPA_SESSION_EXTRA **pp_extra, int32_t *p_num_entries, uint32_t flags)
{
    PPA_SESSION **p_sessions = NULL;
    PPA_SESSION_EXTRA *p_extra = NULL;
    uint32_t num = 0;
    struct session_list_item *p_cur;
    uint32_t i;

    ppa_lock_get(&g_session_list_lock);

    if ( g_session_list_length )
    {
        p_sessions = (PPA_SESSION **)ppa_malloc(sizeof(PPA_SESSION *) * g_session_list_length);
        if ( !p_sessions )
            goto IFX_PPA_SESSION_GET_NO_MEMORY;

        p_extra = (PPA_SESSION_EXTRA *)ppa_malloc(sizeof(PPA_SESSION_EXTRA) * g_session_list_length);
        if ( !p_extra )
            goto IFX_PPA_SESSION_GET_NO_MEMORY;

        for ( i = 0; i < NUM_ENTITY(g_session_list_hash_table); i++ )
            for ( p_cur = g_session_list_hash_table[i]; p_cur; p_cur = p_cur->next )
            {
                p_sessions[num] = p_cur->session;
                p_extra[num] = p_cur->session_extra;
                num++;
            }
    }

    ppa_lock_release(&g_session_list_lock);

    if ( pp_sessions )
        *pp_sessions = p_sessions;
    else
        ppa_free(p_sessions);

    if ( pp_extra )
        *pp_extra = p_extra;
    else
        ppa_free(p_extra);

    if ( p_num_entries )
        *p_num_entries = num;

    return 0;

IFX_PPA_SESSION_GET_NO_MEMORY:
    ppa_lock_release(&g_session_list_lock);
    if ( p_sessions )
        ppa_free(p_sessions);
    return -1;
}

#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API

int32_t ifx_ppa_mc_session_add(PPA_BUF *ppa_buf, PPA_MC_SESSION *p_mc_session, uint32_t flags)
{
    struct session_list_item *p_item;

    if ( ppa_lookup_mc_session(p_mc_session->ip_mc_group, &p_item) == IFX_PPA_SESSION_EXISTS )
        return IFX_PPA_SESSION_EXISTS;

    if ( ppa_add_mc_session(p_mc_session, &p_item, flags) != 0 )
        return IFX_PPA_SESSION_NOT_ADDED;

    if ( ppa_hw_add_mc_session(p_item) != 0 )
        return IFX_PPA_SESSION_NOT_ADDED;

    return IFX_PPA_SESSION_ADDED;
}

int32_t ifx_ppa_mc_session_delete(uint32_t ip_mc_group, uint32_t flags)
{
    struct session_list_item *p_item;

    if ( ppa_lookup_mc_session(ip_mc_group, &p_item) == IFX_PPA_SESSION_EXISTS )
        return -1;

    ppa_hw_del_mc_session(p_item);

    ppa_remove_mc_session_item(p_item);

    ppa_mem_cache_free(p_item, g_session_item_cache);

    return 0;
}

int32_t ifx_ppa_mc_session_modify(uint32_t ip_mc_group, PPA_MC_SESSION *p_mc_session, PPA_SESSION_EXTRA *p_extra, uint32_t flags)
{
    struct session_list_item *p_item;

    if ( ppa_lookup_mc_session(ip_mc_group, &p_item) == IFX_PPA_SESSION_EXISTS )
    {
        ppa_update_mc_session_extra(p_extra, p_item, flags);
        if ( (p_item->flags & SESSION_ADDED_IN_HW) )
            ppa_hw_update_mc_session_extra(p_item, flags);
        return 0;
    }
    else
        return -1;
}

int32_t ifx_ppa_all_mc_session_get(PPA_MC_SESSION ***pp_mc_sessions, PPA_SESSION_EXTRA **pp_extra, int32_t *p_num_entries, uint32_t flags)
{
    PPA_MC_SESSION **p_sessions = NULL;
    PPA_SESSION_EXTRA *p_extra = NULL;
    uint32_t num = 0;
    struct session_list_item *p_cur;
    uint32_t i;

    ppa_lock_get(&g_mc_session_list_lock);

    if ( g_session_list_length )
    {
        p_sessions = (PPA_MC_SESSION **)ppa_malloc(sizeof(PPA_MC_SESSION *) * g_session_list_length);
        if ( !p_sessions )
            goto IFX_PPA_ALL_MC_SESSION_GET_NO_MEMORY;

        p_extra = (PPA_SESSION_EXTRA *)ppa_malloc(sizeof(PPA_SESSION_EXTRA) * g_session_list_length);
        if ( !p_extra )
            goto IFX_PPA_ALL_MC_SESSION_GET_NO_MEMORY;

        for ( i = 0; i < NUM_ENTITY(g_mc_session_list_hash_table); i++ )
            for ( p_cur = g_mc_session_list_hash_table[i]; p_cur; p_cur = p_cur->next )
            {
                p_sessions[num] = (PPA_MC_SESSION *)p_cur->session;
                p_extra[num] = p_cur->session_extra;
                num++;
            }
    }

    ppa_lock_release(&g_mc_session_list_lock);

    if ( pp_mc_sessions )
        *pp_mc_sessions = p_sessions;
    else
        ppa_free(pp_mc_sessions);

    if ( pp_extra )
        *pp_extra = p_extra;
    else
        ppa_free(p_extra);

    if ( p_num_entries )
        *p_num_entries = num;

    return 0;

IFX_PPA_ALL_MC_SESSION_GET_NO_MEMORY:
    ppa_lock_release(&g_session_list_lock);
    if ( p_sessions )
        ppa_free(p_sessions);
    return -1;
}

int32_t ifx_ppa_mc_session_get(uint32_t ip_mc_group, PPA_MC_SESSION **p_mc_session, PPA_SESSION_EXTRA *p_extra, uint32_t flags)
{
    struct session_list_item *p_item;

    if ( ppa_lookup_mc_session(ip_mc_group, &p_item) != IFX_PPA_SESSION_EXISTS )
        return -1;

    if ( p_mc_session )
        *p_mc_session = (PPA_MC_SESSION *)p_item->session;

    if ( p_extra )
        *p_extra = p_item->session_extra;

    return 0;
}

#else

int32_t ifx_ppa_mc_group_update(PPA_MC_GROUP *ppa_mc_entry, uint32_t flags)
{
    struct mc_group_list_item *p_item;

    if ( ppa_lookup_mc_group(ppa_mc_entry->ip_mc_group, &p_item) != IFX_PPA_SESSION_EXISTS )
    {
        //  add

        if ( ppa_add_mc_group(ppa_mc_entry, &p_item, flags) != 0 )
            return IFX_PPA_SESSION_NOT_ADDED;

        if ( ppa_hw_add_mc_group(p_item) != 0 )
        {
            dbg("ppa_hw_add_mc_group(%d.%d.%d.%d): fail", ppa_mc_entry->ip_mc_group >> 24, (ppa_mc_entry->ip_mc_group >> 16) & 0xFF, (ppa_mc_entry->ip_mc_group >> 8) & 0xFF, ppa_mc_entry->ip_mc_group & 0xFF);
            return IFX_PPA_SESSION_NOT_ADDED;
        }

        dbg("hardware added");

        return IFX_PPA_SESSION_ADDED;
    }
    else
    {
        //  update
        //  now, I only add interface, ignore VLAN/PPPoE change
        PPA_NETIF *netif;
        uint32_t idx;
        uint32_t bit;
        uint32_t i;

        //ppa_memset(p_item->netif, 0, MAX_MC_IFS_NUM * sizeof(*p_item->netif));
        //ppa_memset(p_item->ttl, 0, MAX_MC_IFS_NUM * sizeof(*p_item->ttl));

        for ( i = 0, bit = 1, idx = 0; i < MAX_MC_IFS_NUM && idx < ppa_mc_entry->num_ifs; i++, bit <<= 1 )
            if ( (ppa_mc_entry->if_mask & bit) )
            {
                netif = ppa_get_netif(ppa_mc_entry->array_mem_ifs[i].ifname);
                if ( netif && (ppa_get_iftype(netif) &  IFX_PPA_IF_TYPE_WAN) )
                {
                    p_item->netif[idx] = netif;
                    p_item->ttl[idx]   = ppa_mc_entry->array_mem_ifs[i].ttl;
                    idx++;
                }
            }

        p_item->num_ifs = idx;
        p_item->if_mask = (1 << idx) - 1;

        //  if no interface, delete
        if ( p_item->num_ifs == 0 )
        {
            ppa_hw_del_mc_group(p_item);

            ppa_remove_mc_group_item(p_item);

            ppa_mem_cache_free(p_item, g_mc_group_item_cache);
        }
    }

    return 0;
}

int32_t ifx_ppa_mc_group_get(IPADDR ip_mc_group, PPA_MC_GROUP *ppa_mc_entry, uint32_t flags)
{
    struct mc_group_list_item *p_item;
    struct mc_group_list_item item;
    PPA_IFNAME *ifname;
    uint32_t idx;
    uint32_t bit;
    uint32_t i;

//    if ( !ppa_mc_entry )
//        return -1;

    if ( ppa_lookup_mc_group(ip_mc_group, &p_item) != IFX_PPA_SESSION_EXISTS )
        return -1;

    item = *p_item;

    ppa_mc_entry->ip_mc_group = item.ip_mc_group;

    for ( i = 0, bit = 1, idx = 0; i < MAX_MC_IFS_NUM; i++, bit <<= 1 )
        if ( (item.if_mask & bit) )
        {
            ifname = ppa_get_netif_name(item.netif[i]);
            if ( ifname )
            {
                ppa_mc_entry->array_mem_ifs[idx].ifname = ifname;
                ppa_mc_entry->array_mem_ifs[idx].ttl    = item.ttl[i];
                idx++;
            }
        }

    ppa_mc_entry->num_ifs = idx;
    ppa_mc_entry->if_mask = (1 << idx) - 1;

    return 0;
}

#endif

int32_t ifx_ppa_set_session_inactivity(PPA_U_SESSION *p_session, int32_t timeout)
{
#if 0
    //  can't handle bi-directional case

    struct session_list_item *p_item;

    if ( ppa_lookup_session((PPA_SESSION *)p_session, 0, &p_item) != IFX_PPA_SESSION_EXISTS                         //  org dir
        && ppa_lookup_session((PPA_SESSION *)p_session, 1, &p_item) != IFX_PPA_SESSION_EXISTS                       //  reply dir
  #if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
        && ppa_lookup_mc_session(((PPA_MC_SESSION *)p_session)->ip_mc_group, &p_item) != IFX_PPA_SESSION_EXISTS     //  multicast
  #endif
       )
        return -1;

  #if 0
    if ( !(p_item->flags & SESSION_ADDED_IN_HW) )
        return 0;
  #endif

    p_item->timeout = timeout;

    return 0;

#else

    int32_t ret, ret_reply;
    struct session_list_item *p_item, *p_item_reply;

    ret = ppa_lookup_session((PPA_SESSION *)p_session, 0, &p_item);
    ret_reply = ppa_lookup_session((PPA_SESSION *)p_session, 1, &p_item_reply);
    if ( ret != IFX_PPA_SESSION_EXISTS && ret_reply != IFX_PPA_SESSION_EXISTS
  #if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
        && (ret = ppa_lookup_mc_session(((PPA_MC_SESSION *)p_session)->ip_mc_group, &p_item)) != IFX_PPA_SESSION_EXISTS
  #endif
       )
        return -1;

  #if 0
    printk("set timeout %d\n", timeout);
  #endif

    if ( ret == IFX_PPA_SESSION_EXISTS )
        p_item->timeout = timeout;

    if ( ret_reply == IFX_PPA_SESSION_EXISTS )
        p_item_reply->timeout = timeout;

    if ( timeout < g_hit_polling_time )
    {
        //  remove timer
        ppa_timer_del(&g_hit_stat_timer);

        //  timeout can not be zero
        g_hit_polling_time = timeout < 1 ? 1 : timeout;

        //  check hit stat in case the left time is less then the new timeout
        ppa_check_hit_stat(0);  //  timer is added in this function

        //  add timer
//        ppa_timer_add(&g_hit_stat_timer, g_hit_polling_time);
    }

    return 0;

#endif
}

int32_t ifx_ppa_inactivity_status(PPA_U_SESSION *p_session)
{
#if 0
    //  can't handle bi-directional case

    struct session_list_item *p_item;

    if ( ppa_lookup_session((PPA_SESSION *)p_session, 0, &p_item) != IFX_PPA_SESSION_EXISTS                         //  org dir
        && ppa_lookup_session((PPA_SESSION *)p_session, 1, &p_item) != IFX_PPA_SESSION_EXISTS                       //  reply dir
  #if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
        && ppa_lookup_mc_session(((PPA_MC_SESSION *)p_session)->ip_mc_group, &p_item) != IFX_PPA_SESSION_EXISTS     //  multicast
  #endif
       )
        return IFX_PPA_SESSION_NOT_ADDED;

    //  not added in hardware
    if ( !(p_item->flags & SESSION_ADDED_IN_HW) )
        return IFX_PPA_SESSION_NOT_ADDED;

    if ( p_item->timeout + p_item->last_hit_time < ppa_get_time_in_sec() )  //  use < other than <= to avoid "false positives"
        return IFX_PPA_TIMEOUT;

    return IFX_PPA_HIT;

#else

    int f_flag;
    int f_timeout;
    int32_t ret, ret_reply;
    struct session_list_item *p_item, *p_item_reply;

    ret = ppa_lookup_session((PPA_SESSION *)p_session, 0, &p_item);
    ret_reply = ppa_lookup_session((PPA_SESSION *)p_session, 1, &p_item_reply);
    if ( ret != IFX_PPA_SESSION_EXISTS && ret_reply != IFX_PPA_SESSION_EXISTS
  #if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
        && (ret = ppa_lookup_mc_session(((PPA_MC_SESSION *)p_session)->ip_mc_group, &p_item)) != IFX_PPA_SESSION_EXISTS
  #endif
       )
    {
  #if 0
        printk("no such session %08x\n", (unsigned int)p_session);
  #endif
        return IFX_PPA_SESSION_NOT_ADDED;
    }

    f_flag = 0;
    f_timeout = 1;

    if ( ret == IFX_PPA_SESSION_EXISTS && (p_item->flags & SESSION_ADDED_IN_HW) )
    {
        f_flag = 1;
  #if 1
        if ( p_item->timeout + p_item->last_hit_time > ppa_get_time_in_sec() )
            f_timeout = 0;
  #else
        uint32_t tmp = ppa_get_time_in_sec();
        if ( p_item->timeout + p_item->last_hit_time > tmp )
        {
            printk("session %08x, p_item->timeout (%d) + p_item->last_hit_time (%d) > ppa_get_time_in_sec (%d)\n", (unsigned int)p_session, p_item->timeout, p_item->last_hit_time, tmp);
            f_timeout = 0;
        }
        else
            printk("session %08x, p_item->timeout (%d) + p_item->last_hit_time (%d) <= ppa_get_time_in_sec (%d)\n", (unsigned int)p_session, p_item->timeout, p_item->last_hit_time, tmp);
  #endif
    }

    if ( ret_reply == IFX_PPA_SESSION_EXISTS && (p_item_reply->flags & SESSION_ADDED_IN_HW) )
    {
        f_flag = 1;
  #if 1
        if ( p_item_reply->timeout + p_item_reply->last_hit_time > ppa_get_time_in_sec() )
            f_timeout = 0;
  #else
        uint32_t tmp = ppa_get_time_in_sec();
        if ( p_item_reply->timeout + p_item_reply->last_hit_time > tmp )
        {
            printk("session %08x, p_item_reply->timeout (%d) + p_item_reply->last_hit_time (%d) > ppa_get_time_in_sec (%d)\n", (unsigned int)p_session, p_item_reply->timeout, p_item_reply->last_hit_time, tmp);
            f_timeout = 0;
        }
        else
            printk("session %08x, p_item_reply->timeout (%d) + p_item_reply->last_hit_time (%d) <= ppa_get_time_in_sec (%d)\n", (unsigned int)p_session, p_item_reply->timeout, p_item_reply->last_hit_time, tmp);
  #endif
    }

    //  not added in hardware
    if ( !f_flag )
    {
  #if 0
        printk("no such session %08x, !f_flag\n", (unsigned int)p_session);
  #endif
        return IFX_PPA_SESSION_NOT_ADDED;
    }

    return f_timeout ? IFX_PPA_TIMEOUT : IFX_PPA_HIT;

#endif
}

int32_t ifx_ppa_get_if_stats(PPA_IFNAME *ifname, PPA_IF_STATS *p_stats, uint32_t flags)
{
#if 0
    uint32_t if_type;

    uint32_t fast_uc_tcp_pkts;
    uint32_t fast_uc_udp_pkts;
    uint32_t fast_mc_tcp_pkts;
    uint32_t fast_mc_udp_pkts;
    uint32_t fast_drop_tcp_pkts;
    uint32_t fast_drop_udp_pkts;
    uint32_t drop_pkts;

    uint32_t rx_fast_pkts;
    uint32_t rx_fast_bytes;
    uint32_t rx_cpu_pkts;
    uint32_t rx_cpu_bytes;
    uint32_t rx_drop_pkts;
    uint32_t rx_drop_bytes;
    uint32_t tx_fast_pkts;
    uint32_t tx_fast_bytes;

    if ( !p_stats )
        return -1;

    if_type = ppa_get_iftype(ppa_get_netif(ifname));

    if ( if_type == IFX_PPA_IF_NOT_FOUND )
        return -1;

    switch ( g_fwcode )
    {
    case FWCODE_ROUTING_ACC_D2:
        if ( (if_type & IFX_PPA_IF_TYPE_LAN) )
        {
            ppa_memset(p_stats, 0, sizeof(*p_stats));

            get_lan_rx_mib(&fast_uc_tcp_pkts,
                           &fast_uc_udp_pkts,
                           &fast_drop_tcp_pkts,
                           &fast_drop_udp_pkts,
                           &drop_pkts);

            p_stats->rx_pkts         = fast_uc_tcp_pkts + fast_uc_udp_pkts;
            p_stats->rx_discard_pkts = fast_drop_tcp_pkts + fast_drop_udp_pkts;
            p_stats->tx_pkts         = p_stats->rx_pkts;
        }
        else
        {
            ppa_memset(p_stats, 0, sizeof(*p_stats));

            get_wan_rx_mib(&fast_uc_tcp_pkts,
                           &fast_uc_udp_pkts,
                           &fast_mc_tcp_pkts,
                           &fast_mc_udp_pkts,
                           &fast_drop_tcp_pkts,
                           &fast_drop_udp_pkts,
                           &drop_pkts);

            p_stats->rx_pkts         = fast_uc_tcp_pkts + fast_uc_udp_pkts + fast_mc_tcp_pkts + fast_mc_udp_pkts;
            p_stats->rx_discard_pkts = fast_drop_tcp_pkts + fast_drop_udp_pkts;
            p_stats->tx_pkts         = p_stats->rx_pkts;
        }
        break;
    case FWCODE_BRIDGING_ACC_D3:
        if ( (if_type & IFX_PPA_IF_TYPE_LAN) )
        {
            ppa_memset(p_stats, 0, sizeof(*p_stats));

            get_bridging_eth_mib(1,
                                 &rx_fast_pkts,
                                 &rx_fast_bytes,
                                 &rx_cpu_pkts,
                                 &rx_cpu_bytes,
                                 &rx_drop_pkts,
                                 &rx_drop_bytes,
                                 &tx_fast_pkts,
                                 &tx_fast_bytes);

            p_stats->rx_pkts         = rx_fast_pkts;
            p_stats->rx_bytes        = rx_fast_bytes;
            p_stats->rx_discard_pkts = rx_drop_pkts;
            p_stats->tx_pkts         = tx_fast_pkts;
            p_stats->tx_bytes        = tx_fast_bytes;
        }
        else
        {
            ppa_memset(p_stats, 0, sizeof(*p_stats));

            get_bridging_eth_mib(0,
                                 &rx_fast_pkts,
                                 &rx_fast_bytes,
                                 &rx_cpu_pkts,
                                 &rx_cpu_bytes,
                                 &rx_drop_pkts,
                                 &rx_drop_bytes,
                                 &tx_fast_pkts,
                                 &tx_fast_bytes);

            p_stats->rx_pkts         = rx_fast_pkts;
            p_stats->rx_bytes        = rx_fast_bytes;
            p_stats->rx_discard_pkts = rx_drop_pkts;
            p_stats->tx_pkts         = tx_fast_pkts;
            p_stats->tx_bytes        = tx_fast_bytes;
        }
        break;
    }

    return 0;
#else
    return -1;
#endif
}

int32_t ifx_ppa_get_accel_stats(PPA_ACCEL_STATS *p_lan_rx_accel_stats, PPA_ACCEL_STATS *p_wan_rx_accel_stats, uint32_t flags)
{
#if 0
    uint32_t fast_uc_tcp_pkts;
    uint32_t fast_uc_udp_pkts;
    uint32_t fast_mc_tcp_pkts;
    uint32_t fast_mc_udp_pkts;
    uint32_t fast_drop_tcp_pkts;
    uint32_t fast_drop_udp_pkts;
    uint32_t drop_pkts;

    if ( p_lan_rx_accel_stats )
    {
        ppa_memset(p_lan_rx_accel_stats, 0, sizeof(*p_lan_rx_accel_stats));

        get_lan_rx_mib(&fast_uc_tcp_pkts,
                       &fast_uc_udp_pkts,
                       &fast_drop_tcp_pkts,
                       &fast_drop_udp_pkts,
                       &drop_pkts);

        p_lan_rx_accel_stats->fast_tcp_pkt_tx    = fast_uc_tcp_pkts;
        p_lan_rx_accel_stats->fast_udp_pkt_tx    = fast_uc_udp_pkts;
        p_lan_rx_accel_stats->fast_tcp_pkt_drop  = fast_drop_tcp_pkts;
        p_lan_rx_accel_stats->fast_udp_pkt_drop  = fast_drop_udp_pkts;
    }

    if ( p_wan_rx_accel_stats )
    {
        ppa_memset(p_wan_rx_accel_stats, 0, sizeof(*p_wan_rx_accel_stats));

        get_wan_rx_mib(&fast_uc_tcp_pkts,
                       &fast_uc_udp_pkts,
                       &fast_mc_tcp_pkts,
                       &fast_mc_udp_pkts,
                       &fast_drop_tcp_pkts,
                       &fast_drop_udp_pkts,
                       &drop_pkts);

        p_wan_rx_accel_stats->fast_tcp_pkt_tx       = fast_uc_tcp_pkts;
        p_wan_rx_accel_stats->fast_udp_pkt_tx       = fast_uc_udp_pkts;
        p_wan_rx_accel_stats->fast_udp_mcast_pkt_tx = fast_mc_udp_pkts;
        p_wan_rx_accel_stats->fast_tcp_pkt_drop     = fast_drop_tcp_pkts;
        p_wan_rx_accel_stats->fast_udp_pkt_drop     = fast_drop_udp_pkts;
    }

    return 0;
#else
    return -1;
#endif
}

#if 0
int32_t ifx_ppa_filter_add(PPA_FILTER *p_filter, PPA_ACTION *p_action, uint32_t flags)
{
    return 0;
}
#endif

int32_t ifx_ppa_set_if_mac_address(PPA_IFNAME *ifname, uint8_t *mac_addr)
{
    int32_t ret = -1;
    struct netif_info *pinfo;

    lock_netif_info_list();
    for ( pinfo = g_netif; pinfo; pinfo = pinfo->next )
    {
        if ( strcmp(ifname, pinfo->name) == 0 )
        {
            if ( *(uint32_t *)pinfo->mac != *(uint32_t *)mac_addr || *((uint16_t *)pinfo->mac + 2) != *((uint16_t *)mac_addr + 2) )
            {
                *(uint32_t *)pinfo->mac = *(uint32_t *)mac_addr;
                *((uint16_t *)pinfo->mac + 2) = *((uint16_t *)mac_addr + 2);

                if ( (pinfo->flags & NETIF_MAC_ENTRY_CREATED) )
                {
                    del_mac_entry(pinfo->mac_entry);
                    pinfo->mac_entry = ~0;
                    pinfo->flags &= ~NETIF_MAC_ENTRY_CREATED;
                }
                if ( add_mac_entry(mac_addr, &pinfo->mac_entry) == 0 )
                    pinfo->flags |= NETIF_MAC_ENTRY_CREATED;
            }
            break;
        }
    }
    unlock_netif_info_list();

    return ret;
}

int32_t ifx_ppa_get_if_mac_address(PPA_IFNAME *ifname, uint8_t *mac_addr)
{
    int32_t ret = -1;
    struct netif_info *pinfo;

    lock_netif_info_list();
    for ( pinfo = g_netif; pinfo; pinfo = pinfo->next )
    {
        if ( strcmp(ifname, pinfo->name) == 0 )
        {
            *(uint32_t *)mac_addr = *(uint32_t *)pinfo->mac;
            *((uint16_t *)mac_addr + 2) = *((uint16_t *)pinfo->mac + 2);
            break;
        }
    }
    unlock_netif_info_list();

    return ret;
}

int32_t ifx_ppa_bridge_entry_add(uint8_t *mac_addr, PPA_NETIF *netif, uint32_t flags)
{
    struct bridging_session_list_item *p_item;

    if ( ppa_bridging_lookup_session(mac_addr, netif, &p_item) == IFX_PPA_SESSION_EXISTS )
        return IFX_PPA_SESSION_ADDED;
    else if ( ppa_bridging_add_session(mac_addr, netif, &p_item, flags) != 0 )
        return IFX_PPA_SESSION_NOT_ADDED;

    if ( ppa_bridging_hw_add_session(p_item) != 0 )
    {
        dbg("ppa_bridging_hw_add_session(%02x:%02x:%02x:%02x:%02x:%02x): fail", (uint32_t)p_item->mac[0], (uint32_t)p_item->mac[1], (uint32_t)p_item->mac[2], (uint32_t)p_item->mac[3], (uint32_t)p_item->mac[4], (uint32_t)p_item->mac[5]);
        return IFX_PPA_SESSION_NOT_ADDED;
    }

    dbg("hardware added");

    return IFX_PPA_SESSION_ADDED;
}

int32_t ifx_ppa_bridge_entry_delete(uint8_t *mac_addr, uint32_t flags)
{
    struct bridging_session_list_item *p_item;

    if ( ppa_bridging_lookup_session(mac_addr, NULL, &p_item) != IFX_PPA_SESSION_EXISTS )
        return -1;

    dump_bridging_list_item(p_item, "ifx_ppa_bridge_entry_delete");

    ppa_bridging_hw_del_session(p_item);

    ppa_bridging_remove_session_item(p_item);

    ppa_mem_cache_free(p_item, g_bridging_session_item_cache);

    return 0;
}

int32_t ifx_ppa_set_bridge_entry_timeout(uint8_t *mac_addr, uint32_t timeout)
{
    struct bridging_session_list_item *p_item;

    if ( ppa_bridging_lookup_session(mac_addr, NULL, &p_item) != IFX_PPA_SESSION_EXISTS )
        return -1;

    if ( !(p_item->flags & SESSION_STATIC) )
        p_item->timeout = timeout;

    if ( timeout < g_bridging_hit_polling_time )
    {
        //  remove timer
        ppa_timer_del(&g_hit_stat_timer);

        //  timeout can not be zero
        g_bridging_hit_polling_time = timeout < 1 ? 1 : timeout;

        //  check hit stat in case the left time is less then the new timeout
        ppa_bridging_check_hit_stat(0); //  timer is added in this function

        //  add timer
//        ppa_timer_add(&g_hit_stat_timer, g_bridging_hit_polling_time);
    }

    return 0;
}

int32_t ifx_ppa_bridge_entry_inactivity_status(uint8_t *mac_addr)
{
    struct bridging_session_list_item *p_item;

    if ( ppa_bridging_lookup_session(mac_addr, NULL, &p_item) != IFX_PPA_SESSION_EXISTS )
        return IFX_PPA_SESSION_NOT_ADDED;

    //  not added in hardware
    if ( !(p_item->flags & SESSION_ADDED_IN_HW) )
        return IFX_PPA_SESSION_NOT_ADDED;

    if ( (p_item->flags & SESSION_STATIC) )
        return IFX_PPA_HIT;

    if ( p_item->timeout + p_item->last_hit_time < ppa_get_time_in_sec() )  //  use < other than <= to avoid "false positives"
        return IFX_PPA_TIMEOUT;

    return IFX_PPA_HIT;
}

int32_t ifx_ppa_bridge_entry_hit_time(uint8_t *mac_addr, uint32_t *p_hit_time)
{
    struct bridging_session_list_item *p_item;

    if ( ppa_bridging_lookup_session(mac_addr, NULL, &p_item) != IFX_PPA_SESSION_EXISTS )
        return IFX_PPA_SESSION_NOT_ADDED;

    *p_hit_time = p_item->last_hit_time;
    return IFX_PPA_HIT;
}

int32_t ifx_ppa_set_bridge_if_vlan_config(PPA_NETIF *netif, VLAN_TAG_CTRL *vlan_tag_control, VLAN_CFG *vlan_cfg, uint32_t flags)
{
    int32_t ifid;

    ifid = ppa_get_ifid(netif);
    if ( ifid < 0 )
        return -1;

    set_bridge_if_vlan_config(ifid,
                              vlan_tag_control->insertion | vlan_tag_control->replace,
                              vlan_tag_control->remove | vlan_tag_control->replace,
                              vlan_cfg->vlan_aware | (vlan_tag_control->unmodified ? 0 : 1) | vlan_tag_control->insertion | vlan_tag_control->remove | vlan_tag_control->replace,
                              vlan_cfg->src_ip_based_vlan,
                              vlan_cfg->eth_type_based_vlan,
                              vlan_cfg->vlanid_based_vlan
#ifdef CONFIG_IFX_PPA_A4
                              ,
                              vlan_cfg->port_based_vlan,
                              vlan_tag_control->out_insertion | vlan_tag_control->out_replace,
                              vlan_tag_control->out_remove | vlan_tag_control->out_replace,
                              vlan_cfg->out_vlan_aware | (vlan_tag_control->out_unmodified ? 0 : 1) | vlan_tag_control->out_insertion | vlan_tag_control->out_remove | vlan_tag_control->out_replace
#endif
                              );

    return 0;
}

int32_t ifx_ppa_get_bridge_if_vlan_config(PPA_NETIF *netif, VLAN_TAG_CTRL *vlan_tag_control, VLAN_CFG *vlan_cfg, uint32_t flags)
{
    int32_t ifid;
    uint32_t f_eg_vlan_insert;
    uint32_t f_eg_vlan_remove;
    uint32_t f_ig_vlan_aware;
    uint32_t f_ig_src_ip_based;
    uint32_t f_ig_eth_type_based;
    uint32_t f_ig_vlanid_based;
#ifdef CONFIG_IFX_PPA_A4
    uint32_t f_ig_port_based;
    uint32_t f_eg_out_vlan_insert;
    uint32_t f_eg_out_vlan_remove;
    uint32_t f_ig_out_vlan_aware;
#endif

    ifid = ppa_get_ifid(netif);
    if ( ifid < 0 )
        return -1;

    get_bridge_if_vlan_config(ifid,
                              &f_eg_vlan_insert,
                              &f_eg_vlan_remove,
                              &f_ig_vlan_aware,
                              &f_ig_src_ip_based,
                              &f_ig_eth_type_based,
                              &f_ig_vlanid_based
#ifdef CONFIG_IFX_PPA_A4
                              ,
                              &f_ig_port_based,
                              &f_eg_out_vlan_insert,
                              &f_eg_out_vlan_remove,
                              &f_ig_out_vlan_aware
#endif
                              );

    vlan_tag_control->unmodified    = f_ig_vlan_aware ? 0 : 1;
    vlan_tag_control->insertion     = f_eg_vlan_insert ? 1 : 0;
    vlan_tag_control->remove        = f_eg_vlan_remove ? 1 : 0;
    vlan_tag_control->replace       = f_eg_vlan_insert && f_eg_vlan_remove ? 1 : 0;
    vlan_cfg->vlan_aware            = f_ig_vlan_aware ? 1 : 0;
    vlan_cfg->src_ip_based_vlan     = f_ig_src_ip_based ? 1 : 0;
    vlan_cfg->eth_type_based_vlan   = f_ig_eth_type_based ? 1 : 0;
    vlan_cfg->vlanid_based_vlan     = f_ig_vlanid_based ? 1 : 0;
#ifdef CONFIG_IFX_PPA_A4
    vlan_cfg->port_based_vlan       = f_ig_port_based ? 1 : 0;
    vlan_tag_control->out_unmodified    = f_ig_out_vlan_aware ? 0 : 1;
    vlan_tag_control->out_insertion     = f_eg_out_vlan_insert ? 1 : 0;
    vlan_tag_control->out_remove        = f_eg_out_vlan_remove ? 1 : 0;
    vlan_tag_control->out_replace       = f_eg_out_vlan_insert && f_eg_out_vlan_remove ? 1 : 0;
    vlan_cfg->out_vlan_aware            = f_ig_out_vlan_aware ? 1 : 0;
#endif

    return 0;
}

int32_t ifx_ppa_vlan_filter_add(PPA_VLAN_MATCH_FIELD *vlan_match_field, PPA_VLAN_INFO *vlan_info, uint32_t flags)
{
    uint32_t ig_criteria_type;
    uint32_t ig_criteria;
    uint32_t new_vci;
    uint32_t vlan_port_map;
    PPA_NETIF *netif = NULL;
    int32_t ifid;
    int i;

    switch ( vlan_match_field->match_flags )
    {
    case PPA_F_VLAN_FILTER_IFNAME:
        netif = ppa_get_netif(vlan_match_field->match_field.ifname);
        if ( !netif )
            return -1;
        ifid = ppa_get_ifid(netif);
        if ( ifid < 0 )
            return -1;
        ig_criteria_type    = IFX_PPA_BRG_VLAN_IG_COND_TYPE_DEF;
        ig_criteria         = (uint32_t)ifid;
        break;
    case PPA_F_VLAN_FILTER_IP_SRC:
        ig_criteria_type    = IFX_PPA_BRG_VLAN_IG_COND_TYPE_SRC_IP;
        ig_criteria         = vlan_match_field->match_field.ip_src;
        break;
    case PPA_F_VLAN_FILTER_ETH_PROTO:
        ig_criteria_type    = IFX_PPA_BRG_VLAN_IG_COND_TYPE_ETH_TYPE;
        ig_criteria         = vlan_match_field->match_field.eth_protocol;
        break;
    case PPA_F_VLAN_FILTER_VLAN_TAG:
        ig_criteria_type    = IFX_PPA_BRG_VLAN_IG_COND_TYPE_VLAN;
        ig_criteria         = vlan_match_field->match_field.ingress_vlan_tag;
        break;
    default:
        return -1;
    }

    new_vci             = vlan_info->vlan_vci;

    vlan_port_map = 0;
    for ( i = 0; i < vlan_info->num_ifs; i++ )
    {
        netif = ppa_get_netif(vlan_info->vlan_if_membership[i].ifname);
        if ( !netif )
            continue;
        ifid = ppa_get_ifid(netif);
        if ( ifid < 0 )
            continue;
        vlan_port_map |= 1 << ifid;
    }

    return add_vlan_map(ig_criteria_type,
                        ig_criteria,
                        new_vci,
                        vlan_port_map);
}

int32_t ifx_ppa_vlan_filter_del(PPA_VLAN_MATCH_FIELD *vlan_match_field, uint32_t flags)
{
    uint32_t ig_criteria_type;
    uint32_t ig_criteria;
    PPA_NETIF *netif;
    int32_t ifid;

    switch ( vlan_match_field->match_flags )
    {
    case PPA_F_VLAN_FILTER_IFNAME:
        netif = ppa_get_netif(vlan_match_field->match_field.ifname);
        if ( !netif )
            return -1;
        ifid = ppa_get_ifid(netif);
        if ( ifid < 0 )
            return -1;
        ig_criteria_type    = IFX_PPA_BRG_VLAN_IG_COND_TYPE_DEF;
        ig_criteria         = (uint32_t)ifid;
        break;
    case PPA_F_VLAN_FILTER_IP_SRC:
        ig_criteria_type    = IFX_PPA_BRG_VLAN_IG_COND_TYPE_SRC_IP;
        ig_criteria         = vlan_match_field->match_field.ip_src;
        break;
    case PPA_F_VLAN_FILTER_ETH_PROTO:
        ig_criteria_type    = IFX_PPA_BRG_VLAN_IG_COND_TYPE_ETH_TYPE;
        ig_criteria         = vlan_match_field->match_field.eth_protocol;
        break;
    case PPA_F_VLAN_FILTER_VLAN_TAG:
        ig_criteria_type    = IFX_PPA_BRG_VLAN_IG_COND_TYPE_VLAN;
        ig_criteria         = vlan_match_field->match_field.ingress_vlan_tag;
        break;
    default:
        return -1;
    }

    del_vlan_map(ig_criteria_type,
                 ig_criteria);

    return 0;
}

int32_t ifx_ppa_vlan_filter_get_all(uint32_t *num_filters, PPA_VLAN_FILTER_CONFIG *vlan_filters, uint32_t flags)
{
    //  TODO: implementation, too many memory allocation problem

    return -1;
}

int32_t ifx_ppa_add_itf(PPA_IFINFO *p_ifs, uint32_t is_lanif)
{
    struct netif_info *p_netif_info;
    PPA_NETIF *netif;
    uint8_t mac[PPA_ETH_ALEN];

    if ( !netif_info_is_added(p_ifs->ifname, &p_netif_info) )
    {
        p_netif_info = alloc_netif_info();
        if ( !p_netif_info )
            return -1;

        strcpy(p_netif_info->name, p_ifs->ifname);

        netif = ppa_get_netif(p_netif_info->name);
        if ( netif )
        {
            if (  ppa_get_physical_if(netif, NULL, p_netif_info->phys_netif_name) >= 0 )
                p_netif_info->flags |= NETIF_PHY_IF_GOT;

            if ( ppa_check_is_ppp_netif(netif) && ppa_check_is_pppoe_netif(netif) )
                p_netif_info->flags |= NETIF_PPPOE;
            else if ( ppa_if_is_vlan_if(netif, NULL) )
                p_netif_info->flags |= NETIF_VLAN;

#ifdef CONFIG_IFX_PPA_A4
            if ( (p_netif_info->flags & NETIF_PHY_IF_GOT) && (netif = ppa_get_netif(p_netif_info->phys_netif_name)) )
            {
                PPA_VCC *vcc;
                uint32_t dslwan_qid;

                if ( ppa_br2684_get_vcc(netif, &vcc) >= 0 && (dslwan_qid = get_dslwan_qid_with_vcc(vcc)) >= 0 )
                {
                    p_netif_info->flags |= NETIF_BR2684;
                    p_netif_info->dslwan_qid =dslwan_qid;
                }
            }
#endif

            ppa_get_netif_hwaddr(netif, mac);
            ASSERT(*(uint32_t *)mac != 0 || *((uint16_t *)mac + 2) != 0, "MAC address is all zero - %s", p_netif_info->name);

            ppa_memcpy(p_netif_info->mac, mac, PPA_ETH_ALEN);

            if ( add_mac_entry(mac, &p_netif_info->mac_entry) == 0 )
                p_netif_info->flags |= NETIF_MAC_ENTRY_CREATED;
            else
            {
                free_netif_info(p_netif_info);
                err("add_mac_entry failed");
                return -1;
            }
        }

        add_netif_info(p_netif_info);
    }
    else
    {
        netif = ppa_get_netif(p_netif_info->name);
        if ( netif )
        {
            if (  ppa_get_physical_if(netif, NULL, p_netif_info->phys_netif_name) >= 0 )
            p_netif_info->flags |= NETIF_PHY_IF_GOT;

            if ( ppa_check_is_ppp_netif(netif) && ppa_check_is_pppoe_netif(netif) )
                p_netif_info->flags |= NETIF_PPPOE;
            else if ( ppa_if_is_vlan_if(netif, NULL) )
                p_netif_info->flags |= NETIF_VLAN;

#ifdef CONFIG_IFX_PPA_A4
            if ( (p_netif_info->flags & NETIF_PHY_IF_GOT) && (netif = ppa_get_netif(p_netif_info->phys_netif_name)) )
            {
                PPA_VCC *vcc;
                uint32_t dslwan_qid;

                if ( ppa_br2684_get_vcc(netif, &vcc) >= 0 && (dslwan_qid = get_dslwan_qid_with_vcc(vcc)) >= 0 )
                {
                    p_netif_info->flags |= NETIF_BR2684;
                    p_netif_info->dslwan_qid =dslwan_qid;
                }
            }
#endif

            ppa_get_netif_hwaddr(netif, mac);
            ASSERT(*(uint32_t *)mac != 0 || *((uint16_t *)mac + 2) != 0, "MAC address is all zero - %s", p_netif_info->name);

            ppa_memcpy(p_netif_info->mac, mac, PPA_ETH_ALEN);

            if ( add_mac_entry(mac, &p_netif_info->mac_entry) == 0 )
                p_netif_info->flags |= NETIF_MAC_ENTRY_CREATED;
            else
            {
                free_netif_info(p_netif_info);
                err("add_mac_entry failed");
                return -1;
            }
        }
    }

    p_netif_info->flags |= is_lanif ? NETIF_LAN_IF : NETIF_WAN_IF;

    return 0;
}



/*
 * ####################################
 *           Init/Cleanup API
 * ####################################
 */

int __init ifx_ppa_driver_init(void)
{
    if ( ppa_mem_cache_create("ppa_session_item", sizeof(struct session_list_item), &g_session_item_cache) )
        goto IFX_PPA_DRIVER_INIT_SESSION_ITEM_FAIL;

#if defined(USE_NEW_MC_API) && USE_NEW_MC_API
    if ( ppa_mem_cache_create("mc_group_item", sizeof(struct mc_group_list_item), &g_mc_group_item_cache) )
        goto IFX_PPA_DRIVER_INIT_MC_GROUP_SESSION_ITEM_FAIL;
#endif

    if ( ppa_mem_cache_create("bridging_sess_item", sizeof(struct bridging_session_list_item), &g_bridging_session_item_cache) )
        goto IFX_PPA_DRIVER_INIT_BRIDGING_SESSION_ITEM_FAIL;

    proc_file_create();

#if defined(CONFIG_IFX_PPA_MODULE)
  #ifdef CONFIG_IFX_PPA_A4
    ifx_ppa_ppe_a4_hal_init();
  #else
    ifx_ppa_ppe_d4_hal_init();
  #endif
#endif

    ppa_lock_init(&g_netif_lock);

    ppa_lock_init(&g_session_list_lock);
#if !defined(USE_NEW_MC_API) || !USE_NEW_MC_API
    ppa_lock_init(&g_mc_session_list_lock);
#else
    ppa_lock_init(&g_mc_group_list_lock);
#endif

    ppa_lock_init(&g_bridging_session_list_lock);

#if 0
    //  do init for testing purpose
//    g_ppa_init_info.num_lanifs = 1; //  ignore br0 in this step
    ifx_ppa_init(NULL, 0);

    //  enable acceleration
    ifx_ppa_enable(1, 1, 0);
#endif

#if 0   //  move to init part
    //  XXX: start timer at very first time to avoid possible race condition
    ppa_timer_init(&g_hit_stat_timer, ppa_check_hit_stat);
    ppa_timer_add(&g_hit_stat_timer, g_hit_polling_time);
#endif

    ppa_hook_init_fn = ifx_ppa_init;
    ppa_hook_exit_fn = ifx_ppa_exit;

    register_chrdev(PPA_CHR_DEV_MAJOR, PPA_DEV_NAME, &g_ppa_dev_ops);


#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
    ifx_ppa_directpath_init(0);
    g_directpath_enable = 1;
#endif
    printk("PPA API --- init successfully\n");
    return 0;

IFX_PPA_DRIVER_INIT_BRIDGING_SESSION_ITEM_FAIL:
//    printk("IFX_PPA_DRIVER_INIT_BRIDGING_SESSION_ITEM_FAIL\n");
#if defined(USE_NEW_MC_API) && USE_NEW_MC_API
    ppa_mem_cache_destroy(g_mc_group_item_cache);
IFX_PPA_DRIVER_INIT_MC_GROUP_SESSION_ITEM_FAIL:
#endif
    ppa_mem_cache_destroy(g_session_item_cache);
IFX_PPA_DRIVER_INIT_SESSION_ITEM_FAIL:
    return -1;
}

void __exit ifx_ppa_driver_exit(void)
{
    unregister_chrdev(PPA_CHR_DEV_MAJOR, PPA_DEV_NAME);

    ppa_hook_init_fn = NULL;
    ppa_hook_exit_fn = NULL;

#if 0   //  move to exit part
    ppa_timer_del(&g_hit_stat_timer);
#endif

    ifx_ppa_exit();
#if defined(CONFIG_IFX_PPA_DIRECT_FASTPATH) || defined(CONFIG_IFX_PPA_INDIRECT_FASTPATH)
    ifx_ppa_directpath_exit(0);
    g_directpath_enable = 0;
#endif

#if defined(CONFIG_IFX_PPA_MODULE)
  #ifdef CONFIG_IFX_PPA_A4
    ifx_ppa_ppe_a4_hal_exit();
  #else
    ifx_ppa_ppe_d4_hal_exit();
  #endif
#endif

    proc_file_delete();

    ppa_mem_cache_destroy(g_bridging_session_item_cache);

#if defined(USE_NEW_MC_API) && USE_NEW_MC_API
    ppa_mem_cache_destroy(g_mc_group_item_cache);
#endif
    ppa_mem_cache_destroy(g_session_item_cache);
}

module_init(ifx_ppa_driver_init);
module_exit(ifx_ppa_driver_exit);
